# Copyright (c) 2007-2012 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement 

import os
from contextlib import contextmanager

from pshell import CommandFailed, CommandSignalled

from rebar.environment import PLATFORM, VIRTUALIZER
from rebar.buildstep import SimpleBuildStep
from rebar.util import generator, fail_on, join_abs


class Action(object):
    pass


class SimpleAction(Action, SimpleBuildStep):
    pass


class BaseExecuteAction(SimpleAction):
    cmd = None
    values = None

    def __init__(self, env, cmd, *values, **kwargs):
        self.cmd = cmd
        self.values = values

        super(BaseExecuteAction, self).__init__(env, **kwargs)

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        with self.open_shell() as sh:
            with self.output.open_important_stream() as stream:
                sh.follow_write(self.cmd, *self.values, **{'outfile': stream})

    def open_shell(self):
        raise NotImplementedError()


class ExecAction(BaseExecuteAction):
    using = None

    def __init__(self, *args, **kwargs):
        using = kwargs.pop('using', PLATFORM)
        if using not in (PLATFORM, VIRTUALIZER):
            raise ValueError(using)
        self.using = using

        super(ExecAction, self).__init__(*args, **kwargs)

    @contextmanager
    def open_shell(self):
        if self.using == PLATFORM:
            open_platform = self.virtualizer.open_platform
        elif self.using == VIRTUALIZER:
            open_platform = self.virtualizer.open_base_platform
        else:
            raise AssertionError(self.using)

        with open_platform() as platform:
            with platform.open_shell() as sh:
                yield sh


class ShellAction(BaseExecuteAction):
    @contextmanager
    def open_shell(self):
        yield self.user_shell


# TODO: Support permissions?

class CopyToAction(SimpleAction):
    src = None
    dst = None
    recurse = None
    dereference = None
    replace = None
    user = None
    group = None
    using = None

    def __init__(
      self,
      env,
      src,
      dst,
      recurse,
      dereference,
      replace,
      user,
      group,
      using = PLATFORM,
      **kwargs
    ):
        if using not in (PLATFORM, VIRTUALIZER):
            raise ValueError(using)

        self.src = src
        self.dst = dst
        self.recurse = recurse
        self.dereference = dereference
        self.replace = replace
        self.user = user
        self.group = group
        self.using = using

        super(CopyToAction, self).__init__(env, **kwargs)

    # FIXME: We need to handle OSError, and perhaps IOError?
    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        if self.using == PLATFORM:
            open_platform = self.virtualizer.open_platform
        elif self.using == VIRTUALIZER:
            open_platform = self.virtualizer.open_base_platform
        else:
            raise AssertionError(self.using)

        with open_platform() as platform:
            platform.copy_to(
              self.src,
              self.dst,
              recurse = self.recurse,
              dereference = self.dereference,
              replace = self.replace,
              user = self.user,
              group = self.group,
            )


class CopyFromAction(SimpleAction):
    src = None
    dst = None
    recurse = None
    dereference = None

    def __init__(
      self,
      env,
      src,
      dst,
      recurse,
      dereference,
      using = PLATFORM,
      **kwargs
    ):
        if using not in (PLATFORM, VIRTUALIZER):
            raise ValueError(using)

        self.src = src
        self.dst = dst
        self.recurse = recurse
        self.dereference = dereference
        self.using = using

        super(CopyFromAction, self).__init__(env, **kwargs)

    # FIXME: We need to handle OSError, and perhaps IOError?
    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        if self.using == PLATFORM:
            open_platform = self.virtualizer.open_platform
        elif self.using == VIRTUALIZER:
            open_platform = self.virtualizer.open_base_platform
        else:
            raise AssertionError(self.using)

        with open_platform() as platform:
            platform.copy_from(
              self.src,
              self.dst,
              recurse = self.recurse,
              dereference = self.dereference,
            )


class PatchAction(SimpleAction):
    patch = None
    dir = None
    file = None
    strip = None

    def __init__(self, env, patch, dir, file, strip, **kwargs):
        self.patch = patch
        self.dir = dir
        self.file = file
        self.strip = strip

        super(PatchAction, self).__init__(env, **kwargs)

    def get_required_packages(self):
        required_packages = super(PatchAction, self).get_required_packages()
        required_packages.add('patch')
        return required_packages

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        if not any((self.dir, self.file)):
            raise AssertionError('no dir or file specified')

        with open(self.patch, 'rb') as f:
            with self.virtualizer.open_base_platform() as platform:
                with platform.open_shell() as sh:
                    # Make sure patch is usable, otherwise things will fail in a
                    # bad way (we'll write our patch data directly to the shell and
                    # then send an EOF, exiting the shell).
                    sh.execute('which patch >/dev/null')

                    # Note: We use cat here to deal with buffering issues that (at
                    # least in some cases) lead to patch needing to see two EOFs
                    # before exiting.

                    if self.dir is not None:
                        dir_path = join_abs('/platform', self.dir)

                        sh.pushd(dir_path)
                        try:
                            sh.execute(
                              'cat - | patch --strip ?',
                              self.strip,
                              wait = False,
                            )
                            for line in f:
                                sh.write(line)
                            # Send a trailing newline in case the patch was missing
                            # one:
                            sh.write('\n')
                            sh.sendeof()
                            sh.wait_for_prompt()
                            sh.check_last_status()
                        finally:
                            sh.popd()

                    elif self.file is not None:
                        file_path = join_abs('/platform', self.file)

                        sh.execute('cat - | patch ?', file_path, wait = False)
                        for line in f:
                            sh.write(line)
                        # Send a trailing newline in case the patch was missing
                        # one:
                        sh.write('\n')
                        sh.sendeof()
                        sh.wait_for_prompt()
                        sh.check_last_status()
