# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement 

import os, re

from pshell import CommandFailed, CommandSignalled

from rebar.source_package import UnpackedSourcePackage, DscSourcePackage
from rebar.buildstep import SimpleBuildStep, ComplexBuildStep
from rebar.debootstrap import Debootstrapper
from rebar.util import generator, fail_on
from rebar import aptutil
from rebar.exceptions import InvalidSourcePackageError, NoSuchBuildPhase


# TODO: Handle time zone?


REPO_DIR = '/var/local/apt'


class NoSuchPackageError(Exception):
    pass


class PackageAlreadyInstalledError(Exception):
    pass


class VirtualPackageError(Exception):
    requested = None
    selected = None

    def __init__(self, requested, selected):
        super(VirtualPackageError, self).__init__(requested, selected)

        self.requested = requested
        self.selected = selected


class PackageSystem(object):
    INSTALL_PACKAGES = 'The following NEW packages will be installed:'
    UPGRADE_PACKAGES = 'The following packages will be upgraded:'

    GET = 'Get:'
    SELECTING = 'Selecting previously deselected package '
    UNPACKING = 'Unpacking '
    SETTING_UP = 'Setting up '

    REMOVING = 'Removing '
    PURGING = 'Purging configuration files for '

    REGEX_SELECTING = re.compile(
      r'^Note, selecting (?P<selected>\S+) instead of (?P<requested>\S+)'
    )
    REGEX_INST = re.compile(r'^Inst (?P<name>\S+) \((?P<version>\S+) \S+\)')

    def __init__(self, platform):
        super(PackageSystem, self).__init__()

        self.platform = platform

    def update(self):
        with self.platform.open_shell() as sh:
            sh.execute('apt-get --quiet update')

    def clean(self):
        with self.platform.open_shell() as sh:
            sh.execute('apt-get --quiet clean')

    def _calculate_progress(
      self,
      fetch_count,
      unpack_count,
      configure_count,
      package_count,
    ):
        if package_count > 0:
            total_steps = 3 * package_count
            current_steps = fetch_count + unpack_count + configure_count
            progress = 100.0 * (float(current_steps)/float(total_steps))
        else:
            progress = None
        return progress

    def _follow_apt(self, *args, **kwargs):
        # FIXME: Clean up comments.

        packages = []
        package_count = 0
        fetch_count = 0
        unpack_count = 0
        configure_count = 0
        in_package_list = False

        with self.platform.open_shell() as sh:
            for line in sh.follow_lines(*args, **kwargs):
                if in_package_list:
                    if line.startswith('  '):
                        packages.extend(line.strip().split())
                    else:
                        in_package_list = False
                        package_count = len(packages)
                    continue

                elif (
                  line.startswith(self.INSTALL_PACKAGES)
                  or line.startswith(self.UPGRADE_PACKAGES)
                ):
                    in_package_list = True
                    continue

                elif line.startswith(self.GET):
                    arg = line.split()[3]

                    # Check that this line is about a package, not just
                    # some message that resembles a fetch line without
                    # being about a particular package.
                    if arg not in packages:
                        continue

                    desc = 'fetch'

                elif line.startswith(self.SELECTING):
                    # FIXME: "select" actually precedes "unpack".
                    # Maybe they should be combined as "install".

                    ## Strip trailing "."
                    #arg = line.split()[-1][:-1]
                    #desc = 'select'

                    progress = self._calculate_progress(
                      fetch_count,
                      unpack_count,
                      configure_count,
                      package_count,
                    )
                    yield None, None, progress
                    continue

                elif line.startswith(self.UNPACKING):
                    arg = line[len(self.UNPACKING):].split()[0]

                    # Check that this line is about a package, not just
                    # some message that resembles an unpack line without
                    # being about a particular package.
                    if arg not in packages:
                        continue

                    desc = 'unpack'

                elif line.startswith(self.SETTING_UP):
                    arg = line[len(self.SETTING_UP):].split()[0]

                    # Check that this line is about a package, not just
                    # some message that resembles a configure line without
                    # being about a particular package.
                    if arg not in packages:
                        continue

                    desc = 'configure'

                else:
                    # We don't care about this line.  Maybe we should.
                    continue

                progress = self._calculate_progress(
                  fetch_count,
                  unpack_count,
                  configure_count,
                  package_count,
                )

                yield (desc, arg, progress)

                # The idea is that the status changes come when we *start*
                # an action but we should only increment counts after we've
                # *finished* the action.  Then we yield the progress change
                # when starting the *next* action.
                if desc == 'fetch':
                    fetch_count = fetch_count + 1
                elif desc == 'unpack':
                    unpack_count = unpack_count + 1
                elif desc == 'configure':
                    configure_count = configure_count + 1

            progress = self._calculate_progress(
              fetch_count,
              unpack_count,
              configure_count,
              package_count,
            )
            yield (None, None, progress)

    def iter_upgrade(self):
        for action, package, progress in self._follow_apt(
          'apt-get --quiet upgrade'
        ):
            yield action, package, progress

    def _apply_mark_auto(self, packages, mark_auto):
        # Note: Recent versions of apt-get mark packages as manually installed.
        # However, older versions do not (in particular, the version shipped
        # with Ubuntu 8.04 doesn't).  For this reason, we explicitly mark
        # packages as manually installed when mark_auto is False.

        s_packages = ' \\\n'.join(["'%s'" % p for p in packages])

        if mark_auto:
            with self.platform.open_shell() as sh:
                sh.execute('apt-mark markauto %s' % s_packages)
        else:
            with self.platform.open_shell() as sh:
                sh.execute('apt-mark unmarkauto %s' % s_packages)

    def iter_install(self, packages, mark_auto = False):
        if not packages:
            return

        for package in packages:
            if "'" in package:
                raise AssertionError(package)

        s_packages = ' \\\n'.join(["'%s'" % p for p in packages])

        for action, package, progress in self._follow_apt(
          'apt-get --quiet install %s' % s_packages,
        ):
            yield action, package, progress

        self._apply_mark_auto(packages, mark_auto)

    def iter_build_dep(self, packages, mark_auto = False):
        if not packages:
            return

        for package in packages:
            if "'" in package:
                raise AssertionError(package)

        s_packages = ' \\\n'.join(["'%s'" % p for p in packages])
        installed_packages = []

        for action, package, progress in self._follow_apt(
          'apt-get --quiet --only-source build-dep %s' % s_packages,
        ):
            # FIXME: This is brittle due to its tight coupling with the
            # _follow_apt implementation.
            if action == 'configure':
                installed_packages.append(package)
            yield action, package, progress

        if installed_packages:
            self._apply_mark_auto(installed_packages, mark_auto)

    def iter_dpkg_purge(self, packages):
        # FIXME: Clean up comments.
        # FIXME: Calculate progress.

        if not packages:
            return

        with self.platform.open_shell() as sh:
            try:
                for line in sh.follow_lines(
                  'dpkg --abort-after=1000000 --purge %s' % ' \\\n'.join([
                    "'%s'" % p for p in packages
                  ]),
                ):
                    if line.startswith(self.REMOVING):
                        desc = 'remove'
                        prefix = self.REMOVING
                        suffix = ' ...\n'

                    elif line.startswith(self.PURGING):
                        desc = 'purge'
                        prefix = self.PURGING
                        suffix = ' ...\n'

                    else:
                        # We don't think this line is meaningful
                        # for determining what the current minor
                        # step should be.
                        continue

                    start = len(prefix)
                    if suffix:
                        end = -len(suffix)
                    else:
                        end = None
                    arg = line[start:end]

                    # Make sure this line refers to an action taken on an
                    # actual package and is not some other notification.
                    if arg in packages:
                        yield desc, arg

            except CommandFailed, e:
                # Ignore exit status 1.  This is not a fatal error, it only
                # indicates that some requested action could not be completed.
                # We expect this to be the case, because we rely on dpkg
                # refusing to remove packages that would break dependencies.
                if e.status != 1:
                    raise

    def get_installed_packages(self):
        with self.platform.open_shell() as sh:
            status, output = sh.execute(
              'dpkg-query --show --showformat ?',
              r'${Package}\n',
            )
        return output.strip().split('\n')

    def get_binary_candidate_version(self, name):
        with self.platform.open_shell() as sh:
            try:
                status, output = sh.execute(
                  'apt-get --dry-run install ?',
                  name,
                )
            except CommandFailed:
                raise NoSuchPackageError()
        for line in output.strip().split('\n'):
            mo = self.REGEX_SELECTING.match(line)
            if mo:
                selected = mo.group('selected')
                raise VirtualPackageError(name, selected)
            mo = self.REGEX_INST.match(line)
            if mo:
                inst_name = mo.group('name')
                inst_version = mo.group('version')
                if inst_name == name:
                    return inst_version
        raise PackageAlreadyInstalledError()

    def get_binary_installed_version(self, name):
        with self.platform.open_shell() as sh:
            try:
                status, output = sh.execute(
                  'apt-cache show --no-all-versions ?',
                  name,
                )
            except CommandFailed:
                raise NoSuchPackageError()
        for line in output.strip().split('\n'):
            if line.startswith('Version: '):
                return line.split()[1]
        raise AssertionError(output)

    def get_binary_available_version(self, name):
        try:
            # May raise VirtualPackageError.
            return self.get_binary_candidate_version(name)
        except PackageAlreadyInstalledError:
            return self.get_binary_installed_version(name)

    def compare_versions(self, ver1, op, ver2):
        with self.platform.open_shell() as sh:
            try:
                sh.execute('dpkg --compare-versions ? ? ?', ver1, op, ver2)
            except CommandFailed:
                return False
        return True

    def rescan_local_packages(self):
        with self.platform.open_shell() as sh:
            sh.pushd(REPO_DIR)
            try:
                # FIXME: Should these be split out into separate functions?
                sh.execute('dpkg-scansources . >Sources')
                sh.execute('dpkg-scanpackages . >Packages')
            finally:
                sh.popd()

    def import_source_package(self, source_package):
        files = source_package.build_source(self.platform, REPO_DIR)
        return ImportedSourcePackage(self, source_package, files)


class ImportedSourcePackage(object):
    # Note: We keep a reference to the PackageSystem instance, but it
    # is not usable if the Platform is closed.  We rely on our caller
    # to ensure that the Platform is open when our methods are called.
    package_system = None
    source_package = None
    files = None

    _binaries = None
    _provides = None

    def __init__(self, package_system, source_package, files):
        super(ImportedSourcePackage, self).__init__()

        self.package_system = package_system
        self.source_package = source_package
        self.files = files

    def get_name(self):
        return self.source_package.name

    def get_version(self):
        return self.source_package.version

    def get_upstream_version(self):
        return self.source_package.get_upstream_version()

    def get_dsc_filename(self):
        for filename in self.files:
            if filename.endswith('.dsc'):
                return filename
        raise AssertionError('no dsc file?')

    def open_dsc_file(self):
        dsc_filename = self.get_dsc_filename()
        return self.package_system.platform.open_as_temp_file(dsc_filename)

    def get_binaries(self):
        with self.open_dsc_file() as f:
            tagfile = aptutil.parse_tag_file(f)
            while tagfile.Step() == 1:
                if tagfile.Section.has_key('Source'):
                    return [
                      s.strip()
                      for s in tagfile.Section['Binary'].split(',')
                    ]

    def get_binaries_and_provides(self):
        # These are cached to avoid repeated unpacking.
        if (self._binaries is None) or (self._provides is None):
            binaries, provides = self._get_binaries_and_provides()
            self._binaries = tuple(binaries)
            self._provides = tuple(provides)
        return self._binaries, self._provides

    def _get_binaries_and_provides(self):
        binaries = []
        provides = []

        unpack_dir = self.unpack()
        try:
            dirname = '%s-%s' % (
              self.get_name(),
              self.get_upstream_version(),
            )
            control_path = os.path.join(
              unpack_dir,
              dirname,
              'debian',
              'control',
            )
            f = self.package_system.platform.open_as_temp_file(control_path)
            with f:
                tagfile = aptutil.parse_tag_file(f)
                while tagfile.Step() == 1:
                    if not tagfile.Section.has_key('Package'):
                        continue
                    binaries.append(tagfile.Section['Package'].strip())
                    if tagfile.Section.has_key('Provides'):
                        provides.extend([
                          s.strip()
                          for s in tagfile.Section['Provides'].split(',')
                        ])
        finally:
            self.remove(unpack_dir)

        return binaries, provides

    def iter_dependencies(self):
        with self.open_dsc_file() as f:
            tagfile = aptutil.parse_tag_file(f)
            while tagfile.Step() == 1:
                if tagfile.Section.has_key('Source'):
                    for name in ('Build-Depends', 'Build-Depends-Indep'):
                        try:
                            raw_depends = tagfile.Section[name]
                        except KeyError:
                            pass
                        else:
                            for or_group in aptutil.parse_source_depends(raw_depends):
                                yield or_group
                    break

    def get_possible_dependencies(self):
        possible_deps = []
        for or_group in self.iter_dependencies():
            for pkgname, pkgver, pkgop in or_group:
                possible_deps.append(pkgname)
        return possible_deps

    ###

    def check_build_dependencies(self):
        with self.package_system.platform.open_shell() as sh:
            try:
                sh.execute(
                  'apt-get --dry-run --quiet --only-source build-dep ?',
                  self.get_name(),
                )
            except CommandFailed:
                return False
        return True

    def unpack(self):
        with self.package_system.platform.open_shell() as sh:
            status, output = sh.execute('mktemp -d')
            tmp_path = output.strip()
            sh.pushd(tmp_path)
            try:
                sh.execute(
                  'apt-get --quiet --only-source source ?',
                  self.get_name(),
                )

            finally:
                sh.popd()

        return tmp_path

    def build(self, unpack_dir):
        build_files = []

        dirname = '%s-%s' % (
          self.get_name(),
          self.get_upstream_version(),
        )
        build_dir = os.path.join(unpack_dir, dirname)

        with self.package_system.platform.open_shell() as sh:
            sh.pushd(build_dir)
            try:
                # Note: We use follow_lines so that output is logged
                # continously rather than buffered and logged after the command
                # terminates.
                for line in sh.follow_lines('dpkg-buildpackage -uc -us -sa'):
                    pass
            finally:
                sh.popd()

            changes_path = None
            for entry in sh.listdir(unpack_dir):
                if entry.endswith('.changes'):
                    changes_path = os.path.join(unpack_dir, entry)
                    break

            if changes_path is None:
                raise AssertionError()

            f = self.package_system.platform.open_as_temp_file(changes_path)
            with f:
                tagfile = aptutil.parse_tag_file(f)
                while tagfile.Step() == 1:
                    for line in tagfile.Section['Files'].split('\n'):
                        (
                          md5sum,
                          size,
                          section,
                          priority,
                          filename,
                        ) = line.split()

                        build_files.append(
                          os.path.join(unpack_dir, filename),
                        )

        return changes_path, build_files

    def remove(self, unpack_dir):
        with self.package_system.platform.open_shell() as sh:
            sh.execute('rm -fR ?', unpack_dir)


class RebarBuild(ComplexBuildStep):
    build_phase_classes = {}
    build_sequence = [
      'initialize',
      'base',
      'pre-build',
      'build',
      'post-build',
      'pre-install',
      'install',
      'post-install',
      'pre-remove',
      'remove',
      'post-remove',
      'pre-ship',
      'ship',
      'post-ship',
      'clean',
    ]

    @classmethod
    def register_build_phase(cls, build_phase_cls):
        cls.build_phase_classes[build_phase_cls.name] = build_phase_cls

    package_system_initialized = None

    def __init__(self, env, build_phase_names = None, *args, **kwargs):
        if build_phase_names is None:
            build_phase_names = self.build_sequence

        self.package_system_initialized = False

        subtasks = []

        for build_phase_name in build_phase_names:
            try:
                build_phase_cls = self.build_phase_classes[build_phase_name]
            except KeyError:
                raise NoSuchBuildPhase(build_phase_name)
            subtasks.append(build_phase_cls(env))

        super(RebarBuild, self).__init__(
          env,
          name = 'root',
          subtasks = subtasks,
          *args,
          **kwargs
        )

        # FIXME: It would be better to do this in Initialize, however, subtasks
        # have no mechanism to access parent task instances, so there is no way
        # to query the build for required packages from there.
        required_packages = self.get_required_packages()
        for subtask in self.subtasks:
            if isinstance(subtask, Initialize):
                subtask.set_required_packages(required_packages)
                break

    def run_subtask(self, subtask):
        if not self.package_system_initialized:
            self.init_package_system()

        return super(RebarBuild, self).run_subtask(subtask)

    def init_package_system(self):
        # Figure out what architecture the platform is.  If the architecture
        # was specified via the Rebar environment, this should match.
        # Otherwise, it should be the same architecture as the base platform.
        try:
            with self.virtualizer.open_platform() as platform:
                with platform.open_shell() as sh:
                    status, output = sh.execute('dpkg --print-architecture')
                    architecture = output.strip()
        except CommandFailed:
            self.output.print_debug(
              'warning: failed to detect platform architecture'
            )
            return

        # Note: May be None.
        env_arch = self.env['platform']['architecture']
        if env_arch and architecture != env_arch:
            raise AssertionError('%r != %r' % (architecture, env_arch))

        # We initialize APT now that we know what the system architecture is.
        # The APT architecture affects how dependencies are parsed and
        # evaluated (e.g. it impacts the return value of parse_depends,
        # parse_source_depends, etc.).  Note that this configuration is
        # process-wide, so running multiple builds with different architectures
        # in the same process is impossible.
        aptutil.initialize({'APT::Architecture': architecture})

        self.package_system_initialized = True


class UseBasePlatformMixin(object):
    use_base_platform = None

    def __init__(self, *args, **kwargs):
        self.use_base_platform = kwargs.pop('use_base_platform', False)
        super(UseBasePlatformMixin, self).__init__(*args, **kwargs)

    def open_platform(self):
        if self.use_base_platform:
            open_platform = self.virtualizer.open_base_platform
        else:
            open_platform = self.virtualizer.open_platform
        return open_platform()


class AptUpdateBuildStep(UseBasePlatformMixin, SimpleBuildStep):
    name = 'update package cache'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        # FIXME: Progress updates.
        with self.open_platform() as platform:
            package_system = PackageSystem(platform)
            package_system.update()


class AptUpgradeBuildStep(UseBasePlatformMixin, SimpleBuildStep):
    name = 'upgrade packages'

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        with self.open_platform() as platform:
            package_system = PackageSystem(platform)
            for action, package, progress in package_system.iter_upgrade():
                task_changes = []
                if action and package:
                    task_changes.append(self.change(
                      'status',
                      '%s %s' % (action, package),
                    ))
                if (progress is not None) and (progress != self.progress):
                    task_changes.append(self.change('progress', progress))
                yield task_changes


class AptInstallBuildStep(UseBasePlatformMixin, SimpleBuildStep):
    name = 'apt-get install'
    packages = None
    mark_auto = None

    def __init__(self, env, packages = None, *args, **kwargs):
        if packages is None:
            packages = []
        self.packages = packages
        super(AptInstallBuildStep, self).__init__(env, *args, **kwargs)

    def get_required_packages(self):
        packages = super(AptInstallBuildStep, self).get_required_packages()
        # Needed for copying files to and from the virtualizer.
        packages.add('netcat-openbsd')
        # Needed for apt-mark.
        packages.add('python-apt')
        return packages

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        if self.mark_auto is None:
            raise AssertionError('subclass must define attribute mark_auto')

        with self.open_platform() as platform:
            package_system = PackageSystem(platform)

            for action, package, progress in package_system.iter_install(
              self.packages,
              mark_auto = self.mark_auto,
            ):
                task_changes = []
                if action and package:
                    task_changes.append(self.change(
                      'status',
                      '%s %s' % (action, package),
                    ))
                if (progress is not None) and (progress != self.progress):
                    task_changes.append(self.change('progress', progress))
                yield task_changes


class AptCleanBuildStep(UseBasePlatformMixin, SimpleBuildStep):
    name = 'clean package cache'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        with self.open_platform() as platform:
            package_system = PackageSystem(platform)
            package_system.clean()


class DpkgRemoveBuildStep(SimpleBuildStep):
    name = 'dpkg remove'
    packages = None

    def __init__(self, env, packages = None, *args, **kwargs):
        if packages is None:
            packages = []
        self.packages = packages
        super(DpkgRemoveBuildStep, self).__init__(env, *args, **kwargs)

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        with self.virtualizer.open_platform() as platform:
            package_system = PackageSystem(platform)

            for action, package in package_system.iter_dpkg_purge(self.packages):
                task_changes = []
                if action and package:
                    task_changes.append(self.change(
                      'status',
                      '%s %s' % (action, package),
                    ))
                yield task_changes


###


class VirtualizerInstall(AptInstallBuildStep):
    name = 'install packages'
    mark_auto = False

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('use_base_platform', True)
        super(VirtualizerInstall, self).__init__(*args, **kwargs)


class PrepareVirtualizer(ComplexBuildStep):
    name = 'prepare virtualizer'

    required_packages = None

    def __init__(self, env, *args, **kwargs):
        self.required_packages = []

        super(PrepareVirtualizer, self).__init__(
          env,
          subtasks = [
            AptUpdateBuildStep(
              env,
              progress_weight = 1.0,
              use_base_platform = True,
            ),
            AptUpgradeBuildStep(
              env,
              progress_weight = 3.0,
              use_base_platform = True,
            ),
            VirtualizerInstall(
              env,
              packages = self.required_packages,
              progress_weight = 25.0,
            ),
          ],
          *args,
          **kwargs
        )

    def set_required_packages(self, required_packages):
        self.required_packages[:] = required_packages


class Initialize(ComplexBuildStep):
    name = 'initialize'

    def __init__(self, env, *args, **kwargs):
        super(Initialize, self).__init__(
          env,
          subtasks = [
            PrepareVirtualizer(env),
          ],
          *args,
          **kwargs
        )

    def set_required_packages(self, required_packages):
        # Pass required_packages on to PrepareVirtualizer.
        for subtask in self.subtasks:
            if isinstance(subtask, PrepareVirtualizer):
                subtask.set_required_packages(required_packages)
                break

RebarBuild.register_build_phase(Initialize)


class Debootstrap(SimpleBuildStep):
    name = 'bootstrap base system'

    def get_required_packages(self):
        packages = super(Debootstrap, self).get_required_packages()
        packages.add('debootstrap')
        return packages

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        bootstrap_opts = self.env['bootstrap']
        if bootstrap_opts:
            mirror = bootstrap_opts['mirror']
            suite = bootstrap_opts['suite']
            script = bootstrap_opts['script']
            include = bootstrap_opts['include']
            exclude = bootstrap_opts['exclude']
        else:
            # No debootstrap options were specified.  Pull what we need from
            # the first repository.  Note that we ignore repository
            # architectures here.
            repository = self.env['repositories'][0]
            mirror = repository['mirror']
            suite = repository['suite']
            components = repository['components']
            architectures = repository['architectures']
            trusted = repository['trusted']
            script = None
            include = None
            exclude = None

        with self.virtualizer.open_base_platform() as platform:
            with platform.open_shell() as sh:
                debootstrapper = Debootstrapper(
                  sh,
                  target = '/platform',
                  mirror = mirror,
                  suite = suite,
                  script = script,
                  include = include,
                  exclude = exclude,
                  architecture = self.env['platform']['architecture'],
                )
                for action, package, progress in debootstrapper.iter_run():
                    task_changes = []
                    if (action is not None) and (package is not None):
                        task_changes.append(
                          self.change('status', '%s %s' % (action, package))
                        )
                    if (progress is not None) and (progress != self.progress):
                        task_changes.append(self.change('progress', progress))
                    yield task_changes


class RewriteHostname(SimpleBuildStep):
    name = 'write /etc/hostname'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        hostname = self.env['platform']['hostname']
        if hostname is not None:
            with self.virtualizer.open_platform() as platform:
                with platform.open_shell() as sh:
                    sh.execute('echo ? >/etc/hostname', hostname)


class RewriteSourcesList(SimpleBuildStep):
    name = 'write /etc/apt/sources.list'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        lines = []
        for repository in self.env['repositories']:
            parts = ['deb']
            opts = []
            if repository['architectures']:
                opts.append('arch=%s' % ','.join(repository['architectures']))
            if repository['trusted']:
                opts.append('trusted=yes')
            if opts:
                parts.append('[%s]' % ' '.join(opts))
            parts.extend([repository['mirror'], repository['suite']])
            parts.extend(repository['components'])
            lines.append(' '.join(parts))
        with self.virtualizer.open_platform() as platform:
            with platform.open_shell() as sh:
                sh.execute('echo ? >/etc/apt/sources.list', '\n'.join(lines))


class RewriteKernelImgConf(SimpleBuildStep):
    name = 'write /etc/kernel-img.conf'
    kernel_img_conf_content = (
      'do_symlinks = yes\n'
      'relative_links = yes\n'
      'do_bootloader = no\n'
      'do_bootfloppy = no\n'
      'do_initrd = yes\n'
      'link_in_boot = no\n'
    )

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        with self.virtualizer.open_platform() as platform:
            with platform.open_shell() as sh:
                sh.execute(
                  'echo -n ? >/etc/kernel-img.conf',
                  self.kernel_img_conf_content,
                )


class InstallBaseTools(AptInstallBuildStep):
    name = 'install base tools'
    mark_auto = True

    def __init__(self, env, *args, **kwargs):
        # apt-mark needs python-apt.  apt-mark is provided by the apt package
        # but that package does not depend on python-apt.  It is clear that we
        # would not want apt to depend on python-apt, but I think it is a bug
        # that apt installs executables that cannot be used.  apt-mark should
        # either be rewritten in C or it should be installed as a separate
        # package.  Since we need it (we don't currently have any other way to
        # mark a package as automatically installed), we work around the issue
        # by installing python-apt early in the build process.  Note that it is
        # marked as automatically installed.
        super(InstallBaseTools, self).__init__(
          env,
          ['python-apt'],
          *args,
          **kwargs
        )


class Base(ComplexBuildStep):
    name = 'base'
    progress_weight = 10.0

    def __init__(self, env, *args, **kwargs):
        super(Base, self).__init__(
          env,
          subtasks = [
            Debootstrap(env, progress_weight = 95.0),
            RewriteHostname(env, progress_weight = 1.0),
            RewriteSourcesList(env, progress_weight = 1.0),
            RewriteKernelImgConf(env, progress_weight = 1.0),
            AptUpdateBuildStep(env, progress_weight = 1.0),
            AptUpgradeBuildStep(env, progress_weight = 3.0),
            InstallBaseTools(env, progress_weight = 3.0),
          ],
          *args,
          **kwargs
        )

RebarBuild.register_build_phase(Base)


class InstallBuildTools(AptInstallBuildStep):
    name = 'install build tools'
    mark_auto = True

    def __init__(self, env, *args, **kwargs):
        super(InstallBuildTools, self).__init__(
          env,
          ['dpkg-dev', 'build-essential'],
          *args,
          **kwargs
        )

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        # FIXME: Why do we need to check binary_dirs here?
        if self.env['binary_dirs'] or self.env['source_dirs']:
            return super(InstallBuildTools, self).do_run()


class InstallExtraPackages(AptInstallBuildStep):
    name = 'install extra packages'
    mark_auto = True

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        if self.env['install_packages']:
            self.packages.extend(self.env['install_packages'])
            return super(InstallExtraPackages, self).do_run()


class InitializeLocalAptRepository(SimpleBuildStep):
    name = 'initialize custom package repository'

    # FIXME: We should also fail on OSError.
    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        binary_dirs = self.env['binary_dirs']
        source_dirs = self.env['source_dirs']

        if binary_dirs or source_dirs:
            with self.virtualizer.open_platform() as platform:
                with platform.open_shell() as sh:
                    # Make sure REPO_DIR exists.
                    sh.execute('mkdir -p ?', REPO_DIR)

                    for dirname in binary_dirs:
                        for entry in os.listdir(dirname):
                            if entry.endswith('.deb'):
                                src = os.path.join(dirname, entry)

                                # FIXME: We should probably support multi-arch
                                # here e.g. i386 packages on amd64.
                                sys_arch = aptutil.config_get('APT::Architecture')
                                pkg_arch = self._get_package_architecture(src)
                                if pkg_arch not in (sys_arch, 'all'):
                                    self.output.print_important(
                                      'ignoring %s package %s',
                                      pkg_arch,
                                      entry,
                                    )
                                    continue

                                dst = os.path.join(REPO_DIR, entry)
                                platform.copy_to(src, dst)

                    try:
                        sh.execute('ls /etc/apt/source.list.d/rebar.list')
                    except CommandFailed:
                        sh.execute(
                          'echo ? >> /etc/apt/sources.list.d/rebar.list',
                          ((
                            'deb [trusted=yes] file://%s ./\n'
                            'deb-src [trusted=yes] file://%s ./'
                          ) % (REPO_DIR, REPO_DIR)),
                        )

                    package_system = PackageSystem(platform)
                    package_system.rescan_local_packages()

    def _get_package_architecture(self, path):
        status, output = self.user_shell.execute(
          "dpkg-deb --show --showformat '${Architecture}' ?",
          path,
        )
        return output.strip()


class ImportSourcePackages(SimpleBuildStep):
    name = 'import source packages'

    packages = None

    def __init__(self, env, packages, *args, **kwargs):
        self.packages = packages
        super(ImportSourcePackages, self).__init__(env, *args, **kwargs)

    @fail_on(InvalidSourcePackageError)
    def do_run(self):
        # FIXME: Progress.

        source_dirs = self.env['source_dirs']
        if source_dirs:
            with self.virtualizer.open_platform() as platform:
                package_system = PackageSystem(platform)

                for source_dir in source_dirs:
                    for entry in os.listdir(source_dir):
                        path = os.path.join(source_dir, entry)

                        if (
                          os.path.isdir(path)
                          and os.path.isdir(os.path.join(path, 'debian'))
                        ):
                            source_package = UnpackedSourcePackage(path)

                        elif os.path.isfile(path) and path.endswith('.dsc'):
                            source_package = DscSourcePackage(path)

                        else:
                            continue

                        yield [
                          self.change(
                            'status',
                            ('import %s' % source_package.name),
                          ),
                        ]
                        imported_source_package = \
                          package_system.import_source_package(source_package)
                        self.packages.append(imported_source_package)

                package_system.rescan_local_packages()


class SatisfyBuildDependencies(SimpleBuildStep):
    name = 'satisfy build dependencies'

    package = None

    def __init__(self, env, package, *args, **kwargs):
        self.package = package
        super(SatisfyBuildDependencies, self).__init__(env, *args, **kwargs)

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        with self.virtualizer.open_platform() as platform:
            package_system = PackageSystem(platform)

            for action, package, progress in package_system.iter_build_dep(
              [self.package.get_name()],
              mark_auto = True,
            ):
                task_changes = []
                if action and package:
                    task_changes.append(self.change(
                      'status',
                      '%s %s' % (action, package),
                    ))
                if (progress is not None) and (progress != self.progress):
                    task_changes.append(self.change('progress', progress))
                yield task_changes


class _BuildPackage(SimpleBuildStep):
    name = 'build package'

    package = None
    finalize = None

    def __init__(self, env, package, finalize, *args, **kwargs):
        self.package = package
        self.finalize = finalize

        super(_BuildPackage, self).__init__(env, *args, **kwargs)

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        # FIXME: Check if the package to be built is newer than what's already
        # available.  Fail if not (there's no point in building a package that
        # will never be installed).  Or is there?  Someone might request a
        # specific version of a package (even though we don't support that
        # yet).

        # Note: We *must* open the platform before calling methods on the
        # ImportedSourcePackage instance.  The ImportedSourcePackage has a
        # reference to the platform already (via PackageSystem) and assumes
        # that it is open and ready for use.

        with self.virtualizer.open_platform() as platform:
            yield [self.change('status', 'unpack')]
            unpack_dir = self.package.unpack()

            try:
                yield [self.change('status', 'build')]
                changes_file, build_files = self.package.build(unpack_dir)

                export_dirs = self.env['export_dirs']
                if export_dirs:
                    yield [self.change('status', 'export')]
                    # Export the .changes file and any build files (source
                    # and binary packages).
                    export_files = [changes_file]
                    export_files.extend(build_files)
                    for export_dir in export_dirs:
                        self.user_shell.execute('mkdir -p ?', export_dir)
                        for filename in export_files:
                            base = os.path.basename(filename)
                            platform.copy_from(
                              filename,
                              os.path.join(export_dir, base),
                            )

                if callable(self.finalize):
                    self.finalize(self.package, changes_file, build_files)

            finally:
                self.package.remove(unpack_dir)


class BuildPackage(ComplexBuildStep):
    package = None

    def __init__(
      self,
      env,
      package,
      finalize,
      *args,
      **kwargs
    ):
        self.package = package

        super(BuildPackage, self).__init__(
          env,
          name = 'build %s' % package.get_name(),
          subtasks = [
            AptUpdateBuildStep(env, progress_weight = 1.0),
            SatisfyBuildDependencies(env, package, progress_weight = 5.0),
            _BuildPackage(
              env,
              package,
              finalize,
              progress_weight = 20.0,
            ),
          ],
          *args,
          **kwargs
        )

    def check(self, platform):
        return self.package.check_build_dependencies()


class _ImportBinaries(SimpleBuildStep):
    name = 'import'

    parent = None

    def __init__(
      self,
      env,
      parent,
      *args,
      **kwargs
    ):
        self.parent = parent

        super(_ImportBinaries, self).__init__(env, *args, **kwargs)

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        # Note: We *must* open the platform before calling methods on the
        # ImportedSourcePackage instance.  The ImportedSourcePackage has a
        # reference to the platform already (via PackageSystem) and assumes
        # that it is open and ready for use.

        changes_file = self.parent.changes_file
        build_files = self.parent.build_files
        stash_dir = self.parent.stash_dir

        with self.virtualizer.open_platform() as platform:
            with platform.open_shell() as sh:
                for filename in build_files:
                    dst = os.path.join(REPO_DIR, os.path.basename(filename))
                    sh.execute('cp --force --link ? ?', filename, dst)

                package_system = PackageSystem(platform)
                package_system.rescan_local_packages()

                # Build files are no longer needed, so we can clean up the
                # stash directory.
                if stash_dir is not None:
                    sh.execute('rm -fR ?', stash_dir)


class ImportBinaries(ComplexBuildStep):
    package = None
    changes_file = None
    build_files = None

    stash_dir = None

    def __init__(
      self,
      env,
      package,
      changes_file,
      build_files,
      *args,
      **kwargs
    ):
        self.package = package
        self.changes_file = changes_file
        self.build_files = build_files

        super(ImportBinaries, self).__init__(
          env,
          name = 'import %s binaries' % package.get_name(),
          subtasks = [
            _ImportBinaries(env, self, progress_weight = 1.0),
            AptUpdateBuildStep(env, progress_weight = 1.0),
          ],
          *args,
          **kwargs
        )

    def stash_files(self, platform):
        with platform.open_shell() as sh:
            status, output = sh.execute('mktemp -d')
            self.stash_dir = output.strip()
            files = [self.changes_file]
            files.extend(self.build_files)
            stashed_files = []
            for src in files:
                dst = os.path.join(self.stash_dir, os.path.basename(src))
                sh.execute('cp --link --force ? ?', src, dst)
                stashed_files.append(dst)
            self.changes_file = stashed_files[0]
            self.build_files = stashed_files[1:]

    def check(self, platform):
        package_system = PackageSystem(platform)

        source_version = self.package.get_version()
        source_binaries, source_provides = \
          self.package.get_binaries_and_provides()

        with platform.open_shell() as sh:
            for filename in self.build_files:
                if filename.endswith('.deb'):
                    status, output = sh.execute(
                      'dpkg-deb --field ? Depends',
                      filename,
                    )
                    raw_depends = output.strip()
                    for or_group in aptutil.parse_depends(raw_depends):
                        found = False
                        for depname, depver, depop in or_group:
                            # FIXME: What if we get a mismatch between the
                            # source package and what's in the repo?
                            found_virtual = False
                            available_versions = []

                            if depname in source_binaries:
                                available_versions.append(source_version)
                            elif depname in source_provides:
                                found_virtual = True
                                available_versions.append(source_version)

                            try:
                                repo_name = depname
                                while True:
                                    try:
                                        repo_version = \
                                          package_system.get_binary_available_version(
                                            repo_name,
                                          )
                                    except VirtualPackageError, e:
                                        found_virtual = True
                                        repo_name = e.selected
                                    else:
                                        break
                            except NoSuchPackageError:
                                pass
                            else:
                                available_versions.append(repo_version)

                            if found_virtual and depver:
                                raise AssertionError((
                                  'source package %s depends on '
                                  'virtual package %s '
                                  'but specifies a dependency version'
                                ) % (
                                  self.package.get_name(),
                                  depname,
                                ))

                            if not available_versions:
                                # Package referred to by depname is not
                                # available.
                                continue

                            if depver:
                                for available_version in available_versions:
                                    if package_system.compare_versions(
                                      available_version,
                                      depop,
                                      depver,
                                    ):
                                        found = True
                                        break
                            else:
                                found = True

                            if found:
                                # No need to check other or-group members.
                                break

                        if not found:
                            return False

        return True


class BuildPackages(ComplexBuildStep):
    name = 'build packages'

    packages = None

    def __init__(self, env, packages, *args, **kwargs):
        super(BuildPackages, self).__init__(env, *args, **kwargs)
        self.packages = packages

    def do_run(self):
        # Open the platform once and keep it open rather than opening and
        # closing each time we need to access ImportedSourcePackage methods.
        with self.virtualizer.open_platform() as platform:
            # First optimize the build order.  We have to do this here rather
            # than in our initializer because we need access to the platform to
            # sort packages.
            for package in self.sort_packages(self.packages):
                self.schedule_build(package)

            # After each subtask is completed, we search for the next subtask
            # that is ready to be run (i.e. its "check" method returns True).
            # We have to do this because we cannot know the correct build order
            # a priori because:
            #
            # * We do not know the full set of recursive build dependencies for
            #   a given source package until after we've built binaries for its
            #   immediate build dependencies.
            # * We do not know if a source package's binaries are installable
            #   until we've actually built them.  We don't want to import
            #   binaries into the local APT repository until we know they are
            #   installable or we will not be able to fulfill build
            #   dependencies for other source packages.

            for task_changes in super(BuildPackages, self).do_run():
                for task_change in task_changes:
                    if (
                      task_change.parameter == 'running_subtask'
                      and task_change.task is self
                      and task_change.value is None
                    ):
                        prev_subtask = task_change.oldvalue
                        if prev_subtask is None:
                            next_index = 0
                        else:
                            next_index = self.subtasks.index(prev_subtask) + 1
                        try:
                            next_subtask = self.subtasks[next_index]
                        except IndexError:
                            pass
                        else:
                            if (prev_subtask is None) or (not prev_subtask.failed):
                                found = False
                                for subtask in self.subtasks[next_index:]:
                                    if subtask.check(platform):
                                        found = True
                                        break
                                if found and (subtask != next_subtask):
                                    self.remove_subtask(subtask)
                                    self.add_subtask(subtask, next_index)

                yield task_changes

    def schedule_build(self, package):
        self.add_subtask(BuildPackage(
          self.env,
          package,
          self.finalize_build,
        ))

    def finalize_build(self, package, changes_file, build_files):
        subtask = ImportBinaries(
          self.env,
          package,
          changes_file,
          build_files,
        )
        with self.virtualizer.open_platform() as platform:
            subtask.stash_files(platform)
        if self.running_subtask is not None:
            self.add_subtask_after(subtask, self.running_subtask)
        else:
            self.add_subtask(subtask)

    def sort_packages(self, packages):
        # This is a naive approach to sorting source packages so that
        # dependencies are built before the packages that build-depend on them.
        # It considers immediate build dependencies only (i.e. it does not
        # examine recursive dependencies.  This is good enough in most cases to
        # speed up the overall build process and reduce the need for rebuilds.
        # However, rebuilds may be required in some cases (e.g. if a binary
        # uses symbols from a library that it does not have an explicit build
        # dependency for).

        # Open the platform so we can use ImportedSourcePackage methods.
        with self.virtualizer.open_platform():
            sorted_packages = []
            for package in packages:
                index = 0
                for sorted_package in sorted_packages:
                    binaries, provides = package.get_binaries_and_provides()
                    names = list(binaries)
                    names.extend(provides)
                    dependencies = sorted_package.get_possible_dependencies()

                    found = False
                    for name in names:
                        if name in dependencies:
                            found = True
                            break
                    if found:
                        break

                    index = index + 1

                sorted_packages.insert(index, package)

        return sorted_packages


class Build(ComplexBuildStep):
    name = 'build'
    progress_weight = 10.0

    def __init__(self, env, *args, **kwargs):
        # Note: This list is modified by ImportSourcePackages.
        packages = []

        # FIXME: Should we only update the APT cache when we know we've changed
        # something that requires a cache rebuild, or is it okay to do so when
        # we know we need it updated?

        super(Build, self).__init__(
          env,
          subtasks = [
            AptUpdateBuildStep(env, progress_weight = 1.0),

            InstallBuildTools(env, progress_weight = 5.0),
            AptUpdateBuildStep(env, progress_weight = 1.0),

            InstallExtraPackages(env, progress_weight = 5.0),

            InitializeLocalAptRepository(env, progress_weight = 1.0),
            AptUpdateBuildStep(env, progress_weight = 1.0),

            ImportSourcePackages(env, packages, progress_weight = 5.0),
            AptUpdateBuildStep(env, progress_weight = 1.0),

            BuildPackages(env, packages, progress_weight = 50.0),
          ],
          *args,
          **kwargs
        )

RebarBuild.register_build_phase(Build)


class _Install(AptInstallBuildStep):
    name = 'install required packages'
    mark_auto = False

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        self.packages.extend(self.env['required_packages'])
        return super(_Install, self).do_run()


class Install(ComplexBuildStep):
    name = 'install'
    progress_weight = 5.0

    def __init__(self, env):
        super(Install, self).__init__(
          env,
          subtasks = [
            AptUpdateBuildStep(env, progress_weight = 1.0),
            AptUpgradeBuildStep(env, progress_weight = 3.0),
            _Install(env, progress_weight = 25.0),
          ],
        )

RebarBuild.register_build_phase(Install)


class RemoveLocalAptRepository(SimpleBuildStep):
    name = 'remove custom package repository'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        if self.env['binary_dirs'] or self.env['source_dirs']:
            # FIXME: Separate "prune" step?
            with self.virtualizer.open_platform() as platform:
                with platform.open_shell() as sh:
                    sh.execute('rm -fR ?', REPO_DIR)
                    sh.execute('rm -f /etc/apt/sources.list.d/rebar.list')


class RemovePackages(DpkgRemoveBuildStep):
    name = 'remove packages'

    @fail_on(CommandFailed, CommandSignalled)
    def do_run(self):
        required_packages = self.env['required_packages']

        with self.virtualizer.open_platform() as platform:
            package_system = PackageSystem(platform)
            installed_packages = package_system.get_installed_packages()

        remove_packages = [
          p for p in installed_packages
          if p not in required_packages
        ]

        self.packages.extend(remove_packages)

        for task_changes in super(RemovePackages, self).do_run():
            yield task_changes


class Remove(ComplexBuildStep):
    name = 'remove'
    progress_weight = 5.0

    def __init__(self, env, *args, **kwargs):
        super(Remove, self).__init__(
          env,
          subtasks = [
            # Clean out APT package cache *before* removing packages.  We may
            # be removing APT.
            AptCleanBuildStep(env),
            RemoveLocalAptRepository(env, progress_weight = 1.0),
            RemovePackages(env, progress_weight = 25.0),
          ],
          *args,
          **kwargs
        )

RebarBuild.register_build_phase(Remove)


class Ship(ComplexBuildStep):
    name = 'ship'
    progress_weight = 10.0

    def __init__(self, *args, **kwargs):
        super(Ship, self).__init__(*args, **kwargs)
        for target_directive in self.env.get('targets', []):
            self.add_subtask(target_directive.get_build_step(self.env))

RebarBuild.register_build_phase(Ship)


class HookBuildStep(ComplexBuildStep):
    action_group_name = None

    def __init__(self, *args, **kwargs):
        super(HookBuildStep, self).__init__(*args, **kwargs)
        for action_directive in self.env[self.action_group_name]:
            self.add_subtask(action_directive.get_build_step(self.env))


class PreBuild(HookBuildStep):
    name = 'pre-build'
    action_group_name = 'pre_build_actions'

RebarBuild.register_build_phase(PreBuild)


class PostBuild(HookBuildStep):
    name = 'post-build'
    action_group_name = 'post_build_actions'

RebarBuild.register_build_phase(PostBuild)


class PreInstall(HookBuildStep):
    name = 'pre-install'
    action_group_name = 'pre_install_actions'

RebarBuild.register_build_phase(PreInstall)


class PostInstall(HookBuildStep):
    name = 'post-install'
    action_group_name = 'post_install_actions'

RebarBuild.register_build_phase(PostInstall)


class PreRemove(HookBuildStep):
    name = 'pre-remove'
    action_group_name = 'pre_remove_actions'

RebarBuild.register_build_phase(PreRemove)


class PostRemove(HookBuildStep):
    name = 'post-remove'
    action_group_name = 'post_remove_actions'

RebarBuild.register_build_phase(PostRemove)


class PreShip(HookBuildStep):
    name = 'pre-ship'
    action_group_name = 'pre_ship_actions'

RebarBuild.register_build_phase(PreShip)


class PostShip(HookBuildStep):
    name = 'post-ship'
    action_group_name = 'post_ship_actions'

RebarBuild.register_build_phase(PostShip)


class Clean(SimpleBuildStep):
    name = 'clean'

    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        with self.virtualizer.open_base_platform() as platform:
            with platform.open_shell() as sh:
                # These are safety checks; they should never fail.
                for path in (
                  '/platform/proc',
                  '/platform/dev/pts',
                  '/platform/dev/shm',
                ):
                    try:
                        if sh.listdir(path):
                            raise AssertionError('%s is mounted' % path)
                    except CommandFailed:
                        pass

                sh.execute('rm --force --recursive --one-file-system /platform')
                sh.execute('rm --force --recursive --one-file-system /tmp/*')

RebarBuild.register_build_phase(Clean)


class BaseLogin(SimpleBuildStep):
    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        with self.open_platform() as platform:
            with platform.open_shell() as sh:
                sh.interact(handle_sigwinch = True)

    def open_platform(self):
        raise NotImplementedError()


class Login(BaseLogin):
    name = 'login'

    def open_platform(self):
        return self.virtualizer.open_platform()

RebarBuild.register_build_phase(Login)


class VirtLogin(BaseLogin):
    name = 'virt-login'

    def open_platform(self):
        return self.virtualizer.open_base_platform()

RebarBuild.register_build_phase(VirtLogin)
