# Copyright (c) 2007-2013 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, string


class Debootstrapper(object):
    RETRIEVING = 'I: Retrieving '
    VALIDATING = 'I: Validating '
    EXTRACTING = 'I: Extracting '
    UNPACKING = 'I: Unpacking '
    CONFIGURING = 'I: Configuring '

    INSTALLING_CORE_PACKAGES = 'I: Installing core packages'

    DEBOOTSTRAP_DATA_DIR = '/usr/share/debootstrap'

    sh = None
    suite = None
    target = None
    mirror = None
    script = None
    include = None
    exclude = None
    architecture = None

    version = None

    def __init__(
      self,
      sh,
      suite,
      target,
      mirror = None,
      script = None,
      include = None,
      exclude = None,
      architecture = None,
    ):
        if (script is not None) and (mirror is None):
            raise AssertionError(
              'if script is specified, mirror must be specified'
            )

        super(Debootstrapper, self).__init__()

        self.sh = sh
        self.suite = suite
        self.target = target
        self.mirror = mirror
        self.script = script
        self.include = include
        self.exclude = exclude
        self.architecture = architecture

        self.version = self._get_version()

    def _get_version(self):
        status, output = self.sh.execute('debootstrap --version')
        # 'debootstrap 1.2.3ubuntu4'
        version_string = output.strip()
        # '1.2.3ubuntu4'
        version_string = version_string.split()[1]
        # ['1', '2', '3ubuntu4']
        parts = version_string.split('.', 2)

        # ['1', '2', '3', 'ubuntu4']
        last_part = parts.pop()
        found = False
        for index, ch in enumerate(last_part):
            if ch not in string.digits:
                found = True
                break
        if found:
            parts.extend([last_part[:index], last_part[index:]])
        parts.append(last_part)

        # [1, 2, 3, 'ubuntu4']
        for index in range(min(len(parts), 3)):
            try:
                parts[index] = int(parts[index], 10)
            except ValueError:
                pass

        # (1, 2, 3, 'ubuntu4')
        return tuple(parts)

    def _calculate_progress(
      self,
      fetch_count,
      extract_count,
      unpack_count,
      configure_count,
      required_package_count,
      package_count,
    ):
        if package_count > 0:
            # Only required packages have an extract step.
            total_steps = (3 * package_count) + required_package_count
            current_steps = (
              fetch_count + extract_count + unpack_count + configure_count
            )
            progress = 100.0 * (float(current_steps)/float(total_steps))
        else:
            progress = None
        return progress

    def iter_run(self):
        placeholders = []
        args = []

        if self.version >= (1, 0, 30):
            # FIXME: We should provide some way for the user to specify whether
            # or not the release signature should be checked.
            args.append('--no-check-gpg')
            placeholders.append('?')

        if self.architecture is not None:
            placeholders.extend(['?', '?'])
            args.extend(['--arch', self.architecture])

        if self.include:
            placeholders.extend(['?', '?'])
            args.extend(['--include', ','.join(self.include)])

        if self.exclude:
            placeholders.extend(['?', '?'])
            args.extend(['--exclude', ','.join(self.exclude)])

        if self.script is not None:
            if self.mirror is None:
                raise AssertionError()

            script = self.script
            if not script.startswith('/'):
                script = os.path.join(
                  self.DEBOOTSTRAP_DATA_DIR,
                  'scripts',
                  script,
                )

            placeholders.extend(['?', '?', '?', '?'])
            args.extend([self.suite, self.target, self.mirror, script])

        else:
            placeholders.extend(['?', '?'])
            args.extend([self.suite, self.target])

            if self.mirror is not None:
                placeholders.append('?')
                args.append(self.mirror)

        fetch_count = 0
        extract_count = 0
        unpack_count = 0
        configure_count = 0

        status, output = self.sh.execute(
          (
            'debootstrap --print-debs %s 2>/dev/null'
            % ' '.join(placeholders)
          ),
          *args
        )
        output = output.strip()
        packages = output.split()
        package_count = len(packages)

        # Only required packages have an extract step.  The problem is that we
        # don't know which packages are required, so we don't know how many
        # extract steps to expect.
        #
        # To deal with this, we start by assuming that all the packages are
        # required, and then correct required_package_count once we've seen the
        # "Installing core packages" line, at which point all required packages
        # have been extracted.
        required_package_count = package_count

        # Some packages are configured more than once.  We keep track of which
        # have been configured so we don't count them twice.
        configured = set([])

        for line in self.sh.follow_lines(
          'debootstrap %s' % ' '.join(placeholders),
          *args
        ):
            if line.startswith(self.INSTALLING_CORE_PACKAGES):
                # Core packages have been extracted, extract_count indicates
                # the total number of required packages.
                required_package_count = extract_count

            elif line.startswith(self.RETRIEVING):
                prefix = self.RETRIEVING
                suffix = '\n'
                action = 'fetch'

            elif line.startswith(self.VALIDATING):
                # Note: We consider validation part of the fetch step.
                # As such, we ignore validation lines.

                #prefix = self.VALIDATING
                #suffix = '\n'
                #action = 'validate'

                continue

            elif line.startswith(self.EXTRACTING):
                prefix = self.EXTRACTING
                suffix = '...\n'
                action = 'extract'

            elif line.startswith(self.UNPACKING):
                # Note: This matches "Unpacking required packages...".  We
                # filter this below by making sure that package is a package
                # name.

                prefix = self.UNPACKING
                suffix = '...\n'
                action = 'unpack'

            elif line.startswith(self.CONFIGURING):
                prefix = self.CONFIGURING
                suffix = '...\n'
                action = 'configure'

            else:
                # We don't think this line is meaningful for
                # determining what the current minor step should be.
                continue

            start = len(prefix)
            if suffix:
                end = -len(suffix)
            else:
                end = None
            package = line[start:end]

            # Check that this line is about a package, not just some
            # message that resembles an extract/unpack/configure line
            # without being about a particular package.
            if package not in packages:
                continue

            progress = self._calculate_progress(
              fetch_count,
              extract_count,
              unpack_count,
              configure_count,
              required_package_count,
              package_count,
            )

            yield action, package, progress

            if action == 'fetch':
                fetch_count = fetch_count + 1
            elif action == 'extract':
                extract_count = extract_count + 1
            elif action == 'unpack':
                unpack_count = unpack_count + 1
            elif action == 'configure':
                # Only increment configure_count if the package was not
                # already configured.
                if package not in configured:
                    configure_count = configure_count + 1
                    configured.add(package)

        progress = self._calculate_progress(
          fetch_count,
          extract_count,
          unpack_count,
          configure_count,
          required_package_count,
          package_count,
        )
        yield None, None, progress
