# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement 

import os, inspect, re

from fdisk.units import UnitType
from fdisk.mbr import TYPES

from rebar.actions import (
  ExecAction,
  ShellAction,
  CopyToAction,
  CopyFromAction,
  PatchAction,
)
from rebar.targets import (
  PlatformImageTarget,
  DirImageTarget,
  DiskImagingImageTarget,
  DiskImagingDirTarget,
  NetworkImagingDirTarget,
  ActionSequenceTarget,
)
from rebar.exceptions import (
  RebarSyntaxError,
  MissingRequiredArgumentError,
)
from rebar.environment import (
  RebarEnvironment,

  PLATFORM,
  VIRTUALIZER,

  DISK,
  GZIMG,

  GRUB,
  SYSLINUX,

  MBR,
  VBR,

  EXT2,
  VFAT,

  GZIP,
)
from rebar.util import generator


class DirectiveMeta(type):
    def __init__(cls, name, bases, dct):
        super(DirectiveMeta, cls).__init__(name, bases, dct)
        cls.init_argspec = inspect.getargspec(cls.__init__)


class Directive(object):
    __metaclass__ = DirectiveMeta

    def __new__(cls, *args, **kwargs):
        init_args, init_varargs, init_keywords, init_defaults = cls.init_argspec

        # Don't care about "self".
        init_args = init_args[1:]

        if init_defaults is None:
            init_defaults = ()

        if init_keywords:
            for k in kwargs:
                if (k not in init_args) and (k not in cls.keywords):
                    raise RebarSyntaxError(
                      'Unexpected keyword argument to %s: %s' % (
                        cls.__name__,
                        k,
                      ),
                    )

        passed_args = set([])
        # Note [1:] to skip "self".
        for arg_name, arg_value in zip(init_args, args):
            passed_args.add(arg_name)
        passed_args.update(kwargs.keys())

        required_args = init_args[:(len(init_args) - len(init_defaults))]

        for arg_name in required_args:
            if arg_name not in passed_args:
                raise MissingRequiredArgumentError(arg_name)

        instance = super(Directive, cls).__new__(cls)
        instance.init_args = (args, kwargs)

        return instance

    # We define this because object.__init__ cannot be inspected (see
    # DirectiveMeta).
    def __init__(self):
        super(Directive, self).__init__()

    def __unicode__(self):
        return self.format()

    def __repr__(self):
        return self.format()

    # This is useful for testing.
    def __eq__(self, other):
        return (
          (type(self) is type(other))
          and (self.__dict__ == other.__dict__)
        )

    def format(self):
        args, kwargs = self.init_args

        parts = [repr(arg) for arg in args]
        for k, v in kwargs.iteritems():
            parts.append('%s=%r' % (k, v))
        formatted_values = ', '.join(parts)

        return u'%s(%s)' % (
          self.__class__.__name__,
          formatted_values,
        )

    ###

    @classmethod
    def set_defaults(cls, environment):
        pass

    def apply(self, environment):
        pass

    def finalize(self, environment):
        pass


class ListExtendDirective(Directive):
    name = None
    value = None

    def __init__(self, *value):
        self.value = value
        super(ListExtendDirective, self).__init__()

    @classmethod
    def set_defaults(cls, environment):
        super(ListExtendDirective, cls).set_defaults(environment)

        environment[cls.name] = []

    def apply(self, environment):
        super(ListExtendDirective, self).apply(environment)

        environment[self.name].extend(self.value)


class ApplyOnceMixin(object):
    def apply(self, environment):
        name = type(self).__name__
        k = '_%s_applied' % name
        if environment.get(k, False):
            raise RebarSyntaxError(
              '%s directive can only be used once.'
              % name
            )

        super(ApplyOnceMixin, self).apply(environment)

        environment[k] = True


###


class Platform(ApplyOnceMixin, Directive):
    architecture = None
    hostname = None

    def __init__(self, architecture = None, hostname = None):
        self.architecture = architecture
        self.hostname = hostname

        super(Platform, self).__init__()

    @classmethod
    def set_defaults(cls, environment):
        super(Platform, cls).set_defaults(environment)

        environment['platform'] = {
          'architecture': None,
          'hostname': None,
        }

    def apply(self, environment):
        super(Platform, self).apply(environment)

        environment['platform'].update({
          'architecture': self.architecture,
          'hostname': self.hostname,
        })

RebarEnvironment.register_directive(Platform)


class Bootstrap(ApplyOnceMixin, Directive):
    mirror = None
    suite = None
    script = None
    include = None
    exclude = None

    def __init__(self, mirror, suite, script, include = (), exclude = ()):
        self.mirror = mirror
        self.suite = suite
        self.script = script
        self.include = include
        self.exclude = exclude

        super(Bootstrap, self).__init__()

    @classmethod
    def set_defaults(cls, environment):
        super(Bootstrap, cls).set_defaults(environment)

        environment['bootstrap'] = None

    def apply(self, environment):
        super(Bootstrap, self).apply(environment)

        environment['bootstrap'] = {
          'mirror': self.mirror,
          'suite': self.suite,
          'script': self.script,
          'include': self.include,
          'exclude': self.exclude,
        }

RebarEnvironment.register_directive(Bootstrap)


class Repository(Directive):
    mirror = None
    suite = None
    components = None
    architectures = None
    trusted = None

    def __init__(
      self,
      mirror,
      suite,
      components,
      architectures = None,
      trusted = False,
    ):
        self.mirror = mirror
        self.suite = suite
        self.components = components
        self.architectures = architectures
        self.trusted = trusted

        super(Repository, self).__init__()

    @classmethod
    def set_defaults(cls, environment):
        super(Repository, cls).set_defaults(environment)

        environment['repositories'] = []

    def apply(self, environment):
        super(Repository, self).apply(environment)

        environment['repositories'].append({
          'mirror': self.mirror,
          'suite': self.suite,
          'components': self.components,
          'architectures': self.architectures,
          'trusted': self.trusted,
        })

    def finalize(self, environment):
        super(Repository, self).finalize(environment)

        if not environment.get('repositories', None):
            raise RebarSyntaxError('At least one repository must be specified.')

RebarEnvironment.register_directive(Repository)


class PackageListExtendDirective(ListExtendDirective):
    def __init__(self, *pkgnames):
        name = type(self).__name__

        for pkgname in pkgnames:
            for ch in '?*|[^$':
                if ch in pkgname:
                    raise RebarSyntaxError(
                      '%s: regular expressions are not supported' % name
                    )

            # FIXME: A more generic test (probably regular expression-based)
            # would be better.
            if "'" in pkgname:
                raise RebarSyntaxError(
                  '%s: illegal character in package name: %s'
                  % (name, pkgname)
                )

        super(PackageListExtendDirective, self).__init__(*pkgnames)


class Require(PackageListExtendDirective):
    name = 'required_packages'

RebarEnvironment.register_directive(Require)


class Install(PackageListExtendDirective):
    name = 'install_packages'

RebarEnvironment.register_directive(Install)


class BinaryDirs(ListExtendDirective):
    name = 'binary_dirs'

RebarEnvironment.register_directive(BinaryDirs)


class SourceDirs(ListExtendDirective):
    name = 'source_dirs'

RebarEnvironment.register_directive(SourceDirs)


class ExportDirs(ListExtendDirective):
    name = 'export_dirs'

RebarEnvironment.register_directive(ExportDirs)


###


class ActionGroupDirective(Directive):
    name = None

    def __init__(self, *action_directives):
        # Check that we were not passed something we didn't expect
        # (particularly a TargetDirective).
        for action_directive in action_directives:
            if not isinstance(action_directive, ActionDirective):
                raise RebarSyntaxError(
                  'arguments to %s must be actions'
                  % type(self).__name__
                )

        # All we do here is set the action_group_name for the directives we
        # were passed.
        for action_directive in action_directives:
            action_directive.action_group_name = self.name

        super(ActionGroupDirective, self).__init__()

    @classmethod
    def set_defaults(cls, environment):
        super(ActionGroupDirective, cls).set_defaults(environment)

        environment[cls.name] = []


class PreBuild(ActionGroupDirective):
    name = 'pre_build_actions'

RebarEnvironment.register_directive(PreBuild)


class PostBuild(ActionGroupDirective):
    name = 'post_build_actions'

RebarEnvironment.register_directive(PostBuild)


class PreInstall(ActionGroupDirective):
    name = 'pre_install_actions'

RebarEnvironment.register_directive(PreInstall)


class PostInstall(ActionGroupDirective):
    name = 'post_install_actions'

RebarEnvironment.register_directive(PostInstall)


class PreRemove(ActionGroupDirective):
    name = 'pre_remove_actions'

RebarEnvironment.register_directive(PreRemove)


class PostRemove(ActionGroupDirective):
    name = 'post_remove_actions'

RebarEnvironment.register_directive(PostRemove)


class PreShip(ActionGroupDirective):
    name = 'pre_ship_actions'

RebarEnvironment.register_directive(PreShip)


class PostShip(ActionGroupDirective):
    name = 'post_ship_actions'

RebarEnvironment.register_directive(PostShip)


class BuildStepDirective(Directive):
    def get_build_step(self, env):
        raise NotImplementedError()


class ActionDirective(BuildStepDirective):
    action_group_name = PostRemove.name

    def apply(self, environment):
        super(ActionDirective, self).apply(environment)
        environment[self.action_group_name].append(self)


class Exec(ActionDirective):
    keywords = {'using': PLATFORM}

    cmd = None
    values = None
    using = None

    def __init__(self, cmd, *values, **kwargs):
        using = kwargs.pop('using', self.keywords['using'])

        if using not in (PLATFORM, VIRTUALIZER):
            raise RebarSyntaxError('Invalid argument for "using": %r' % using)

        super(Exec, self).__init__()

        self.cmd = cmd
        self.values = values
        self.using = using

    def get_build_step(self, env):
        return ExecAction(
          env,
          self.cmd,
          using = self.using,
          name = unicode(self),
          *self.values
        )

RebarEnvironment.register_directive(Exec)


class Shell(ActionDirective):
    cmd = None
    values = None

    def __init__(self, cmd, *values):
        super(Shell, self).__init__()

        self.cmd = cmd
        self.values = values

    def get_build_step(self, env):
        return ShellAction(env, self.cmd, name = unicode(self), *self.values)

RebarEnvironment.register_directive(Shell)


class CopyTo(ActionDirective):
    src = None
    dst = None
    recurse = None
    dereference = None
    replace = None
    user = None
    group = None
    using = None

    def __init__(
      self,
      src,
      dst,
      recurse = False,
      dereference = False,
      replace = False,
      user = None,
      group = None,
      using = PLATFORM,
    ):
        if not dst.startswith('/'):
            raise RebarSyntaxError(
              'Destination must be an absolute path '
              'specifying the location within the chroot.'
            )

        if src.endswith('/') or dst.endswith('/'):
            raise RebarSyntaxError('CopyTo paths must not end with "/"')

        super(CopyTo, self).__init__()

        self.src = src
        self.dst = dst
        self.recurse = recurse
        self.dereference = dereference
        self.replace = replace
        self.user = user
        self.group = group
        self.using = using

    def get_build_step(self, env):
        return CopyToAction(
          env,
          self.src,
          self.dst,
          recurse = self.recurse,
          dereference = self.dereference,
          replace = self.replace,
          user = self.user,
          group = self.group,
          name = unicode(self),
          using = self.using,
        )

RebarEnvironment.register_directive(CopyTo)


class CopyFrom(ActionDirective):
    src = None
    dst = None
    recurse = None
    dereference = None
    using = None

    def __init__(
      self,
      src,
      dst,
      recurse = False,
      dereference = False,
      using = PLATFORM,
    ):
        if not src.startswith('/'):
            raise RebarSyntaxError(
              'Source must be an absolute path '
              'specifying the location within the chroot.'
            )

        if src.endswith('/') or dst.endswith('/'):
            raise RebarSyntaxError('CopyFrom paths must not end with "/"')

        super(CopyFrom, self).__init__()

        self.src = src
        self.dst = dst
        self.recurse = recurse
        self.dereference = dereference
        self.using = using

    def get_build_step(self, env):
        return CopyFromAction(
          env,
          self.src,
          self.dst,
          recurse = self.recurse,
          dereference = self.dereference,
          name = unicode(self),
          using = self.using,
        )

RebarEnvironment.register_directive(CopyFrom)


class Patch(ActionDirective):
    patch = None
    dir = None
    file = None
    strip = None

    def __init__(self, patch, dir = None, file = None, strip = 0):
        if (file is None) and (dir is None):
            raise RebarSyntaxError(
              'One of "file" or "dir" must be specified.'
            )

        super(Patch, self).__init__()

        self.patch = patch
        self.dir = dir
        self.file = file
        self.strip = strip

    def get_build_step(self, env):
        return PatchAction(
          env,
          patch = self.patch,
          dir = self.dir,
          file = self.file,
          strip = self.strip,
          name = unicode(self),
        )

RebarEnvironment.register_directive(Patch)


###


def check_geometry(geometry):
    if geometry is not None:
        if not (isinstance(geometry, tuple) and len(geometry) == 3):
            raise RebarSyntaxError('geometry must be a tuple (C, H, S)')


def check_image_type(type):
    if type not in (DISK, GZIMG):
        raise RebarSyntaxError('Unsupported image type %r.' % type)


def check_image_compression(compression):
    if compression not in (None, GZIP):
        raise RebarSyntaxError('Unsupported compression type %r.' % compression)


def check_partition(partition):
    if not isinstance(partition, dict):
        raise RebarSyntaxError('Partition must be a dictionary: %r' % partition)

    if 'type' not in partition:
        raise RebarSyntaxError('Partition type must be specified.')

    if 'size' not in partition:
        raise RebarSyntaxError('Partition size must be specified.')

    for k, v in partition.items():
        if k == 'type':
            if not (
              (isinstance(v, (int, long)) and (0 <= v < 256))
              or v in TYPES
            ):
                raise RebarSyntaxError('Invalid partition type: %r' % v)
        elif k == 'size':
            if not isinstance(v, UnitType):
                raise RebarSyntaxError(
                  'Partition size units must be specified.',
                )
        elif k == 'bootable':
            if k not in (True, False):
                raise RebarSyntaxError(
                  'Partition bootable flag must be True or False.',
                )
        else:
            raise RebarSyntaxError('Invalid partition parameter: %r' % k)


def check_partitions(partitions):
    for partition in partitions:
        check_partition(partition)


class TargetDirective(BuildStepDirective):
    @classmethod
    def set_defaults(cls, environment):
        super(TargetDirective, cls).set_defaults(environment)

        environment['targets'] = []

    def apply(self, environment):
        super(TargetDirective, self).apply(environment)

        environment['targets'].append(self)


class PlatformImage(TargetDirective):
    target_path = None
    geometry = None
    filesystem = None
    label = None
    bootloader = None
    fsexec = None
    devexec = None
    prepend_partitions = None
    type = None
    compression = None

    def __init__(
      self,
      target_path,
      geometry = None,
      filesystem = EXT2,
      label = None,
      bootloader = GRUB,
      fsexec = (),
      devexec = (),
      prepend_partitions = (),
      type = DISK,
      compression = None,
    ):
        check_geometry(geometry)
        check_image_type(type)
        check_image_compression(compression)

        check_partitions(prepend_partitions)
        if len(prepend_partitions) > 3:
            raise RebarSyntaxError('Unsupported partition layout.')

        super(PlatformImage, self).__init__()

        self.target_path = target_path
        self.geometry = geometry
        self.filesystem = filesystem
        self.label = label
        self.bootloader = bootloader
        self.fsexec = fsexec
        self.devexec = devexec
        self.prepend_partitions = prepend_partitions
        self.type = type
        self.compression = compression

    def get_build_step(self, env):
        prepend_partitions = [dict(p) for p in self.prepend_partitions]
        for partition in prepend_partitions:
            if isinstance(partition['type'], basestring):
                partition['type'] = TYPES[partition['type']]
            if 'bootable' not in partition:
                partition['bootable'] = False

        return PlatformImageTarget(
          env,
          self.target_path,
          geometry = self.geometry,
          filesystem = self.filesystem,
          label = self.label,
          bootloader = self.bootloader,
          fsexec = self.fsexec,
          devexec = self.devexec,
          prepend_partitions = prepend_partitions,
          type = self.type,
          compression = self.compression,
          name = unicode(self),
        )

RebarEnvironment.register_directive(PlatformImage)


class DirImage(TargetDirective):
    target_path = None
    source_dir = None
    geometry = None
    filesystem = None
    label = None
    bootloader = None
    fsexec = None
    type = None
    compression = None

    def __init__(
      self,
      target_path,
      source_dir,
      geometry = None,
      filesystem = VFAT,
      label = None,
      bootloader = SYSLINUX,
      fsexec = (),
      type = DISK,
      compression = None,
    ):
        check_geometry(geometry)
        check_image_type(type)
        check_image_compression(compression)

        super(DirImage, self).__init__()

        self.target_path = target_path
        self.source_dir = source_dir
        self.geometry = geometry
        self.filesystem = filesystem
        self.label = label
        self.bootloader = bootloader
        self.fsexec = fsexec
        self.type = type
        self.compression = compression

    def get_build_step(self, env):
        return DirImageTarget(
          env,
          self.target_path,
          self.source_dir,
          geometry = self.geometry,
          filesystem = self.filesystem,
          label = self.label,
          bootloader = self.bootloader,
          fsexec = self.fsexec,
          type = self.type,
          compression = self.compression,
          name = unicode(self),
        )

RebarEnvironment.register_directive(DirImage)


class ImagingTargetDirective(TargetDirective):
    @classmethod
    def set_defaults(cls, environment):
        super(ImagingTargetDirective, cls).set_defaults(environment)

        environment.setdefault('required_packages', [])

    def apply(self, environment):
        super(ImagingTargetDirective, self).apply(environment)

        # FIXME: Are other packages are required? Perhaps coreutils?
        # Add packages with utilities needed by gzimg_write.
        environment['required_packages'].extend(['gawk'])


class DiskImagingDir(ImagingTargetDirective):
    target_path = None
    source_image = None

    def __init__(self, target_path, source_image):
        super(DiskImagingDir, self).__init__()

        self.target_path = target_path
        self.source_image = source_image

    def get_build_step(self, env):
        return DiskImagingDirTarget(
          env,
          self.target_path,
          self.source_image,
          name = unicode(self),
        )

RebarEnvironment.register_directive(DiskImagingDir)


class DiskImagingImage(ImagingTargetDirective):
    target_path = None
    source_image = None
    geometry = None
    filesystem = None
    label = None
    bootloader = None
    fsexec = None
    type = None
    compression = None

    def __init__(
      self,
      target_path,
      source_image,
      geometry = None,
      filesystem = VFAT,
      label = None,
      bootloader = SYSLINUX,
      fsexec = (),
      type = DISK,
      compression = None,
    ):
        check_geometry(geometry)
        check_image_type(type)
        check_image_compression(compression)

        super(DiskImagingImage, self).__init__()

        self.target_path = target_path
        self.source_image = source_image
        self.geometry = geometry
        self.filesystem = filesystem
        self.label = label
        self.bootloader = bootloader
        self.fsexec = fsexec
        self.type = type
        self.compression = compression

    def get_build_step(self, env):
        return DiskImagingImageTarget(
          env,
          self.target_path,
          self.source_image,
          geometry = self.geometry,
          filesystem = self.filesystem,
          label = self.label,
          bootloader = self.bootloader,
          fsexec = self.fsexec,
          type = self.type,
          compression = self.compression,
          name = unicode(self),
        )

RebarEnvironment.register_directive(DiskImagingImage)


class NetworkImagingDir(ImagingTargetDirective):
    target_path = None
    source_image = None
    image_url = None

    def __init__(self, target_path, source_image, image_url):
        super(NetworkImagingDir, self).__init__()

        self.target_path = target_path
        self.source_image = source_image
        self.image_url = image_url

    def get_build_step(self, env):
        return NetworkImagingDirTarget(
          env,
          self.target_path,
          self.source_image,
          self.image_url,
          name = unicode(self),
        )

RebarEnvironment.register_directive(NetworkImagingDir)


class Target(TargetDirective):
    target_path = None
    action_directives = None

    def __init__(self, target_path, action_directives):
        super(Target, self).__init__()

        self.target_path = target_path
        self.action_directives = action_directives

    def apply(self, environment):
        super(Target, self).apply(environment)

        # Remove our directives from the environment.
        for action_directive in self.action_directives:
            environment[action_directive.action_group_name].remove(
              action_directive,
            )

    def get_build_step(self, env):
        actions = []
        for action_directive in self.action_directives:
            actions.append(action_directive.get_build_step(env))

        return ActionSequenceTarget(
          env,
          self.target_path,
          actions,
          name = unicode(self),
        )

RebarEnvironment.register_directive(Target)


###


class Extension(Directive):
    name_re = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*')

    name = None
    functions = None

    def __init__(self, name, functions):
        if not self.name_re.match(name):
            raise RebarSyntaxError('Invalid extension name: %r' % name)

        for function in functions:
            if not callable(function):
                raise RebarSyntaxError('Not a function: %r' % function)

        super(Extension, self).__init__()

        self.name = name
        self.functions = functions

    def apply(self, environment):
        super(Extension, self).apply(environment)

        d = dict((f.__name__, staticmethod(f)) for f in self.functions)
        extension = type(self.name, (object,), d)
        try:
            environment.contribute_to_parent(self.name, extension)
        except KeyError:
            raise RebarSyntaxError('Extension name clash: %s', extname)

RebarEnvironment.register_directive(Extension)
