# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement

import os, sys, re, tempfile

from pshell import CommandFailed
from fdisk.mbr import MBRDisk, TYPES

from rebar.util import locate_data_file
from rebar.exceptions import DiskImageError


# FIXME: We should probably pull this from the chroot.
MBR_FILE = '/usr/lib/syslinux/mbr.bin'


def check_partition_index(partition_index):
    if not (
      isinstance(partition_index, (int, long))
      and partition_index >= 0
    ):
        raise ValueError(partition_index)


class DiskImage(object):
    platform = None
    filename = None
    mbr_disk = None

    device_path = None

    # FIXME: Support geometry detection here?
    def __init__(self, platform, filename, geometry):
        super(DiskImage, self).__init__()

        self.platform = platform
        self.filename = filename
        self.mbr_disk = MBRDisk(geometry)

        # Ignore failure because image file may not exist yet.
        self._open(ignore_failure = True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.closed:
            self.close()

    ###

    def _require_open(self):
        if self.closed:
            raise AssertionError('closed')

    def _open(self, ignore_failure = False):
        with self.platform.open_shell() as sh:
            try:
                status, output = sh.execute(
                  'losetup --show --find ?',
                  self.filename,
                )
            except CommandFailed:
                if not ignore_failure:
                    raise
            else:
                device_path = output.strip()
                if not device_path:
                    raise AssertionError(device_path)
                self.device_path = device_path

    closed = property(lambda self: (not self.device_path))

    def close(self):
        # FIXME: Prevent closing disk image with open partitions?

        self._require_open()

        with self.platform.open_shell() as sh:
            # To prevent possible EBUSY releasing the loopback device:
            sh.execute('sync')
            sh.execute('losetup --detach ?', self.device_path)

        del self.device_path

    def create(self):
        if not self.closed:
            self.close()

        # Create sparse file.
        with self.platform.open_shell() as sh:
            sh.execute(
              'dd if=/dev/null of=? bs=512 seek=?',
              self.filename,
              long(self.mbr_disk.size.as_sectors()),
            )

        self._open()

    def install_mbr(self):
        self._require_open()

        # FIXME: We should ship mbr.bin or install it from the archives.
        # Copying it from the host system is reasonable, but introduces a
        # dependency on syslinux that otherwise wouldn't be needed (since we
        # need syslinux in the virtualized platform anyway).
        with self.platform.open_shell() as sh:
            sh.execute('dd if=? of=? conv=notrunc', MBR_FILE, self.device_path)

    def create_partition_table(self):
        self._require_open()

        # FIXME: Consider using sfdisk dump input instead of interactive input.
        lines = []
        for line in self.mbr_disk.iter_sfdisk_input_lines():
            if line.endswith('\n'):
                line = line[:-1]
            lines.append(line)

        with self.platform.open_shell() as sh:
            C, H, S = self.mbr_disk.geometry
            sh.execute(
              'sfdisk -uS -x -C? -H? -S? ?',
              C,
              H,
              S,
              self.device_path,
              wait = False,
            )

            for line in lines:
                sh.sendline(line)

            sh.expect(re.escape('[ynq]'))
            sh.sendline('y')

            sh.wait_for_prompt(timeout = 600)
            sh.check_last_status()

    def open_partition(self, index):
        self._require_open()
        return DiskImagePartition(self, index)

    def save_copy(self, path):
        self.platform.copy_from(self.filename, path)

    def remove(self):
        with self.platform.open_shell() as sh:
            sh.execute('rm -f ?', self.filename)

    def install_grub(self, root_index = 0, grub_index = None):
        self._require_open()

        check_partition_index(root_index)

        if grub_index is None:
            # Install GRUB in MBR.
            grub_device = '(hd0)'
        else:
            # Install GRUB in VBR.
            check_partition_index(grub_index)
            grub_device = '(hd0,%u)' % grub_index

        with self.platform.open_shell() as sh:
            with self.open_partition(root_index) as partition:
                mount_point = partition.mount()
                try:
                    # FIXME: There is probably a better way to detect the
                    # platform architecture.
                    # FIXME: It should be possible to use GRUB on other
                    # architectures, but there's no simple way to detect the
                    # GRUB basic machine type without a hard-coded mapping like
                    # we have here.
                    status, output = sh.execute(
                      'chroot ? /usr/bin/dpkg --print-architecture',
                      mount_point,
                    )
                    dpkg_arch = output.strip()
                    if dpkg_arch == 'i386':
                        grub_machine = 'i386-pc'
                    elif dpkg_arch == 'amd64':
                        grub_machine = 'x86_64-pc'
                    else:
                        raise AssertionError(
                          'invalid architecture: %r' % dpkg_arch
                        )

                    sh.execute('mkdir -p ?/boot/grub', mount_point)
                    # Note: We use --no-clobber to avoid overwriting files
                    # that the user already copied to /boot/grub.
                    sh.execute(
                      (
                        'cp --archive --one-file-system --no-clobber '
                        '?/usr/lib/grub/?/* ?/boot/grub/'
                      ),
                      mount_point,
                      grub_machine,
                      mount_point,
                    )
                finally:
                    partition.umount(mount_point)

            # FIXME: We assume that the base platform's GRUB is 100% compatible
            # with the target platform.  Is this a reasonable assumption with
            # differing architectures?

            # Make sure grub is usable, otherwise things will fail in a bad way
            # (we'll write grub input directly to the shell and then send an
            # EOF, exiting the shell).
            sh.execute('which grub >/dev/null')

            # NOTE: GRUB chokes ("Error 22: No such partition") if we give it
            # the loopback device, not sure why.  So we detach the loopback
            # device and work with the image directly.
            self.close()
            try:
                sh.execute(
                  'grub --batch --no-curses --device-map=/dev/null --no-floppy',
                  wait = False,
                )
                self._post_grub_command(sh)

                sh.sendline('device (hd0) %s' % self.filename)
                self._post_grub_command(sh)

                sh.sendline('geometry (hd0) %s %s %s' % self.mbr_disk.geometry)
                self._post_grub_command(sh)

                sh.sendline('root (hd0,%u)' % root_index)
                self._post_grub_command(sh)

                sh.sendline('setup %s' % grub_device)
                self._post_grub_command(sh)

                sh.sendline('quit')
                sh.wait_for_prompt(timeout = 10)
                sh.check_last_status()

            finally:
                self._open()

    def _post_grub_command(self, sh):
        sh.expect('\r\ngrub\> ', timeout = 30)
        if 'Error' in sh.before:
            msg = sh.before.strip()
            # FIXME: It should be sufficient to send an EOL, right?
            sh.interrupt()
            raise DiskImageError(msg)

    def install_grub2(self, root_index = 0, grub_index = None):
        self._require_open()

        check_partition_index(root_index)

        with self.open_partition(root_index) as partition:
            if grub_index is None:
                grub_device_path = self.device_path
            else:
                check_partition_index(grub_index)
                grub_device_path = partition.device_path

            with open(locate_data_file('grub-probe.sh'), 'rb') as f:
                probe_content = f.read()

            mount_point = partition.mount()
            try:
                for name, value in (
                  ('SYSTEM_DRIVE', self.device_path),
                  ('SYSTEM_PARTITION', partition.device_path),
                  # FIXME: We should confirm that the filesystem is
                  # one of ext2, ext3, ext4.
                  ('FS', 'ext2'),
                  ('PARTMAP', 'msdos'),
                  ('ABSTRACTION', ''),
                  ('GRUB_DRIVE', 'hd0'),
                  ('GRUB_PARTITION', ('msdos%u' % (partition.index + 1))),
                ):
                    probe_content = probe_content.replace(
                      ('%%' + name + '%%'),
                      value,
                    )

                with tempfile.NamedTemporaryFile('wb') as f:
                    f.write(probe_content)
                    f.flush()

                    with self.platform.open_shell() as sh:
                        status, output = sh.execute('mktemp')
                        grub_probe = output.strip()
                        try:
                            self.platform.copy_to(f.name, grub_probe)
                            sh.execute('chmod a+x ?', grub_probe)
                            sh.execute(
                              (
                                'grub-install --debug --no-floppy --force '
                                '--boot-directory=?/boot --grub-probe=? ?'
                              ),
                              mount_point,
                              grub_probe,
                              grub_device_path,
                            )
                            sh.execute(
                              (
                                'cp ?/usr/sbin/grub-probe '
                                '?/usr/sbin/grub-probe.bak'
                              ),
                              mount_point,
                              mount_point,
                            )
                            try:
                                sh.execute(
                                  'cp -af ? ?/usr/sbin/grub-probe',
                                  grub_probe,
                                  mount_point,
                                )
                                # NOTE: We assume that the GRUB configuration
                                # (i.e. /etc/default/grub) was already written.
                                sh.execute(
                                  'chroot ? /usr/sbin/update-grub',
                                  mount_point,
                                )
                            finally:
                                sh.execute(
                                  (
                                    'mv -f ?/usr/sbin/grub-probe.bak '
                                    '?/usr/sbin/grub-probe'
                                  ),
                                  mount_point,
                                  mount_point,
                                )
                        finally:
                            sh.execute('rm -f ?', grub_probe)

            finally:
                partition.umount(mount_point)

    def get_size(self):
        return self.mbr_disk.size.normalize()


class DiskImagePartition(object):
    disk_image = None
    index = None

    closed = None

    def __init__(self, disk_image, index):
        check_partition_index(index)

        super(DiskImagePartition, self).__init__()

        self.disk_image = disk_image
        self.index = index

        self._open()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.closed:
            self.close()

    def _require_open(self):
        if self.closed:
            raise AssertionError('closed')

    def _open(self):
        found = False
        for index, partition in enumerate(
          self.disk_image.mbr_disk.iter_partitions_recursive()
        ):
            if index == self.index:
                if partition.is_empty():
                    # FIXME: Better exception?
                    raise AssertionError("can't open an empty partition")
                if partition.is_extended():
                    # FIXME: Better exception?
                    raise AssertionError("can't open an extended partition")
                found = True
                break

        if not found:
            # FIXME: Better exception?
            raise AssertionError(
              'partition with index %u not found' % self.index
            )

        with self.disk_image.platform.open_shell() as sh:
            sh.execute(
              'partx --add ? ?',
              (self.index + 1),
              self.disk_image.device_path,
            )

        self.closed = False

    def close(self):
        self._require_open()

        with self.disk_image.platform.open_shell() as sh:
            sh.execute(
              'partx --delete ? ?',
              (self.index + 1),
              self.disk_image.device_path,
            )

        self.closed = True

    def create_filesystem(self, filesystem_name, label = None):
        if filesystem_name not in ('ext2', 'ext3', 'ext4', 'msdos', 'vfat'):
            raise ValueError(filesystem_name)

        options = []

        if label is not None:
            if filesystem_name in ('ext2', 'ext3', 'ext4'):
                options.extend(['-L', label])
            elif filesystem_name in ('msdos', 'vfat'):
                options.extend(['-n', label])
            else:
                raise AssertionError(filesystem_name)

        self._require_open()

        with self.disk_image.platform.open_shell() as sh:
            if options:
                cmd = 'mkfs -t ? %s ?' % ' '.join(options)
            else:
                cmd = 'mkfs -t ? ?'
            sh.execute(cmd, filesystem_name, self.device_path)

    def install_syslinux(self):
        self._require_open()

        with self.disk_image.platform.open_shell() as sh:
            sh.execute('syslinux ?', self.device_path)

    def mount(self):
        self._require_open()

        with self.disk_image.platform.open_shell() as sh:
            status, output = sh.execute('mktemp -d')
            mount_point = output.strip()
            try:
                sh.execute('mount -n ? ?', self.device_path, mount_point)
            except:
                sh.execute('rmdir ?', mount_point)
                raise
        return mount_point

    def umount(self, mount_point):
        with self.disk_image.platform.open_shell() as sh:
            sh.execute('umount -n ?', mount_point)
            sh.execute('rmdir ?', mount_point)

    @property
    def device_path(self):
        return '%sp%u' % (self.disk_image.device_path, (self.index + 1))
