# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, copy, errno

from fdisk import units

from rebar.const import SymbolicConstant, BootLoader


PLATFORM = SymbolicConstant('PLATFORM')
VIRTUALIZER = SymbolicConstant('VIRTUALIZER')

DISK = SymbolicConstant('DISK')
GZIMG = SymbolicConstant('GZIMG')

MBR = SymbolicConstant('MBR')
VBR = SymbolicConstant('VBR')

GRUB = BootLoader('GRUB', (MBR, VBR))
GRUB2 = BootLoader('GRUB2', (MBR, VBR))
SYSLINUX = BootLoader('SYSLINUX', (VBR,))

EXT2 = SymbolicConstant('EXT2')
EXT3 = SymbolicConstant('EXT3')
EXT4 = SymbolicConstant('EXT4')
VFAT = SymbolicConstant('VFAT')
MSDOS = SymbolicConstant('MSDOS')

GZIP = SymbolicConstant('GZIP')


class RebarEnvironment(dict):
    directive_types = []

    @classmethod
    def register_directive(cls, directive_type):
        for directive_type_ in cls.directive_types:
            if directive_type_.__name__ == directive_type.__name__:
                raise ValueError(
                  'directive with name %s is already registered'
                  % directive_type_.__name__
                )
        cls.directive_types.append(directive_type)

    locals_stack = None
    to_apply = None
    to_finalize = None
    include_path = None

    def __init__(self, *args):
        self.locals_stack = []
        self.to_apply = []
        self.to_finalize = []

        self.set_defaults()

        super(RebarEnvironment, self).__init__(*args)

    def push_locals(self, name):
        locals = copy.copy(self.get_initial_locals())
        locals['__name__'] = name
        self.locals_stack.append(locals)

        return locals

    def pop_locals(self):
        return self.locals_stack.pop()

    def get_locals(self):
        return self.locals_stack[-1]

    def get_parent_locals(self):
        return self.locals_stack[-2]

    def set_include_path(self, include_path):
        prev_include_path = self.include_path
        self.include_path = include_path
        return prev_include_path

    def load_from_file(self, filename, ignore_missing = False):
        def do_exec(locals):
            try:
                execfile(filename, locals)
            except IOError, e:
                if not (ignore_missing and e.errno == errno.ENOENT):
                    raise

        # FIXME: Do we need to do any path validation here?
        include_path = os.path.dirname(filename)
        prev_include_path = self.set_include_path(include_path)
        try:
            name = '__' + os.path.basename(filename).replace('.', '_') + '__'
            self._load(name, do_exec)
        finally:
            self.set_include_path(prev_include_path)

    def load_from_string(self, s, name = '__rebar__'):
        def do_exec(locals):
            exec s in locals

        prev_include_path = self.set_include_path(None)
        try:
            self._load(name, do_exec)
        finally:
            self.set_include_path(prev_include_path)

    def _load(self, name, do_exec):
        locals = self.push_locals(name)
        try:
            # Execute the Rebar file contents.  This populates the to_apply
            # list since our directive wrapper functions are called for each
            # time a directive is used in the Rebar file.
            for directive_type in self.directive_types:
                f = self._make_directive_wrapper(directive_type)
                locals[directive_type.__name__] = f

            # Execute the Rebar file contents.  This populates the directives
            # list since our directive wrapper functions are called for each
            # time a directive is used in the Rebar file.
            do_exec(locals)

            self.apply_directives()
            self.finalize_directives()

        finally:
            self.pop_locals()

    def _make_directive_wrapper(self, cls):
        def directive_wrapper(*args, **kwargs):
            directive_instance = cls(*args, **kwargs)
            self.to_apply.append(directive_instance)
            # FIXME: We have to return the instance so that e.g.
            # ActionGroupDirectives can have access to them, but that exposes
            # our internal directive classes inside the Rebar file locals
            # namespace, which is potentially a security problem in the future?
            return directive_instance
        return directive_wrapper

    def include(self, filename, ignore_missing = False):
        # FIXME: Should we disallow includes when include_path is None
        # (i.e. when we are not loading from a file)?
        if self.include_path is not None:
            filename = os.path.join(self.include_path, filename)
        self.load_from_file(filename, ignore_missing = ignore_missing)

    def contribute_to_parent(self, name, value):
        parent_locals = self.get_parent_locals()
        if name in parent_locals:
            raise KeyError(name)
        parent_locals[name] = value

    def set_defaults(self):
        for directive_type in self.directive_types:
            directive_type.set_defaults(self)

    def apply_directives(self):
        while self.to_apply:
            directive = self.to_apply.pop(0)
            directive.apply(self)
            self.to_finalize.append(directive)

    def finalize_directives(self):
        while self.to_finalize:
            directive = self.to_finalize.pop(0)
            directive.finalize(self)

    def get_initial_locals(self):
        return {
          '__builtins__': None,
          'True': True,
          'False': False,
          'Include': self.include,

          'PLATFORM': PLATFORM,
          'VIRTUALIZER': VIRTUALIZER,

          'DISK': DISK,
          'GZIMG': GZIMG,

          'GRUB': GRUB,
          'GRUB2': GRUB2,
          'SYSLINUX': SYSLINUX,

          'MBR': MBR,
          'VBR': VBR,

          'EXT2': EXT2,
          'EXT3': EXT3,
          'EXT4': EXT4,
          'MSDOS': MSDOS,
          'VFAT': VFAT,

          'GZIP': GZIP,

          'B': units.B,
          'bytes': units.bytes,
          'kB': units.kB,
          'kilobytes': units.kilobytes,
          'MB': units.MB,
          'megabytes': units.megabytes,
          'GB': units.GB,
          'gigabytes': units.gigabytes,
        }
