# Copyright (c) 2007-2010 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement 

import sys, signal
from optparse import OptionParser

try:
    from contextlib import closing
except ImportError:
    from rebar.legacy_support import closing

from sclapp2.safeio import with_safe_stdio
from sclapp2.signals import with_standard_signals
from sclapp2.ioenc import StdioEncodingContext

from rebar.virtualizer.chroot import ChrootVirtualizer
from rebar.environment import RebarEnvironment
from rebar.build import RebarBuild
from rebar.shells import Shell
from rebar.output import (
  ConciseOutput,
  DetailedOutput,
  VerboseOutput,
  ProgressIndicationMixin,
  LogFileOutputMixin,
)
from rebar.util import mixin_instance, print_rebar_file_traceback
from rebar.tasks_ import ComplexTask
from rebar.exceptions import NoSuchBuildPhase, VirtualizerInitializationError
from rebar.distinfo import version as rebar_version


REBAR_FILE = 'Rebar'


@with_standard_signals
@with_safe_stdio
def main(argv):
    encoding_context = StdioEncodingContext()
    with encoding_context:
        parser = OptionParser(
          usage = '%prog [OPTIONS] phase',
          version = rebar_version,
        )
        parser.add_option('--virtualizer', default = 'chroot:chroot')
        parser.add_option('--create-virtualizer', action = 'store_true')
        parser.add_option(
          '--output',
          choices = ('concise', 'detailed', 'verbose'),
          default = 'concise',
        )
        parser.add_option('--log', default = None)
        parser.add_option(
          '--no-progress',
          action = 'store_false',
          dest = 'progress',
          default = True,
        )
        parser.add_option('--debug', action = 'store_true', default = False)

        opts, args = parser.parse_args(argv[1:])

        if opts.output == 'concise':
            output = ConciseOutput(sys.stdout)
        elif opts.output == 'detailed':
            output = DetailedOutput(sys.stdout)
        elif opts.output == 'verbose':
            output = VerboseOutput(sys.stdout)
        else:
            raise AssertionError(opts.output)

        if opts.output in ('concise', 'detailed'):
            if sys.stdout.isatty() and opts.progress:
                mixin_instance(output, ProgressIndicationMixin)

        logfile = None
        if opts.log is not None:
            # Note: log file must be line buffered.
            logfile = open(opts.log, 'w', 1)
            try:
                mixin_instance(output, LogFileOutputMixin)
                output.set_logfile(logfile)
            except:
                logfile.close()
                raise

        try:
            parts = opts.virtualizer.split(':', 1)
            try:
                virtualizer_name, virtualizer_details = parts
            except ValueError:
                virtualizer_name = opts.virtualizer
                virtualizer_details = None

            if virtualizer_name == 'chroot':
                virtualizer_cls = ChrootVirtualizer
            else:
                parser.error('unsupported virtualizer: %s' % virtualizer_name)

            output.print_important('rebar %s' % rebar_version)
            output.print_important('Copyright Outpost Embedded, LLC and others.')
            output.print_important(
              'This is free software with ABSOLUTELY NO WARRANTY.')
            output.print_important('')

            if opts.create_virtualizer:
                # Disable encoding context for interactive initialization to
                # work properly (stdio encoding interferes with password
                # prompts).
                with encoding_context.disabled():
                    try:
                        with virtualizer_cls.Create(
                          virtualizer_details,
                          args,
                          command_log = logfile,
                          output_log = logfile,
                          interactive = True,
                        ) as task:
                            # Reenable stdio encoding.
                            with encoding_context.enabled():
                                # Note: We setup a task tree like this:
                                #
                                # root
                                # `- create
                                #    `- [virtualizer create task]
                                #
                                # This is necessary to fulfill the output
                                # module's expectations for task structures,
                                # which are ultimately based around
                                # RebarBuilds (create is equivalent to a build
                                # phase).
                                create = ComplexTask(
                                  'create',
                                  subtasks = [task],
                                )
                                root = ComplexTask(
                                  'root',
                                  subtasks = [create],
                                )
                                for task_changes in root.run():
                                    output.handle_task_changes(task_changes)
                    except VirtualizerInitializationError, e:
                        parser.error(unicode(e))
                        return 1
                return 0

            env = RebarEnvironment()

            try:
                env.load_from_file(REBAR_FILE)
            except KeyboardInterrupt:
                raise
            except Exception:
                if opts.debug:
                    import traceback
                    traceback.print_exc()
                else:
                    print_rebar_file_traceback()
                return 1

            try:
                build = RebarBuild(env, (args or None))
            except NoSuchBuildPhase, e:
                parser.error(unicode(e))

            user_shell = Shell(command_log = logfile, output_log = logfile)
            with closing(user_shell):
                # Disable encoding context for interactive initialization to
                # work properly (stdio encoding interferes with password
                # prompts).
                with encoding_context.disabled():
                    # FIXME: We should somehow check that the virtualizer
                    # architecture and distribution match the platform
                    # architecture and distribution.
                    with virtualizer_cls.Initialized(
                      virtualizer_details,
                      env,
                      command_log = logfile,
                      output_log = logfile,
                      interactive = True,
                    ) as virtualizer:
                        # Reenable stdio encoding.
                        with encoding_context.enabled():
                            virtualizer.start()
                            try:
                                build.bind(virtualizer, user_shell, output)
                                try:
                                    for task_changes in build.run():
                                        output.handle_task_changes(task_changes)
                                finally:
                                    build.unbind()
                            finally:
                                virtualizer.stop()

        finally:
            if logfile is not None:
                logfile.close()

        if build.failed:
            return 1
        return 0
