# Copyright (c) 2009-2010 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys
from contextlib import contextmanager

from sclapp2.termcontrol import get_caps

from rebar.tasks_ import ComplexTask


class NullFile(object):
    def write(self, s):
        pass

    def flush(self):
        pass

    def close(self):
        pass


class Output(object):
    f_out = None

    def __init__(self, f_out):
        self.f_out = f_out

    ###

    def format_message(self, msg, *args):
        if args:
            return msg % args
        return msg

    def print_debug(self, msg, *args):
        self.f_out.write('%s\n' % self.format_message(msg, *args))

    def print_verbose(self, msg, *args):
        self.f_out.write('%s\n' % self.format_message(msg, *args))

    def print_important(self, msg, *args):
        self.f_out.write('%s\n' % self.format_message(msg, *args))

    ###

    @contextmanager
    def open_debug_stream(self):
        yield self.f_out

    @contextmanager
    def open_verbose_stream(self):
        yield self.f_out

    @contextmanager
    def open_important_stream(self):
        yield self.f_out

    ###

    def handle_task_changes(self, task_changes):
        for task_change in task_changes:
            self.print_important(unicode(task_change))


class ConciseOutput(Output):
    build_phase = None

    def print_debug(self, msg, *args):
        pass

    def print_verbose(self, msg, *args):
        pass

    @contextmanager
    def open_debug_stream(self):
        yield NullFile()

    @contextmanager
    def open_verbose_stream(self):
        yield NullFile()

    @contextmanager
    def open_important_stream(self):
        yield self.f_out

    def truncate(self, s, length):
        len_s = len(s)
        if len_s <= length:
            return s
        s = '%s...' % s[:(length - 3)]
        return s

    def handle_task_changes(self, task_changes):
        for task_change in task_changes:
            if task_change.parameter == 'running':
                if task_change.task.depth == 1:
                    if task_change.value:
                        self.f_out.write('[%s]\n' % task_change.task.name)
                    else:
                        self.f_out.write('\n')

                elif task_change.task.depth == 2:
                    if task_change.value:
                        self.f_out.write('%s\n' % task_change.task.name)

            elif (
              task_change.task.depth == 0
            ) and (
              task_change.parameter == 'failed'
            ) and (
              task_change.value
            ):
                if task_change.task.failure is None:
                    self.f_out.write('FAIL\n')
                else:
                    self.f_out.write('FAIL: %s\n' % task_change.task.failure)


class DetailedOutput(ConciseOutput):
    def handle_task_changes(self, task_changes):
        super(DetailedOutput, self).handle_task_changes(task_changes)
        for task_change in task_changes:
            if task_change.task.depth >= 2:
                desc = None

                if isinstance(task_change.task, ComplexTask):
                    if (
                      task_change.parameter == 'running_subtask'
                    ) and (
                      task_change.value is not None
                    ):
                        desc = task_change.value.name
                else:
                    if task_change.parameter == 'status':
                        desc = task_change.value

                if desc is not None:
                    self.f_out.write('%s%s\n' % (
                      '  ' * (task_change.task.depth - 1),
                      desc,
                    ))


class VerboseOutput(Output):
    pass


class NullOutput(Output):
    def __init__(self, f_out = None):
        super(NullOutput, self).__init__(NullFile())


class Spinner(object):
    current = None
    STATES = ('-', '\\', '|', '/')

    def __init__(self):
        self.current = 0

    def next(self):
        val = self.STATES[self.current]
        self.current = self._get_next()
        return val

    def _get_next(self):
        index = self.current + 1
        if index >= len(self.STATES):
            index = 0
        return index


# Note: ProgressBar writes directly to sys.stdout.  There's no point in
# abstracting this away, because we rely on ncurses to tell us what the
# terminal capabilities are, and there's no way (that I'm aware of) to
# tell it to use a different file object.

class ProgressBar(object):
    width = None
    spinner = None

    caps = None

    def __init__(self, width = None):
        self.width = width
        self.spinner = Spinner()

    def clear(self):
        sys.stdout.write('%(BOL)s%(CLEAR_EOL)s' % self.caps)

    def draw(self, progress):
        self.caps = get_caps()

        if self.width is not None:
            width = self.width
        else:
            width = self.caps['COLS']
            if width is None:
                raise AssertionError()

        sys.stdout.write('%(BOL)s' % self.caps)

        bar_width = width - 7 - 2

        filled = int((progress * bar_width) / 100)

        if filled == bar_width:
            bar = filled * '='
        else:
            joiner = self.spinner.next()
            empty = bar_width - filled - 1
            bar = '%s%s%s' % (filled * '=', joiner, empty * '-')

        output = '%3u%% [%s]' % (progress, bar)
        output = output.ljust(width)

        sys.stdout.write(output)


class ProgressIndicationMixin(object):
    def handle_task_changes(self, task_changes):
        if not hasattr(self, 'current_progress'):
            self.current_progress = 0
            self.progress_bar = ProgressBar()
            self.set_progress_bar_enabled(True)
        else:
            if getattr(self, 'progress_bar_enabled', False):
                self.clear()

        super(ProgressIndicationMixin, self).handle_task_changes(task_changes)

        exiting = False
        for task_change in task_changes:
            if task_change.task.depth == 0:
                if task_change.parameter == 'progress':
                    self.current_progress = task_change.value
                elif (
                  task_change.parameter == 'running'
                  and not task_change.value
                ):
                    exiting = True

        if not exiting:
            if getattr(self, 'progress_bar_enabled', False):
                self.write_progress()

    def clear(self):
        self.progress_bar.clear()
        self.f_out.write('%(UP)s%(BOL)s%(CLEAR_EOL)s' % self.caps)
        self.progress_bar_visible = False

    def write_progress(self):
        self.caps = get_caps()
        self.f_out.write('\n')
        self.progress_bar.draw(self.current_progress)
        self.progress_bar_visible = True

    def set_progress_bar_enabled(self, value):
        old_value = getattr(self, 'progress_bar_enabled', False)
        self.progress_bar_enabled = value
        if value:
            if not getattr(self, 'progress_bar_visible', False):
                self.write_progress()
        else:
            if getattr(self, 'progress_bar_visible', False):
                self.clear()
        return old_value

    def print_debug(self, msg, *args):
        enabled = self.set_progress_bar_enabled(False)
        super(ProgressIndicationMixin, self).print_debug(msg, *args)
        self.set_progress_bar_enabled(enabled)

    def print_verbose(self, msg, *args):
        enabled = self.set_progress_bar_enabled(False)
        super(ProgressIndicationMixin, self).print_verbose(msg, *args)
        self.set_progress_bar_enabled(enabled)

    def print_important(self, msg, *args):
        enabled = self.set_progress_bar_enabled(False)
        super(ProgressIndicationMixin, self).print_important(msg, *args)
        self.set_progress_bar_enabled(enabled)

    @contextmanager
    def open_debug_stream(self):
        with super(ProgressIndicationMixin, self).open_debug_stream() as stream:
            enabled = self.set_progress_bar_enabled(False)
            yield stream
        self.set_progress_bar_enabled(enabled)

    @contextmanager
    def open_verbose_stream(self):
        with super(ProgressIndicationMixin, self).open_verbose_stream() as stream:
            enabled = self.set_progress_bar_enabled(False)
            yield stream
        self.set_progress_bar_enabled(enabled)

    @contextmanager
    def open_important_stream(self):
        with super(ProgressIndicationMixin, self).open_important_stream() as stream:
            enabled = self.set_progress_bar_enabled(False)
            yield stream
        self.set_progress_bar_enabled(enabled)


class LogFileOutputMixin(object):
    f_log = None

    def set_logfile(self, f_log):
        self.f_log = f_log

    def handle_task_changes(self, task_changes):
        super(LogFileOutputMixin, self).handle_task_changes(task_changes)

        for task_change in task_changes:
            if 0 < task_change.task.depth <= 2:
                if task_change.parameter == 'running':
                    if task_change.value:
                        self.f_log.write(
                          'RUNNING: %s\n' % task_change.task.name
                        )

                elif task_change.parameter == 'failed':
                    if task_change.value:
                        self.f_log.write(
                          'FAILED: %s\n' % task_change.task.name
                        )

                elif task_change.parameter == 'completed':
                    if task_change.value:
                        self.f_log.write(
                          'COMPLETED: %s\n' % task_change.task.name
                        )
