# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, tempfile

from pshell import CommandFailed

from rebar.diskimage import DiskImage
from rebar.util import join_abs


class PlatformContext(object):
    __platform = None
    __count = None

    def __init__(self, platform):
        super(PlatformContext, self).__init__()

        self.__platform = platform

        self.__count = 0

    def __enter__(self):
        if self.__count == 0:
            self.__platform._open()
        self.__count += 1
        return self.__platform

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__count -= 1
        if self.__count == 0:
            self.__platform._close()

    def suspend(self):
        if self.is_active():
            self.__platform._suspend()

    def resume(self):
        if self.is_active():
            self.__platform._resume()

    def is_active(self):
        return (self.__count > 0)


class ShellContext(object):
    __platform = None

    __count = None

    def __init__(self, platform):
        super(ShellContext, self).__init__()

        self.__platform = platform

        self.__count = 0

    def __enter__(self):
        self.__platform._check_is_open()
        if self.__count == 0:
            self.__platform._enter_shell_context()
        self.__count += 1
        return self.__platform._get_shell()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__count -= 1
        if self.__count == 0:
            self.__platform._exit_shell_context()

    def suspend(self):
        if self.is_active():
            self.__platform._exit_shell_context()

    def resume(self):
        if self.is_active():
            self.__platform._enter_shell_context()

    def is_active(self):
        return (self.__count > 0)


class Platform(object):
    __virtualizer = None

    _sc_cleanup = None
    _shell_context = None
    _is_open = None

    def __init__(self, virtualizer):
        super(Platform, self).__init__()

        self.__virtualizer = virtualizer

        self._sc_cleanup = []
        self._shell_context = ShellContext(self)
        self._is_open = False

    def _check_is_open(self):
        if not self._is_open:
            raise AssertionError('not open')

    def _open(self):
        if self._is_open:
            raise AssertionError('already open')
        self._is_open = True

    def _close(self):
        self._check_is_open()

        if self._shell_context.is_active():
            # FIXME: Better exception?
            raise AssertionError('shell context is active')

        self._is_open = False

    def _suspend(self):
        self._shell_context.suspend()

    def _resume(self):
        self._shell_context.resume()

    def _enter_shell_context(self):
        if self._sc_cleanup:
            raise AssertionError(
              'entering context with existing cleanup actions',
            )

        sh = self._get_shell()

        self._sc_add_cleanup(
          'umount -n /lib/modules/*/volatile || '
          'umount -nfl /lib/modules/*/volatile || '
          'true'
        )

        sh.execute('mount -nt proc proc /proc')
        self._sc_add_cleanup('umount -n /proc || umount -nfl /proc')

        if sh.eval('grep -q devtmpfs /proc/filesystems') == 0:
            sh.execute('mount -nt devtmpfs udev /dev')
            self._sc_add_cleanup('umount -n /dev || umount -nfl /dev')

        sh.execute('mount -nt devpts devpts /dev/pts')
        self._sc_add_cleanup('umount -n /dev/pts || umount -nfl /dev/pts')

        if sh.eval('[ -L /dev/shm ]') == 0:
            if sh.eval('[ -d "$(readlink -f /dev/shm)" ]') != 0:
                sh.execute('mkdir -p "$(readlink -f /dev/shm)"')
                self._sc_add_cleanup('rmdir "$(readlink -f /dev/shm)"')

            sh.execute('mount -nt tmpfs devshm "$(readlink -f /dev/shm)"')
            self._sc_add_cleanup(
              'umount -n "$(readlink -f /dev/shm)" '
              '|| umount -nfl "$(readlink -f /dev/shm)"'
            )

        elif sh.eval('[ -d /dev/shm ]') == 0:
            sh.execute('mount -nt tmpfs devshm /dev/shm')
            self._sc_add_cleanup('umount -n /dev/shm || umount -nfl /dev/shm')

        sh.env['HOME'] = '/root'

        if 'TERM' not in sh.env:
            sh.env['TERM'] = 'ansi'

        sh.execute('cd /')

        sh.env['LANG'] = 'C'
        sh.env['DEBIAN_FRONTEND'] = 'noninteractive'
        sh.env['DEBIAN_PRIORITY'] = 'critical'

        # FIXME: Do we want to set DEBCONF_FRONTEND?

        # Backup existing policy-rc.d.
        if sh.eval('[ -e /usr/sbin/policy-rc.d -o -L /usr/sbin/policy-rc.d ]') == 0:
            sh.execute('mv /usr/sbin/policy-rc.d /usr/sbin/policy-rc.d.bak')
            self._sc_add_cleanup('mv /usr/sbin/policy-rc.d.bak /usr/sbin/policy-rc.d')

        # Write a new policy-rc.d.
        sh.execute('echo "#!/bin/sh\nexit 101" > /usr/sbin/policy-rc.d')
        sh.execute('chmod a+x /usr/sbin/policy-rc.d')
        self._sc_add_cleanup('rm -f /usr/sbin/policy-rc.d')

        # Divert initctl to prevent upstart jobs from being started/stopped.
        # Using dpkg-divert is important because if upstart is upgraded, the
        # new version of initctl is installed at the diverted location.
        if sh.eval('[ -e /sbin/initctl ]') == 0:
            diverted = False

            if sh.eval('which dpkg-divert >/dev/null') == 0:
                try:
                    sh.execute(
                      'dpkg-divert --local --divert '
                      '/sbin/initctl.orig --rename /sbin/initctl'
                    )
                except CommandFailed:
                    pass
                else:
                    self._sc_add_cleanup(
                      'dpkg-divert --local --rename --remove /sbin/initctl',
                    )
                    diverted = True

            if not diverted:
                sh.execute('mv /sbin/initctl /sbin/initctl.orig')
                self._sc_add_cleanup('mv /sbin/initctl.orig /sbin/initctl')

        # Our version of initctl does nothing.
        content = '#!/bin/sh\nexit 101\n'
        sh.execute('echo ? >/sbin/initctl', content)
        self._sc_add_cleanup('rm -f /sbin/initctl')
        sh.execute('chmod a+x /sbin/initctl')
          
        content = '\n'.join([
          '#clear APT::Install-Recommends-Sections;',
          'APT::Install-Recommends 0;',
          'APT::Get::Assume-Yes 1;',
          'Acquire::Retries 5;',
          # Note: This option makes assumptions about the build proces simpler
          # (i.e. files can be overwritten without accidentally updating those
          # in the local APT archive, which would be at the other end of any
          # symlinks).
          'Acquire::Source-Symlinks false;',
          'Dpkg::Options {"--force-confnew";};',
        ])
        sh.execute('echo ? >/etc/apt/apt.conf.d/90rebar', content)
        self._sc_add_cleanup('rm -f /etc/apt/apt.conf.d/90rebar')

        if sh.eval('[ -e /etc/mtab -o -L /etc/mtab ]') == 0:
            sh.execute('mv /etc/mtab /etc/mtab.bak')
            self._sc_add_cleanup('mv /etc/mtab.bak /etc/mtab')

        sh.execute('touch /etc/mtab')
        self._sc_add_cleanup('rm -f /etc/mtab')

    def _exit_shell_context(self):
        self._sc_do_cleanup()

    def _sc_add_cleanup(self, *args):
        self._sc_cleanup.append(args)

    def _sc_do_cleanup(self):
        sh = self._get_shell()

        while self._sc_cleanup:
            # FIXME: Handle exceptions here.  Do we continue with other cleanup
            # tasks in that case?
            args = self._sc_cleanup.pop()
            sh.execute(*args)

    def _get_shell(self):
        return self.__virtualizer._get_shell()

    ###

    def open_shell(self):
        self._check_is_open()
        return self._shell_context

    def _virtualizer_copy_to(self, src, dst, recurse, dereference, replace):
        self.__virtualizer.copy_to(src, dst, recurse, dereference, replace)

    def copy_to(
      self,
      src,
      dst,
      recurse = False,
      dereference = False,
      replace = False,
      user = None,
      group = None,
    ):
        self._check_is_open()

        self._virtualizer_copy_to(src, dst, recurse, dereference, replace)

        sh = self._get_shell()

        if user:
            options = []
            if recurse:
                options.append('--recursive')

            if options:
                sh.execute(
                  'chown %s ? ?' % ' '.join(options),
                  user,
                  dst,
                )
            else:
                sh.execute('chown ? ?', user, dst)

        if group:
            options = []
            if recurse:
                options.append('--recursive')

            if options:
                sh.execute(
                  'chgrp %s ? ?' % ' '.join(options),
                  group,
                  dst,
                )
            else:
                sh.execute('chgrp ? ?', group, dst)

    def _virtualizer_copy_from(self, src, dst, recursive, dereference, root):
        self.__virtualizer.copy_from(src, dst, recursive, dereference, root)

    def copy_from(self, src, dst, recurse = False, dereference = False):
        self._check_is_open()
        self._virtualizer_copy_from(src, dst, recurse, dereference, '/')

    ###

    def create_disk_image(self, geometry):
        self._check_is_open()
        sh = self._get_shell()
        status, output = sh.execute('mktemp')
        filename = output.strip()
        disk_image = self.open_disk_image(filename, geometry)
        try:
            disk_image.create()
            return disk_image
        except:
            disk_image.close()
            disk_image.remove()
            raise

    def open_disk_image(self, filename, geometry):
        return DiskImage(self, filename, geometry)

    def open_as_temp_file(self, path):
        f = tempfile.NamedTemporaryFile('rb')
        try:
            self.copy_from(path, f.name)
            return f
        except:
            f.close()
            raise


class BasePlatform(Platform):
    def export_filesystem(self, directory):
        with self.open_shell() as sh:
            sh.execute(
              (
                r'''find /platform -mindepth 1 -maxdepth 1 '''
                r'''-exec cp --archive --one-file-system '{}' ? \;'''
              ),
              directory,
            )


class ChrootPlatform(Platform):
    def _open(self):
        self._enter_chroot()
        super(ChrootPlatform, self)._open()

    def _close(self):
        super(ChrootPlatform, self)._close()
        self._exit_chroot()

    def _suspend(self):
        super(ChrootPlatform, self)._suspend()
        self._exit_chroot()

    def _resume(self):
        self._enter_chroot()
        super(ChrootPlatform, self)._resume()

    def _enter_chroot(self):
        sh = self._get_shell()
        sh.execute('chroot /platform /bin/sh')
        sh.setup_shell()

    def _exit_chroot(self):
        sh = self._get_shell()
        # FIXME: Check shell pid to make sure we are exiting the right shell.
        sh.execute('exit')

    ###

    def _virtualizer_copy_to(self, src, dst, recurse, dereference, replace):
        dst = join_abs('/platform', dst)
        super(ChrootPlatform, self)._virtualizer_copy_to(
          src,
          dst,
          recurse,
          dereference,
          replace,
        )

    def _virtualizer_copy_from(self, src, dst, recurse, dereference, root):
        root = join_abs('/platform', root)
        super(ChrootPlatform, self)._virtualizer_copy_from(
          src,
          dst,
          recurse,
          dereference,
          root,
        )
