# Copyright (c) 2010-2012 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, re

from rebar import aptutil
from rebar.exceptions import InvalidSourcePackageError


def _parse_files(s):
    for l in s.split('\n'):
        md5sum, size, filename = l.split()
        # This is for basic validation only at this point.
        size = int(size)
        yield filename


class SourcePackage(object):
    name = None
    version = None

    regex_epoch = re.compile(r'^(?:([1-9][0-9]*):)?(.*$)')

    def build_source(self, platform, dstdir):
        raise NotImplementedError()

    def get_upstream_version(self):
        return aptutil.upstream_version(self.version)

    def get_version_without_epoch(self):
        return self.regex_epoch.match(self.version).group(2)


class DscSourcePackage(SourcePackage):
    parent = None
    filename = None

    _files = None

    def __init__(self, filename):
        self.filename = os.path.abspath(filename)
        self.parent = os.path.dirname(self.filename)

        self._files = []

        self._read_dsc_file()

    def build_source(self, platform, dstdir):
        # Nothing to build, just copy files.
        files_to_copy = [os.path.basename(self.filename)]
        files_to_copy.extend(self._files)
        files = []
        for filename in files_to_copy:
            full_dst = os.path.join(dstdir, filename)
            platform.copy_to(os.path.join(self.parent, filename), full_dst)
            files.append(full_dst)
        return files

    def _read_dsc_file(self):
        with open(self.filename, 'rb') as f:
            tagfile = aptutil.parse_tag_file(f)

            while tagfile.Step() == 1:
                if tagfile.Section.has_key('Source'):
                    if self.name is not None:
                        raise InvalidSourcePackageError(
                          '%s contains more than one Source: line'
                          % self.filename
                        )

                    self.name = tagfile.Section['Source']
                    self.version = tagfile.Section['Version']
                    self._files.extend(_parse_files(tagfile.Section['Files']))


class UnpackedSourcePackage(SourcePackage):
    parent = None
    dirname = None

    def __init__(self, dirname):
        self.dirname = os.path.abspath(dirname.rstrip('/'))
        self.parent = os.path.dirname(self.dirname)

        self._read_control()
        self._read_changelog()

    def build_source(self, platform, dst):
        files = []

        with platform.open_shell() as sh:
            status, output = sh.execute('mktemp -d')
            tmp_path = output.strip()
            try:
                source_dir = '%s-%s' % (
                  self.name,
                  self.get_upstream_version(),
                )
                platform.copy_to(
                  self.dirname,
                  os.path.join(tmp_path, source_dir),
                  recurse = True,
                )

                orig_archive_prefix = '%s_%s.orig.tar' % (
                  self.name,
                  self.get_upstream_version(),
                )
                for ext in ('gz', 'bz2'):
                    basename = '%s.%s' % (orig_archive_prefix, ext)
                    orig_archive = os.path.join(self.parent, basename)
                    if os.path.exists(orig_archive):
                        platform.copy_to(
                          orig_archive,
                          os.path.join(tmp_path, basename),
                        )
                        break

                sh.pushd(tmp_path)
                try:
                    # Note: We use follow_lines so that output is logged
                    # continously rather than buffered and logged after the
                    # command terminates.
                    for line in sh.follow_lines('dpkg-source -b ?', source_dir):
                        pass
                finally:
                    sh.popd()

                dsc_filename = '%s_%s.dsc' % (
                  self.name,
                  self.get_version_without_epoch(),
                )
                dsc_path = os.path.join(tmp_path, dsc_filename)
                source_files = [dsc_filename]
                with platform.open_as_temp_file(dsc_path) as f:
                    tagfile = aptutil.parse_tag_file(f)
                    while tagfile.Step() == 1:
                        if tagfile.Section.has_key('Source'):
                            source_files.extend(
                              _parse_files(tagfile.Section['Files']),
                            )

                if not source_files:
                    raise AssertionError(source_files)

                for source_file in source_files:
                    source_path = os.path.join(tmp_path, source_file)
                    full_dst = os.path.join(dst, source_file)
                    sh.execute('cp --force --link ? ?', source_path, full_dst)
                    files.append(full_dst)

            finally:
                sh.execute('rm -fR ?', tmp_path)

        return files

    def _read_control(self):
        filename = os.path.join(self.dirname, 'debian', 'control')

        with open(filename, 'rb') as f:
            tagfile = aptutil.parse_tag_file(f)
            while tagfile.Step() == 1:
                if tagfile.Section.has_key('Source'):
                    if self.name is not None:
                        raise InvalidSourcePackageError(
                          '%s contains more than one Source: line' % filename
                        )

                    self.name = tagfile.Section['Source']

    def _read_changelog(self):
        # FIXME: This is pretty naive.
        f = open(os.path.join(self.dirname, 'debian', 'changelog'))
        try:
            line = f.readline()
        finally:
            f.close()
        start = line.index('(') + 1
        end = line.index(')')
        self.version = line[start:end]
