# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from __future__ import with_statement 

import os

from pshell import CommandFailed, CommandSignalled
from fdisk.mbr import TYPES, is_not_empty

from rebar.buildstep import SimpleBuildStep, ComplexBuildStep
from rebar.util import (
  generator,
  fail_on,
  locate_data_file,
  locate_script,
)
from rebar.environment import (
  DISK,
  GZIMG,

  GRUB,
  GRUB2,
  SYSLINUX,

  MBR,
  VBR,

  EXT2,
  EXT3,
  EXT4,
  VFAT,
  MSDOS,

  GZIP,
)
from rebar.exceptions import DiskImageError


class Target(object):
    target_path = None

    def __init__(self, env, target_path, **kwargs):
        super(Target, self).__init__(env, **kwargs)
        self.target_path = target_path


class SimpleTarget(Target, SimpleBuildStep):
    pass


class ComplexTarget(Target, ComplexBuildStep):
    pass


class DirectoryTarget(SimpleTarget):
    @fail_on(CommandFailed, CommandSignalled)
    @generator
    def do_run(self):
        self.user_shell.execute('mkdir -p ?', self.target_path)
        self.populate()

    def populate(self):
        raise NotImplementedError()


class ImageTarget(SimpleTarget):
    geometry = None
    filesystem = None
    label = None
    bootloader = None
    fsexec = None
    type = None
    compression = None

    def __init__(
      self,
      env,
      target_path,
      geometry,
      filesystem,
      label,
      bootloader,
      fsexec,
      type,
      compression,
      **kwargs
    ):
        self.geometry = geometry
        self.filesystem = filesystem
        self.label = label
        self.bootloader = bootloader
        self.fsexec = fsexec
        self.type = type
        self.compression = compression

        super(ImageTarget, self).__init__(env, target_path, **kwargs)

    def get_required_packages(self):
        required_packages = super(
          ImageTarget,
          self,
        ).get_required_packages()

        # Needed for sfdisk -- should be installed anyway.
        required_packages.add('util-linux')

        required_packages.add('gzip')

        # sed is used by gzimg_build
        required_packages.add('sed')

        if self.filesystem in (EXT2, EXT3, EXT4):
            required_packages.add('e2fsprogs')
        if self.filesystem in (MSDOS, VFAT):
            required_packages.add('dosfstools')

        # Needed for mbr.bin, and also if boot loader is syslinux:
        required_packages.add('syslinux')

        if self.bootloader == GRUB:
            required_packages.add('grub')
        elif self.bootloader == GRUB2:
            # FIXME: This will fail on architectures other than i386/amd64.
            required_packages.add('grub-pc')
        elif self.bootloader == SYSLINUX:
            # syslinux does not depend on mtools, but mcopy is required to
            # install syslinux in most (all?) cases.
            # See https://bugs.launchpad.net/ubuntu/+source/syslinux/+bug/75765
            required_packages.add('mtools')

        return required_packages

    @fail_on(CommandFailed, CommandSignalled, DiskImageError)
    @generator
    def do_run(self):
        self.output.print_verbose('Building disk image...')
        self.build_image()

    ###

    def build_image(self):
        raise NotImplementedError()

    def post_process_image(self, disk_image):
        if self.type == GZIMG:
            self.output.print_verbose('Converting to gzimg...')

            with disk_image.platform.open_shell() as sh:
                status, output = sh.execute('mktemp')
                tmp_path = output.strip()
                try:
                    disk_image.platform.copy_to(
                      locate_script('gzimg_build'),
                      tmp_path,
                      dereference = True,
                    )
                    sh.execute(
                      'mv ? ?.in',
                      disk_image.filename,
                      disk_image.filename,
                    )
                    sh.execute(
                      '? ?.in ? && rm ?.in',
                      tmp_path,
                      disk_image.filename,
                      disk_image.filename,
                      disk_image.filename,
                    )
                finally:
                    sh.execute('rm -f ?', tmp_path)

    def compress_image(self, disk_image):
        if self.compression is None:
            return

        self.output.print_verbose('Compressing image..')

        if self.compression == GZIP:
            with disk_image.platform.open_shell() as sh:
                sh.execute(
                  'mv ? ?.in && gzip <?.in >? && rm -f ?.in',
                  disk_image.filename,
                  disk_image.filename,
                  disk_image.filename,
                  disk_image.filename,
                  disk_image.filename,
                )
        else:
            raise AssertionError()

    ###

    def get_geometry(self, size):
        if self.geometry is not None:
            return self.geometry

        heads = 255
        sectors_per_track = 63
        bytes_per_cylinder = heads * sectors_per_track * 512
        cylinders = (size / bytes_per_cylinder) + 1
        # FIXME: Can we get away with less?  This is pretty fuzzy, and it
        # depends on the filesystem and filesystem options.
        # Increase size by 50% to give us some head room.
        cylinders = long(1.5 * cylinders)
        return cylinders, heads, sectors_per_track

    def apply_fsexec(self, partition_index, disk_image):
        if self.fsexec:
            with disk_image.open_partition(partition_index) as partition:
                with disk_image.platform.open_shell() as sh:
                    for cmd in self.fsexec:
                        sh.execute(cmd, partition.device_path)

    def get_bootloader_name(self):
        if self.bootloader == GRUB:
            return 'grub'
        if self.bootloader == GRUB2:
            return 'grub2'
        if self.bootloader == SYSLINUX:
            return 'syslinux'
        raise ValueError(self.bootloader)

    def get_filesystem_name(self):
        if self.filesystem == EXT2:
            return 'ext2'
        if self.filesystem == EXT3:
            return 'ext3'
        if self.filesystem == EXT4:
            return 'ext4'
        if self.filesystem == VFAT:
            return 'vfat'
        if self.filesystem == MSDOS:
            return 'msdos'
        raise ValueError(self.filesystem)


class PlatformImageTarget(ImageTarget):
    devexec = None
    prepend_partitions = None

    def __init__(self, env, target_path, devexec, prepend_partitions, **kwargs):
        super(PlatformImageTarget, self).__init__(env, target_path, **kwargs)

        self.devexec = devexec
        self.prepend_partitions = prepend_partitions

    def build_image(self):
        if self.filesystem not in (EXT2, EXT3, EXT4):
            # FIXME: Better exception?
            raise AssertionError('unsupported filesystem: %r' % self.filesystem)

        if self.bootloader not in (GRUB, GRUB2):
            # FIXME: Better exception?
            raise AssertionError('unsupported bootloader: %r' % self.bootloader)

        if self.bootloader.location not in (MBR, VBR):
            # FIXME: Better exception?
            raise AssertionError(
              'unsupported bootloader location: %r' % self.bootloader.location
            )

        with self.virtualizer.open_base_platform() as platform:
            with platform.open_shell() as sh:
                status, output = sh.execute('du --bytes --summarize /platform')
                size = long(output.split()[0].strip())

        geometry = self.get_geometry(size)
        root_index = 0

        with self.virtualizer.open_base_platform() as platform:
            disk_image = platform.create_disk_image(geometry)
            with disk_image:
                try:
                    for partition_info in self.prepend_partitions:
                        disk_image.mbr_disk.add_partition(**partition_info)
                        root_index += 1

                    disk_image.mbr_disk.add_partition(
                      type = TYPES['Linux'],
                      bootable = (self.bootloader.location == VBR),
                    )
                    disk_image.install_mbr()
                    disk_image.create_partition_table()

                    with disk_image.open_partition(root_index) as partition:
                        partition.create_filesystem(
                          self.get_filesystem_name(),
                          label = self.label,
                        )
                        mount_point = partition.mount()
                        try:
                            platform.export_filesystem(mount_point)
                        finally:
                            partition.umount(mount_point)

                    if self.bootloader.location == VBR:
                        grub_index = root_index
                    else:
                        grub_index = None

                    if self.bootloader == GRUB:
                        disk_image.install_grub(
                          root_index = root_index,
                          grub_index = grub_index,
                        )
                    elif self.bootloader == GRUB2:
                        disk_image.install_grub2(
                          root_index = root_index,
                          grub_index = grub_index,
                        )

                    self.apply_fsexec(root_index, disk_image)
                    self.apply_devexec(root_index, disk_image)

                except:
                    # Must close before we can remove.
                    disk_image.close()
                    disk_image.remove()
                    raise

            self.post_process_image(disk_image)
            self.compress_image(disk_image)

            disk_image.save_copy(self.target_path)
            disk_image.remove()

    def apply_devexec(self, partition_index, disk_image):
        if self.devexec:
            partitions = []
            pt = disk_image.mbr_disk.boot_record.partition_table
            mbr_partitions = pt.get_partitions(is_not_empty)
            try:
                for index in range(len(mbr_partitions)):
                    partition = disk_image.open_partition(index)
                    partitions.append(partition)

                if index != partition_index:
                    raise AssertionError((index, partition_index))

                mount_point = partition.mount()
                try:
                    with disk_image.platform.open_shell() as sh:
                        # FIXME: The chroot enter/exit logic is largely
                        # duplicated from Platform.  We should find some way to
                        # share this code.
                        sh.execute('mount -nt proc proc ?/proc', mount_point)
                        try:
                            sh.execute(
                              'mount -n --bind /dev ?/dev',
                              mount_point,
                            )
                            try:
                                sh.execute('chroot ? /bin/sh', mount_point)
                                try:
                                    sh.setup_shell()
                                    sh.execute(
                                      'test -b ?',
                                      disk_image.device_path,
                                    )
                                    for cmd in self.devexec:
                                        sh.execute(cmd, disk_image.device_path)
                                finally:
                                    # FIXME: Check shell pid to make sure we
                                    # are exiting the right shell?
                                    sh.execute('exit')
                            finally:
                                sh.execute(
                                  'sync; umount -n ?/dev || umount -nfl ?/dev',
                                  mount_point,
                                  mount_point,
                                )
                        finally:
                            sh.execute(
                              'umount -n ?/proc || umount -nfl ?/proc',
                              mount_point,
                              mount_point,
                            )
                finally:
                    partition.umount(mount_point)
            finally:
                while partitions:
                    partitions.pop().close()


class DirImageTarget(ImageTarget):
    def __init__(
      self,
      env,
      target_path,
      source_dir,
      geometry,
      filesystem,
      label,
      bootloader,
      fsexec,
      type,
      compression,
      **kwargs
    ):
        self.source_dir = source_dir
        super(DirImageTarget, self).__init__(
          env,
          target_path,
          geometry,
          filesystem,
          label,
          bootloader,
          fsexec,
          type,
          compression,
          **kwargs
        )

    def build_image(self):
        if self.filesystem not in (MSDOS, VFAT, EXT2, EXT3, EXT4):
            # FIXME: Better exception?
            raise AssertionError('unsupported filesystem: %r' % self.filesystem)

        if self.bootloader not in (SYSLINUX, GRUB, GRUB2):
            # FIXME: Better exception?
            raise AssertionError('unsupported bootloader: %r' % self.bootloader)

        if self.bootloader.location not in (MBR, VBR):
            # FIXME: Better exception?
            raise AssertionError(
              'unsupported bootloader location: %r' % self.bootloader.location
            )

        status, output = self.user_shell.execute(
          'du --bytes --summarize ?',
          self.source_dir,
        )
        size = long(output.split()[0].strip())

        geometry = self.get_geometry(size)
        root_index = 0

        with self.virtualizer.open_base_platform() as platform:
            disk_image = platform.create_disk_image(geometry)
            with disk_image:
                try:
                    if self.filesystem == MSDOS:
                        ptype = TYPES['FAT16']
                    elif self.filesystem == VFAT:
                        ptype = TYPES['W95 FAT32']
                    elif self.filesystem in (EXT2, EXT3, EXT4):
                        ptype = TYPES['Linux']
                    else:
                        raise AssertionError()

                    disk_image.mbr_disk.add_partition(
                      type = ptype,
                      bootable = (self.bootloader.location == VBR),
                    )
                    disk_image.install_mbr()
                    disk_image.create_partition_table()

                    with disk_image.open_partition(root_index) as partition:
                        partition.create_filesystem(
                          self.get_filesystem_name(),
                          label = self.label,
                        )

                        # SYSLINUX is always installed in the VBR.
                        if self.bootloader == SYSLINUX:
                            partition.install_syslinux()

                        mount_point = partition.mount()
                        try:
                            for entry in os.listdir(self.source_dir):
                                src = os.path.join(self.source_dir, entry)
                                dst = os.path.join(mount_point, entry)
                                platform.copy_to(src, dst, recurse = True)
                        finally:
                            partition.umount(mount_point)

                    if self.bootloader.location == VBR:
                        grub_index = root_index
                    else:
                        grub_index = None

                    if self.bootloader == GRUB:
                        disk_image.install_grub(
                          root_index = root_index,
                          grub_index = grub_index,
                        )
                    elif self.bootloader == GRUB2:
                        disk_image.install_grub2(
                          root_index = root_index,
                          grub_index = grub_index,
                        )

                    self.apply_fsexec(root_index, disk_image)

                except:
                    # Must close before we can remove.
                    disk_image.close()
                    disk_image.remove()
                    raise

            self.post_process_image(disk_image)
            self.compress_image(disk_image)

            disk_image.save_copy(self.target_path)
            disk_image.remove()


class DiskImagingDirTarget(DirectoryTarget):
    source_image = None

    def __init__(
      self,
      env,
      target_path,
      source_image,
      **kwargs
    ):
        self.source_image = source_image
        super(DiskImagingDirTarget, self).__init__(env, target_path, **kwargs)

    def populate(self):
        image_filename = os.path.basename(self.source_image)

        # Build and install initramfs.
        self._build_initramfs(
          os.path.join(self.target_path, 'imaging.ird'),
        )
        # Install kernel.
        with self.virtualizer.open_platform() as platform:
            platform.copy_from(
              '/vmlinuz',
              os.path.join(self.target_path, 'imaging.vml'),
              dereference = True,
            )
        # Install source image.
        self.user_shell.execute(
          'cp --link --dereference ? ?',
          self.source_image,
          os.path.join(self.target_path, image_filename),
        )
        # Install pxelinux.cfg.
        with open(os.path.join(self.target_path, 'syslinux.cfg'), 'w') as f:
            f.write(
              (
                'DEFAULT imaging.vml initrd=imaging.ird ro quiet '
                 'image_filename=%s\n'
                'PROMPT 0\n'
                'NOESCAPE 1\n'
              ) % image_filename
            )

    def _build_initramfs(self, dst):
        with self.virtualizer.open_platform() as platform:
            platform.copy_to(
              locate_script('gzimg_write'),
              '/etc/initramfs-tools/scripts/gzimg_write',
              dereference = True,
            )
            platform.copy_to(
              locate_data_file('imaging/disk_imaging.script', ('indep', 'pkg')),
              '/etc/initramfs-tools/scripts/disk_imaging',
            )
            platform.copy_to(
              locate_data_file('imaging/imaging_functions', ('indep', 'pkg')),
              '/etc/initramfs-tools/scripts/imaging_functions',
            )
            platform.copy_to(
              locate_data_file('imaging/disk_imaging.hook', ('indep', 'pkg')),
              '/etc/initramfs-tools/hooks/disk_imaging',
            )
            with platform.open_shell() as sh:
                sh.execute(
                  'sed -i.bak '
                  "'s/^BOOT=.*$/BOOT=disk_imaging/' "
                  '/etc/initramfs-tools/initramfs.conf'
                )
                sh.execute('update-initramfs -u')

                platform.copy_from(
                  '/initrd.img',
                  dst,
                  dereference = True,
                )

                sh.execute(
                  'cat /etc/initramfs-tools/initramfs.conf.bak '
                  '>/etc/initramfs-tools/initramfs.conf'
                )
                sh.execute('rm /etc/initramfs-tools/initramfs.conf.bak')
                sh.execute('rm /etc/initramfs-tools/scripts/imaging_functions')
                sh.execute('rm /etc/initramfs-tools/scripts/disk_imaging')
                sh.execute('rm /etc/initramfs-tools/hooks/disk_imaging')
                sh.execute('update-initramfs -u')


class DiskImagingImageTarget(ImageTarget):
    source_image = None

    def __init__(
      self,
      env,
      target_path,
      source_image,
      geometry,
      filesystem,
      label,
      bootloader,
      fsexec,
      type,
      compression,
      **kwargs
    ):
        self.source_image = source_image
        super(DiskImagingImageTarget, self).__init__(
          env,
          target_path = target_path,
          geometry = geometry,
          filesystem = filesystem,
          label = label,
          bootloader = bootloader,
          fsexec = fsexec,
          type = type,
          compression = compression,
          **kwargs
        )

    def build_image(self):
        if self.filesystem not in (MSDOS, VFAT):
            # FIXME: Better exception?
            raise AssertionError('unsupported filesystem: %r' % self.filesystem)

        if self.bootloader != SYSLINUX:
            # FIXME: Better exception?
            raise AssertionError('unsupported bootloader: %r' % self.bootloader)

        if self.bootloader.location != VBR:
            # FIXME: Better exception?
            raise AssertionError(
              'unsupported bootloader location: %r' % self.bootloader.location
            )

        status, output = self.user_shell.execute('mktemp -d')
        tmpdir = output.strip()
        try:
            # FIXME: It's unfortunate that we have to pass a name argument
            # here, but that is how Targets currently need to be handled.
            # It would be nice to simplify this a bit.
            dir_ta = DiskImagingDirTarget(
              self.env,
              tmpdir,
              name = 'DiskImagingDirTarget',
              source_image = self.source_image,
            )
            dir_ta.bind(self.virtualizer, self.user_shell, self.output)
            try:
                dir_ta.populate()
            finally:
                dir_ta.unbind()

            status, output = self.user_shell.execute(
              'du --bytes --summarize ?',
              tmpdir,
            )
            size = long(output.strip().split()[0])

            geometry = self.get_geometry(size)
            root_index = 0

            with self.virtualizer.open_base_platform() as platform:
                disk_image = platform.create_disk_image(geometry)
                with disk_image:
                    try:
                        if self.filesystem == MSDOS:
                            ptype = TYPES['FAT16']
                        elif self.filesystem == VFAT:
                            ptype = TYPES['W95 FAT32']
                        else:
                            raise AssertionError()

                        # syslinux is always in the VBR
                        bootable = (self.bootloader.location == VBR)
                        if not bootable:
                            raise AssertionError(self.bootloader.location)

                        disk_image.mbr_disk.add_partition(
                          type = ptype,
                          bootable = bootable,
                        )
                        disk_image.install_mbr()
                        disk_image.create_partition_table()

                        with disk_image.open_partition(root_index) as partition:
                            partition.create_filesystem(
                              self.get_filesystem_name(),
                            )
                            partition.install_syslinux()

                            mount_point = partition.mount()
                            try:
                                for entry in self.user_shell.listdir(tmpdir):
                                    platform.copy_to(
                                      os.path.join(tmpdir, entry),
                                      os.path.join(mount_point, entry),
                                      recurse = True,
                                    )
                            finally:
                                partition.umount(mount_point)

                        self.apply_fsexec(root_index, disk_image)

                    except:
                        # Must close before we can remove.
                        disk_image.close()
                        disk_image.remove()
                        raise

                self.post_process_image(disk_image)
                self.compress_image(disk_image)

                disk_image.save_copy(self.target_path)
                disk_image.remove()

        finally:
            self.user_shell.execute('rm -fR ?', tmpdir)


class NetworkImagingDirTarget(DirectoryTarget):
    source_image = None
    image_url = None

    def __init__(self, env, target_path, source_image, image_url, **kwargs):
        self.source_image = source_image
        self.image_url = image_url
        super(NetworkImagingDirTarget, self).__init__(
          env,
          target_path,
          **kwargs
        )

    def populate(self):
        # FIXME: This is an abuse of os.path.basename.  We only support *nix
        # systems, so it should work, but it is still not correct.
        image_filename = os.path.basename(self.image_url)

        # Build and install initramfs.
        self._build_initramfs(
          os.path.join(self.target_path, 'imaging.ird'),
        )
        # Install kernel.
        with self.virtualizer.open_platform() as platform:
            platform.copy_from(
              '/vmlinuz',
              os.path.join(self.target_path, 'imaging.vml'),
              dereference = True,
            )
        # Install source image.
        self.user_shell.execute(
          'cp --link --dereference ? ?',
          self.source_image,
          os.path.join(self.target_path, image_filename),
        )
        # Install pxelinux.cfg.
        with open(os.path.join(self.target_path, 'pxelinux.cfg'), 'w') as f:
            f.write(
              'DEFAULT imaging.vml initrd=imaging.ird ro quiet image_url=%s\n'
              % self.image_url
            )

    def _build_initramfs(self, dst):
        with self.virtualizer.open_platform() as platform:
            platform.copy_to(
              locate_script('gzimg_write'),
              '/etc/initramfs-tools/scripts/gzimg_write',
              dereference = True,
            )
            platform.copy_to(
              locate_data_file('imaging/network_imaging.script', ('indep', 'pkg')),
              '/etc/initramfs-tools/scripts/network_imaging',
            )
            platform.copy_to(
              locate_data_file('imaging/imaging_functions', ('indep', 'pkg')),
              '/etc/initramfs-tools/scripts/imaging_functions',
            )
            platform.copy_to(
              locate_data_file('imaging/network_imaging.hook', ('indep', 'pkg')),
              '/etc/initramfs-tools/hooks/network_imaging',
            )
            with platform.open_shell() as sh:
                sh.execute(
                  "sed -i.bak 's/^BOOT=.*$/BOOT=network_imaging/' "
                  '/etc/initramfs-tools/initramfs.conf'
                )
                sh.execute('update-initramfs -u')

                platform.copy_from('/initrd.img', dst, dereference = True)

                sh.execute(
                  'cat /etc/initramfs-tools/initramfs.conf.bak '
                  '>/etc/initramfs-tools/initramfs.conf'
                )
                sh.execute('rm /etc/initramfs-tools/initramfs.conf.bak')
                sh.execute(
                  'rm /etc/initramfs-tools/scripts/imaging_functions')
                sh.execute('rm /etc/initramfs-tools/scripts/network_imaging')
                sh.execute('rm /etc/initramfs-tools/hooks/network_imaging')
                sh.execute('update-initramfs -u')


class ActionSequenceTarget(ComplexTarget):
    actions = None

    def __init__(self, env, target_path, actions, **kwargs):
        super(ActionSequenceTarget, self).__init__(
          env,
          target_path,
          **kwargs
        )
        for action in actions:
            self.add_subtask(action)
