# Copyright (c) 2009-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, sys

from sclapp2.encodings import get_user_encoding

try:
    from functools import wraps
except ImportError:
    from rebar.legacy_support import wraps


def generator(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        retval = fn(*args, **kwargs)
        if retval is None:
            pass
        elif hasattr(retval, 'next'):
            for x in retval:
                yield x
        else:
            raise TypeError('Function returned non-iterator %r.' % retval)
    return wrapped


def fail_on(*exc_types):
    def decorator(fn):
        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            try:
                for task_changes in fn(self, *args, **kwargs):
                    yield task_changes
            except exc_types, e:
                self.fail(unicode(e))
                return
        return wrapped
    return decorator


def mixin_instance(instance, mixin_cls):
    instance_cls = type(instance)
    d = dict(instance_cls.__dict__)
    # mixin_cls.__dict__ should be empty for __class__ assignment?
    d.update(mixin_cls.__dict__)
    new_cls = type(instance_cls.__name__, (mixin_cls, instance_cls), d)
    instance.__class__ = new_cls


class StrFromUnicodeMixin(object):
    def __str__(self):
        return unicode(self).encode(get_user_encoding())


def print_rebar_file_traceback():
    import traceback

    exc_type, exc_value, exc_traceback = sys.exc_info()

    try:
        original_stack = traceback.extract_tb(exc_traceback)
    finally:
        # Avoid a circular reference (see the Python Standard Library
        # documentation).
        del exc_traceback

    modified_stack = []

    for stack_entry in original_stack:
        if stack_entry[2] == '<module>':
            modified_stack.append(stack_entry)

    sys.stderr.write('Error in Rebar file:\n')
    sys.stderr.write('\n')
    if modified_stack:
        for line in traceback.format_list(modified_stack):
            sys.stderr.write(line)
        sys.stderr.write('\n')

    for line in traceback.format_exception_only(exc_type, exc_value):
        sys.stderr.write(line)


def locate_data_file(path, flags = ()):
    from rebar import distinfo

    if 'arch' in flags:
        if 'pkg' in flags:
            base_path = distinfo.install_data_arch_pkg
        else:
            base_path = distinfo.install_data_arch
    elif 'indep' in flags:
        if 'pkg' in flags:
            base_path = distinfo.install_data_indep_pkg
        else:
            base_path = distinfo.install_data_indep
    else:
        base_path = distinfo.install_data

    return os.path.join(base_path, path)


def locate_script(name):
    from rebar import distinfo

    if hasattr(distinfo, 'install_scripts'):
        return os.path.join(distinfo.install_scripts, name)

    if os.environ.get('PATH', None):
        for path in os.environ['PATH'].split(':'):
            p = os.path.join(path, name)
            if os.path.exists(p):
                return p

    raise AssertionError('unable to find script %r' % name)


def read_iter_chunks(f, size, chunk_size):
    nbytes = 0
    while True:
        remaining = size - nbytes
        if remaining < chunk_size:
            read_size = remaining
        else:
            read_size = chunk_size

        chunk = f.read(read_size)

        if len(chunk) != read_size:
            raise AssertionError()

        nbytes += read_size

        yield chunk

        if len(chunk) < read_size:
            break

        if nbytes >= size:
            break


def join_abs(*parts):
    parts = list(parts)
    for index in range(1, len(parts)):
        parts[index] = parts[index].lstrip('/')
    return os.path.join(*parts)
