# Copyright (c) 2010-2012 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, stat, time, socket, errno, random, string
from contextlib import contextmanager

from pshell import CommandFailed

from rebar.platform import (
  PlatformContext,
  BasePlatform,
  ChrootPlatform,
)
from rebar.util import read_iter_chunks, join_abs


# Note: Chunk size must be multiple of 4.
CHUNK_SIZE = 512

TO = object()
FROM = object()


class BaseVirtualizer(object):
    __base_platform_context = None
    __platform_context = None

    @classmethod
    def Initialized(
      cls,
      details,
      env,
      command_log = None,
      output_log = None,
      interactive = False,
    ):
        raise NotImplementedError()

    @classmethod
    def Create(
      cls,
      details,
      args,
      command_log = None,
      output_log = None,
      interactive = False,
    ):
        raise NotImplementedError()

    def __init__(self):
        super(BaseVirtualizer, self).__init__()

        self.__base_platform_context = self._init_base_platform_context()
        self.__platform_context = self._init_platform_context()

    def _init_base_platform_context(self):
        return PlatformContext(BasePlatform(self))

    def _init_platform_context(self):
        return PlatformContext(ChrootPlatform(self))

    def start(self):
        pass

    # FIXME: Should we allow the virtualizer to be stopped if the platform is
    # still open?
    def stop(self):
        pass

    def is_started(self):
        raise NotImplementedError()

    def _get_shell(self):
        raise NotImplementedError()

    ###

    def open_base_platform(self):
        if self.__platform_context.is_active():
            # FIXME: Better exception?
            raise AssertionError('platform context is active')
        return self.__base_platform_context

    def open_platform(self):
        if self.__base_platform_context.is_active():
            # FIXME: Better exception?
            raise AssertionError('base platform context is active')
        return self.__platform_context

    ###

    def get_ipv4_address(self):
        raise NotImplementedError()

    def _generate_transfer_key(self):
        return ''.join([random.choice(string.lowercase) for x in range(8)])

    @contextmanager
    def _open_netcat(self, direction, path):
        host = self.get_ipv4_address()

        sh = self._get_shell()
        sh.execute('test -x /bin/nc.openbsd')

        success = False
        # FIXME: Should we search randomly?
        for port in xrange(1024, 2**16):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                try:
                    sock.bind((host, port))
                    sock.settimeout(60)
                    sock.listen(1)
                except socket.error, e:
                    if e.errno != errno.EADDRINUSE:
                        raise
                    continue
                else:
                    success = True
                    break
            except:
                sock.close()
                raise
        if not success:
            raise AssertionError()

        try:
            key = self._generate_transfer_key()
            # Note: Key will be written to stdin.
            parts = []
            args = []
            if direction == FROM:
                # Concatenate key and file data and send the result back to
                # rebar.
                parts.append('cat - ? | ')
                args.append(path)
            parts.append('/bin/nc.openbsd -w 60 -n ? ?')
            args.extend([host, port])
            if direction == TO:
                # Incoming data is written to path.  Only the key data is sent
                # back to rebar.
                parts.append(' > ?')
                args.append(path)
            cmd = ''.join(parts)

            sh.execute(cmd, wait = False, *args)
            try:
                sh.write(key)
                sh.sendeof()
                if direction == FROM:
                    # FIXME: This is coupled to cat's stdio handling.  Is there
                    # a better way?
                    # EOF does not follow a newline, so we have to send it
                    # twice for cat to notice.
                    sh.sendeof()
                while True:
                    conn, addr = sock.accept()
                    try:
                        if conn.recv(len(key)) == key:
                            yield conn
                            break
                    finally:
                        conn.close()
            except:
                # FIXME: We should probably log a traceback right away so we
                # don't lose the exception if something else goes wrong.
                # FIXME: If netcat is not hung we lose exit status information.
                sh.interrupt()
                raise
            else:
                output = sh.wait_for_prompt()
                try:
                    sh.check_last_status()
                except CommandFailed, e:
                    e.output = output
                    raise

        finally:
            sock.close()

    def _copy_to_copy_file_contents(self, src, dst):
        # FIXME: Can we preserve sparseness of files?
        with open(src, 'rb') as f:
            with self._open_netcat(TO, dst) as conn:
                while True:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    conn.send(chunk)

    def copy_to(self, src, dst, recurse, dereference, replace):
        if src.endswith('/'):
            # FIXME: Better exception?
            raise AssertionError('%r ends with a slash' % src)

        if dst.endswith('/'):
            # FIXME: Better exception?
            raise AssertionError('%r ends with a slash' % dst)

        self.__base_platform_context.suspend()
        try:
            self.__platform_context.suspend()
            try:
                self._copy_to(src, dst, recurse, dereference, replace)
            finally:
                self.__platform_context.resume()
        finally:
            self.__base_platform_context.resume()

    def _copy_to(self, src, dst, recurse, dereference, replace):
        sh = self._get_shell()
        stat_result = os.lstat(src)
        st_mode = stat_result.st_mode
        st_atime = stat_result.st_atime
        st_mtime = stat_result.st_mtime
        copy_attrs = False

        if stat.S_ISREG(st_mode):
            # Note: If dst is already a symlink, we'll end up copying file
            # content to the symlink target, not replacing the symlink with a
            # regular file.  This is actually consistent with cp's behavior, so
            # I'm leaving it.  It caught me by surprise, though.
            if replace:
                sh.execute('rm -fR ?', dst)
            self._copy_to_copy_file_contents(src, dst)
            copy_attrs = True

        elif stat.S_ISLNK(st_mode):
            if replace:
                sh.execute('rm -fR ?', dst)
            if dereference:
                real_src = os.path.realpath(src)
                if not os.path.exists(real_src):
                    # FIXME: Better exception.
                    raise AssertionError('%r does not exist' % real_src)
                self._copy_to(real_src, dst, recurse, dereference, replace)
            else:
                sh.execute('ln -s ? ?', os.readlink(src), dst)

        elif stat.S_ISDIR(st_mode):
            # FIXME: Should we raise an exception if we were asked to copy a
            # directory with recurse = False?
            if recurse:
                if replace:
                    # If the destination exists but it is not a directory,
                    # remove it.
                    try:
                        sh.execute('test ! -e ? || test -d ?', dst, dst)
                    except CommandFailed:
                        sh.execute('rm -f ?', dst)
                if replace:
                    sh.execute('mkdir -p ?', dst)
                else:
                    sh.execute('mkdir ?', dst)
                copy_attrs = True
                for entry in os.listdir(src):
                    self._copy_to(
                      os.path.join(src, entry),
                      os.path.join(dst, entry),
                      recurse,
                      dereference,
                      replace,
                    )

        else:
            # FIXME: Better exception?
            raise AssertionError(
              '%r is not a regular file, symlink, or directory' % src
            )

        if copy_attrs:
            sh.execute('chmod ? ?', oct(stat.S_IMODE(st_mode)), dst)
            sh.execute('touch -a -t ? ?', self._format_secs(st_atime), dst)
            sh.execute('touch -m -t ? ?', self._format_secs(st_mtime), dst)

    def _format_secs(self, secs):
        return time.strftime('%Y%m%d%H%M.%S', time.gmtime(secs))

    def _copy_from_copy_file_contents(self, src, dst):
        # FIXME: Can we preserve sparseness of files?
        with open(dst, 'wb') as f:
            with self._open_netcat(FROM, src) as conn:
                while True:
                    chunk = conn.recv(CHUNK_SIZE)
                    if not chunk:
                        break
                    f.write(chunk)

    # Note: We deliberately do not support a replace argument to copy_from.  It
    # is dangerous and I don't see a use case for it.

    def copy_from(self, src, dst, recurse, dereference, root):
        if src.endswith('/'):
            # FIXME: Better exception?
            raise AssertionError('%r ends with a slash' % src)

        if dst.endswith('/'):
            # FIXME: Better exception?
            raise AssertionError('%r ends with a slash' % dst)

        self.__base_platform_context.suspend()
        try:
            self.__platform_context.suspend()
            try:
                self._copy_from(src, dst, recurse, dereference, root)
            finally:
                self.__platform_context.resume()
        finally:
            self.__base_platform_context.resume()

    def _copy_from(self, src, dst, recurse, dereference, root):
        real_src = join_abs(root, src)

        sh = self._get_shell()
        status, output = sh.execute("stat --format '%f %X %Y' ?", real_src)
        _mode, _atime, _mtime = output.strip().split()
        st_mode = int(_mode, 16)
        st_atime = int(_atime, 10)
        st_mtime = int(_mtime, 10)
        copy_attrs = False

        if stat.S_ISREG(st_mode):
            self._copy_from_copy_file_contents(real_src, dst)
            copy_attrs = True

        elif stat.S_ISLNK(st_mode):
            if dereference:
                status, output = sh.execute('readlink ?', real_src)
                target = output.strip()
                if not target.startswith('/'):
                    target = os.path.join(os.path.dirname(src), target)
                try:
                    sh.execute('test -e ?', join_abs(root, target))
                except CommandFailed:
                    # FIXME: Better exception.
                    raise AssertionError('%r does not exist' % target)
                self._copy_from(target, dst, recurse, dereference, root)
            else:
                status, output = sh.execute('readlink ?', real_src)
                target = output.strip()
                os.symlink(target, dst)

        elif stat.S_ISDIR(st_mode):
            # FIXME: Should we raise an exception if we were asked to copy a
            # directory with recurse = False?
            if recurse:
                os.mkdir(dst)
                copy_attrs = True
                for entry in sh.listdir(real_src):
                    self._copy_from(
                      os.path.join(src, entry),
                      os.path.join(dst, entry),
                      recurse,
                      dereference,
                      root,
                    )

        else:
            # FIXME: Better exception?
            raise AssertionError(
              '%r is not a regular file, symlink, or directory' % real_src
            )

        if copy_attrs:
            os.chmod(dst, stat.S_IMODE(st_mode))
            os.utime(dst, (st_atime, st_mtime))
