# Copyright (c) 2010 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os
from contextlib import contextmanager, closing

from pshell import PasswordRequired

from rebar.shells import SudoShell
from rebar.virtualizer.base import BaseVirtualizer
from rebar.exceptions import VirtualizerInitializationError
from rebar.tasks_ import Task, ComplexTask
from rebar.build import Debootstrapper


INSIDE = 'INSIDE'
OUTSIDE = 'OUTSIDE'


class DebootstrapTask(Task):
    name = 'bootstrap base system'

    root_shell = None
    debootstrapper = None

    def __init__(
      self,
      root_shell,
      suite,
      target,
      mirror = None,
      script = None,
      include = None,
      exclude = None,
      architecture = None,
    ):
        self.root_shell = root_shell

        super(DebootstrapTask, self).__init__()

        self.debootstrapper = Debootstrapper(
          root_shell,
          suite = suite,
          target = target,
          mirror = mirror,
          script = script,
          include = include,
          exclude = exclude,
          architecture = architecture,
        )

    def do_run(self):
        for action, package, progress in self.debootstrapper.iter_run():
            task_changes = []
            if (action is not None) and (package is not None):
                task_changes.append(
                  self.change('status', '%s %s' % (action, package))
                )
            if (progress is not None) and (progress != self.progress):
                task_changes.append(self.change('progress', progress))
            yield task_changes


class ChrootVirtualizer(BaseVirtualizer):
    directory = None

    __sh = None
    __in_chroot = None

    def __init__(self, sh, directory):
        super(ChrootVirtualizer, self).__init__()

        self.directory = directory

        self.__sh = sh
        self.__in_chroot = False

    @classmethod
    @contextmanager
    def Initialized(
      cls,
      directory,
      env,
      command_log = None,
      output_log = None,
      interactive = False,
    ):
        password = None
        if interactive:
            virt_shell, password = SudoShell.InitializeInteractive(
              command_log = command_log,
              output_log = output_log,
            )
        else:
            try:
                virt_shell = SudoShell(
                  command_log = command_log,
                  output_log = output_log,
                )
            except PasswordRequired:
                raise VirtualizerInitializationError(
                  'interactive initialization or passwordless '
                  'sudo privileges required'
                )
        with closing(virt_shell):
            del password
            yield cls(virt_shell, directory)

    @classmethod
    @contextmanager
    def Create(
      cls,
      directory,
      args,
      command_log = None,
      output_log = None,
      interactive = False,
    ):
        # FIXME: The command-line interface for specifying debootstrap options
        # is not ideal.  We're trying to avoid requiring additional
        # command-line options.  Is there a better way?

        mirror = None
        script = None
        architecture = None

        len_args = len(args)
        if len_args == 0:
            raise VirtualizerInitializationError(
              'missing required argument suite'
            )
        elif len_args == 1:
            suite = args[0]
        elif len_args == 2:
            suite, mirror = args
        elif len_args == 3:
            suite, mirror, script = args
        elif len_args == 4:
            suite, mirror, script, architecture = args
        else:
            raise VirtualizerInitializationError('too many arguments')

        if mirror == '-':
            mirror = None

        if script == '-':
            script = None

        password = None
        if interactive:
            root_shell, password = SudoShell.InitializeInteractive(
              command_log = command_log,
              output_log = output_log,
            )
        else:
            try:
                root_shell = SudoShell(
                  command_log = command_log,
                  output_log = output_log,
                )
            except PasswordRequired:
                raise VirtualizerInitializationError(
                  'interactive initialization or passwordless '
                  'sudo privileges required'
                )
        with closing(root_shell):
            del password
            yield DebootstrapTask(
              root_shell,
              suite = suite,
              target = directory,
              mirror = mirror,
              script = script,
              architecture = architecture,
            )

    ###

    def start(self):
        self.__sh.execute('chroot ? ?', self.directory, '/bin/sh')
        self.__sh.setup_shell()
        self.__in_chroot = True

        super(ChrootVirtualizer, self).start()

    def stop(self):
        super(ChrootVirtualizer, self).stop()

        # FIXME: Check shell pid to make sure we are exiting the right shell.
        self.__sh.execute('exit 0')
        self.__in_chroot = False

    def is_started(self):
        return self.__in_chroot

    def _get_shell(self):
        if not self.__in_chroot:
            # FIXME: Better exception?
            raise AssertionError('not started')
        return self.__sh

    def get_ipv4_address(self):
        return '127.0.0.1'
