# Copyright (c) 2007-2010 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, tempfile

from unittest import TestCase

from rebar.environment import (
  RebarEnvironment,
  PLATFORM,
  VIRTUALIZER,
)
from rebar.actions import (
  ExecAction,
  ShellAction,
  CopyToAction,
  PatchAction,
)
from rebar.output import NullOutput

from manager import manager
from common import BuildStepTestCase


class TestExecAction(BuildStepTestCase):
    def test_using_platform(self):
        action = ExecAction(
          self.env,
          'echo foo',
          using = PLATFORM,
          name = 'Exec',
        )
        self.run_build_step(action)
        # FIXME: We have no way to differentiate between commands executed in
        # the platform and those executed directly in the virtualizer.
        self.assertVirtualizerCommand('echo foo', outfile = self.output.f_out)

    def test_using_virtualizer(self):
        action = ExecAction(
          self.env,
          'echo foo',
          using = VIRTUALIZER,
          name = 'Exec',
        )
        self.run_build_step(action)
        # FIXME: We have no way to differentiate between commands executed in
        # the platform and those executed directly in the virtualizer.
        self.assertVirtualizerCommand('echo foo', outfile = self.output.f_out)

manager.add_test_case_class(TestExecAction)


class TestShellAction(BuildStepTestCase):
    def test(self):
        action = ShellAction(
          self.env,
          'echo foo',
          name = 'Shell',
        )
        self.run_build_step(action)
        self.assertUserCommand('echo foo', outfile = self.output.f_out)

manager.add_test_case_class(TestShellAction)


class TestCopyToAction(BuildStepTestCase):
    def test_defaults(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          False,
          False,
          False,
          None,
          None,
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst')

    def test_recurse(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          True,
          False,
          False,
          None,
          None,
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst', recurse = True)

    def test_user(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          False,
          False,
          False,
          'foo',
          None,
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst')
        self.assertVirtualizerCommand('chown ? ?', 'foo', '/dst')

    def test_group(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          False,
          False,
          False,
          None,
          'foo',
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst')
        self.assertVirtualizerCommand('chgrp ? ?', 'foo', '/dst')

    def test_user_group(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          False,
          False,
          False,
          'foo',
          'bar',
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst')
        self.assertVirtualizerCommand('chown ? ?', 'foo', '/dst')
        self.assertVirtualizerCommand('chgrp ? ?', 'bar', '/dst')

    def test_user_group_recurse(self):
        action = CopyToAction(
          self.env,
          'src',
          '/dst',
          True,
          False,
          False,
          'foo',
          'bar',
          name = 'CopyTo',
        )
        self.run_build_step(action)
        self.assertCopyToPlatform('src', '/dst', recurse = True)
        self.assertVirtualizerCommand(
          'chown --recursive ? ?',
          'foo',
          '/dst',
        )
        self.assertVirtualizerCommand(
          'chgrp --recursive ? ?',
          'bar',
          '/dst',
        )

manager.add_test_case_class(TestCopyToAction)


# FIXME: Test other options.

class TestPatchAction(BuildStepTestCase):
    def test_with_dir(self):
        tmp_fd, tmp_path = tempfile.mkstemp()
        try:
            os.write(tmp_fd, 'one\ntwo\nthree\n')
            os.close(tmp_fd)

            action = PatchAction(
              self.env,
              tmp_path,
              '/bar',
              None,
              0,
              name = 'Patch',
            )
            self.run_build_step(action)
            self.assertVirtualizerCommand('which patch >/dev/null')
            self.assertVirtualizerCommand('PUSHD', '/platform/bar')
            self.assertVirtualizerCommand(
              'cat - | patch --strip ?',
              0,
              wait = False,
            )
            self.assertVirtualizerCommand('WRITE', 'one\n')
            self.assertVirtualizerCommand('WRITE', 'two\n')
            self.assertVirtualizerCommand('WRITE', 'three\n')
            self.assertVirtualizerCommand('WRITE', '\n')
            self.assertVirtualizerCommand('SENDEOF')
            self.assertVirtualizerCommand('POPD')
        finally:
            try:
                os.close(tmp_fd)
            except OSError:
                pass
            os.unlink(tmp_path)

    def test_with_file(self):
        tmp_fd, tmp_path = tempfile.mkstemp()
        try:
            os.write(tmp_fd, 'one\ntwo\nthree\n')
            os.close(tmp_fd)

            action = PatchAction(
              self.env,
              tmp_path,
              None,
              '/bar',
              0,
              name = 'Patch',
            )
            self.run_build_step(action)
            self.assertVirtualizerCommand('which patch >/dev/null')
            self.assertVirtualizerCommand(
              'cat - | patch ?',
              '/platform/bar',
              wait = False,
            )
            self.assertVirtualizerCommand('WRITE', 'one\n')
            self.assertVirtualizerCommand('WRITE', 'two\n')
            self.assertVirtualizerCommand('WRITE', 'three\n')
            self.assertVirtualizerCommand('WRITE', '\n')
            self.assertVirtualizerCommand('SENDEOF')
        finally:
            try:
                os.close(tmp_fd)
            except OSError:
                pass
            os.unlink(tmp_path)

manager.add_test_case_class(TestPatchAction)
