# Copyright (c) 2007-2010 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, re
from unittest import TestCase

from pshell import shinterp

from rebar.environment import RebarEnvironment
from rebar.shells import RebarShellMixin
from rebar.virtualizer.base import BaseVirtualizer
from rebar.platform import (
  BasePlatform,
  ChrootPlatform,
  PlatformContext,
)
from rebar.output import NullOutput
from rebar.util import join_abs


class MockShell(RebarShellMixin):
    before = ''

    shell = None

    failure_exceptions = None
    signal_exceptions = None
    trace = None

    interpolate = staticmethod(shinterp.interpolate)

    env = None

    dir_cache = None

    shell_encoding = 'utf-8'

    always_group_cmd = None

    interactive_ps1 = '$ '
    interactive_ps2 = '> '
    interactive_ps3 = ''
    interactive_ps4 = '+ '

    interactive_exit_string = 'PSHELL_INTERACTIVE_SHELL_EXITING'

    _last_cmd_values = None


    exec_cache = None
    result_cache = None


    def __init__(self, exec_cache = None, result_cache = None):
        self.env = {}

        if exec_cache is None:
            exec_cache = []

        if result_cache is None:
            result_cache = {}

        self.exec_cache = exec_cache
        self.result_cache = result_cache

    def setup_shell(self):
        pass

    def detect_shell_encoding(self):
        return 'UTF-8'

    def wait_for_prompt(self, **kwargs):
        pass

    def interrupt(self, wait_for_prompt = True, timeout = -1):
        raise NotImplementedError()

    def execute(self, cmd, *values, **kwargs):
        wait = kwargs.get('wait', True)

        self._last_cmd_values = (cmd, values)
        self._execute(cmd, *values, **kwargs)

        if wait:
            return self.check_last_status()

    def _execute(self, cmd, *values, **kwargs):
        self.exec_cache.append((self, cmd, values, kwargs))

    def check_last_status(self):
        cmd, values = self._last_cmd_values
        try:
            return self.result_cache[(cmd, values)]
        except KeyError:
            return 0, ''

    def sendline(self, line):
        self.exec_cache.append((self, 'SENDLINE', (line,), {}))

    def sendeof(self):
        self.exec_cache.append((self, 'SENDEOF', (), {}))

    def write(self, s):
        self.exec_cache.append((self, 'WRITE', (s,), {}))

    def expect(self, *args, **kwargs):
        pass

    def encode(self, s):
        return s.encode('utf-8')

    def decode(self, s):
        return s.decode('utf-8')

    def follow(self, cmd, *values, **kwargs):
        raise NotImplementedError()

    def follow_callback(self, cmd, *values, **kwargs):
        raise NotImplementedError()

    def follow_write(self, cmd, *values, **kwargs):
        self.execute(cmd, *values, **kwargs)

    def follow_write_return(self, cmd, *values, **kwargs):
        raise NotImplementedError()

    def exit(self, timeout = -1):
        pass

    prompt = '$ '

    prompt_sep = property(lambda self: re.escape(self.prompt))

    def pwd(self):
        raise NotImplementedError()

    def pushd(self, dir):
        self.exec_cache.append((self, 'PUSHD', (dir,), {}))

    def popd(self):
        self.exec_cache.append((self, 'POPD', (), {}))

    def cd(self, dir):
        raise NotImplementedError()

    def fit_window(self):
        pass

    def interact(self, fitted = False):
        raise NotImplementedError()

    def interact_with(self, *args, **kwargs):
        raise NotImplementedError()


class MockVirtualizer(BaseVirtualizer):
    started = None
    shell = None
    copy_cache = None

    def __init__(self, shell = None, copy_cache = None):
        super(MockVirtualizer, self).__init__()

        if shell is None:
            shell = MockShell()

        if copy_cache is None:
            copy_cache = []

        self.started = False
        self.shell = shell
        self.copy_cache = copy_cache

    def _init_base_platform_context(self):
        return PlatformContext(MockBasePlatform(self))

    def _init_platform_context(self):
        return PlatformContext(MockChrootPlatform(self))

    def start(self):
        self.started = True

        super(MockVirtualizer, self).start()

    def stop(self):
        super(MockVirtualizer, self).stop()

        self.started = False

    def is_started(self):
        return self.started

    def _get_shell(self):
        return self.shell

    ###

    def copy_to(self, *args):
        self.copy_cache.append((self, 'COPY_TO', args, {}))

    def copy_from(self, *args):
        self.copy_cache.append((self, 'COPY_FROM', args, {}))


class MockPlatformMixin(object):
    def _enter_shell_context(self):
        pass

    def _exit_shell_context(self):
        pass


class MockBasePlatform(MockPlatformMixin, BasePlatform):
    pass


class MockChrootPlatform(MockPlatformMixin, ChrootPlatform):
    def _enter_chroot(self):
        pass

    def _exit_chroot(self):
        pass


class BuildStepTestCase(TestCase):
    exec_cache = None
    user_shell = None
    virtualizer = None
    env = None
    output = None

    def setUp(self):
        self.exec_cache = []
        self.user_shell = MockShell(exec_cache = self.exec_cache)
        self.virtualizer = MockVirtualizer(
          shell = MockShell(exec_cache = self.exec_cache),
          copy_cache = self.exec_cache,
        )

        self.env = RebarEnvironment()
        self.env.load_from_string(
          '''Repository('http://archive.ubuntu.com/ubuntu', 'hardy', ['main'])'''
        )

        self.output = NullOutput()

    def tearDown(self):
        self.assertEqual(self.exec_cache, [])

        del self.user_shell
        del self.virtualizer
        del self.env
        del self.output

    def assertShellCommand(self, sh, cmd, *values, **kwargs):
        self.assertEqual(self.exec_cache.pop(0), (sh, cmd, values, kwargs))

    def assertUserCommand(self, cmd, *values, **kwargs):
        self.assertShellCommand(self.user_shell, cmd, *values, **kwargs)

    def assertVirtualizerCommand(self, cmd, *values, **kwargs):
        self.assertShellCommand(self.virtualizer.shell, cmd, *values, **kwargs)

    def assertCopyTo(
      self,
      src,
      dst,
      recurse = False,
      dereference = False,
      replace = False,
    ):
        self.assertEqual(
          self.exec_cache.pop(0),
          (self.virtualizer, 'COPY_TO', (src, dst, recurse, dereference, replace), {}),
        )

    def assertCopyFrom(
      self,
      src,
      dst,
      recurse = False,
      dereference = False,
      root = '/',
    ):
        self.assertEqual(
          self.exec_cache.pop(0),
          (
            self.virtualizer,
            'COPY_FROM',
            (src, dst, recurse, dereference, root),
            {},
          ),
        )

    def assertCopyToPlatform(self, src, dst, **kwargs):
        dst = join_abs('/platform', dst)
        self.assertCopyTo(src, dst, **kwargs)

    def assertCopyFromPlatform(self, src, dst, **kwargs):
        root = kwargs.pop('root', '/')
        kwargs['root'] = join_abs('/platform', root)
        self.assertCopyFrom(src, dst, **kwargs)

    def run_build_step(self, build_step):
        self.virtualizer.start()
        try:
            build_step.bind(self.virtualizer, self.user_shell, self.output)
            try:
                for task_changes in build_step.run():
                    pass
            finally:
                build_step.unbind()
        finally:
            self.virtualizer.stop()
