# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from unittest import TestCase

from rebar.environment import (
  RebarEnvironment,
  PLATFORM,
  VIRTUALIZER,
)
from rebar.directives import (
  Platform,
  Bootstrap,
  Repository,
  Require,
  Install,
  BinaryDirs,
  SourceDirs,
  ExportDirs,

  PreBuild,
  PostBuild,
  PreInstall,
  PostInstall,
  PreRemove,
  PostRemove,
  PreShip,
  PostShip,

  Exec,
  Shell,
  CopyTo,
  CopyFrom,
  Patch,

  PlatformImage,
  DirImage,
  DiskImagingDir,
  DiskImagingImage,
  NetworkImagingDir,
  Target,
)
from rebar.actions import ExecAction
from rebar.exceptions import RebarSyntaxError, MissingRequiredArgumentError

from manager import manager


class TestPlatform(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(
          environment['platform'],
          {
            'architecture': None,
            'hostname': None,
          },
        )

    def test_apply(self):
        environment = RebarEnvironment()
        directive = Platform(architecture = 'foo', hostname = 'bar')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['platform'],
          {
            'architecture': 'foo',
            'hostname': 'bar',
          },
        )

    def test_apply_twice(self):
        environment = RebarEnvironment()
        directive = Platform()
        directive.apply(environment)
        directive.finalize(environment)
        self.assertRaises(RebarSyntaxError, directive.apply, environment)

manager.add_test_case_class(TestPlatform)


class TestBootstrap(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['bootstrap'], None)

    def test_apply_simple(self):
        environment = RebarEnvironment()
        directive = Bootstrap('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['bootstrap'],
          {
            'mirror': 'foo',
            'suite': 'bar',
            'script': 'baz',
            'include': (),
            'exclude': (),
          },
        )

    def test_apply_include_exclude(self):
        environment = RebarEnvironment()
        directive = Bootstrap(
          'foo',
          'bar',
          'baz',
          include = ['abc'],
          exclude = ['xyz'],
        )
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['bootstrap'],
          {
            'mirror': 'foo',
            'suite': 'bar',
            'script': 'baz',
            'include': ['abc'],
            'exclude': ['xyz'],
          },
        )

    def test_apply_twice(self):
        environment = RebarEnvironment()
        directive = Bootstrap('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertRaises(RebarSyntaxError, directive.apply, environment)

manager.add_test_case_class(TestBootstrap)


class TestRepository(TestCase):
    def test_without_architectures(self):
        environment = RebarEnvironment()
        directive = Repository('foo', 'bar', ['baz'])
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['repositories'],
          [{
            'mirror': 'foo',
            'suite': 'bar',
            'components': ['baz'],
            'architectures': None,
            'trusted': False,
          }],
        )

    def test_with_architectures(self):
        environment = RebarEnvironment()
        directive = Repository('foo', 'bar', ['baz'], ['arch1', 'arch2'])
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['repositories'],
          [{
            'mirror': 'foo',
            'suite': 'bar',
            'components': ['baz'],
            'architectures': ['arch1', 'arch2'],
            'trusted': False,
          }],
        )

manager.add_test_case_class(TestRepository)


class TestRequire(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['required_packages'], [])

    def test_apply(self):
        environment = RebarEnvironment()
        directive = Require('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['required_packages'],
          ['foo', 'bar', 'baz'],
        )

    def test_apply_twice(self):
        environment = RebarEnvironment()

        directive = Require('foo', 'bar')
        directive.apply(environment)
        directive.finalize(environment)

        directive = Require('baz', 'qux')
        directive.apply(environment)
        directive.finalize(environment)

        self.assertEqual(
          environment['required_packages'],
          ['foo', 'bar', 'baz', 'qux'],
        )

    def test_apply_with_regex(self):
        self.assertRaises(RebarSyntaxError, Require, 'foo.*')

manager.add_test_case_class(TestRequire)


class TestInstall(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['install_packages'], [])

    def test_apply(self):
        environment = RebarEnvironment()
        directive = Install('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(
          environment['install_packages'],
          ['foo', 'bar', 'baz'],
        )

    def test_apply_twice(self):
        environment = RebarEnvironment()

        directive = Install('foo', 'bar')
        directive.apply(environment)
        directive.finalize(environment)

        directive = Install('baz', 'qux')
        directive.apply(environment)
        directive.finalize(environment)

        self.assertEqual(
          environment['install_packages'],
          ['foo', 'bar', 'baz', 'qux'],
        )

    def test_apply_with_regex(self):
        self.assertRaises(RebarSyntaxError, Install, 'foo.*')

manager.add_test_case_class(TestInstall)


class TestBinaryDirs(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['binary_dirs'], [])

    def test_apply(self):
        environment = RebarEnvironment()
        directive = BinaryDirs('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(environment['binary_dirs'], ['foo', 'bar', 'baz'])

    def test_apply_twice(self):
        environment = RebarEnvironment()

        directive = BinaryDirs('foo', 'bar')
        directive.apply(environment)
        directive.finalize(environment)

        directive = BinaryDirs('baz', 'qux')
        directive.apply(environment)
        directive.finalize(environment)

        self.assertEqual(
          environment['binary_dirs'],
          ['foo', 'bar', 'baz', 'qux'],
        )

manager.add_test_case_class(TestBinaryDirs)


class TestSourceDirs(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['source_dirs'], [])

    def test_apply(self):
        environment = RebarEnvironment()
        directive = SourceDirs('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(environment['source_dirs'], ['foo', 'bar', 'baz'])

    def test_apply_twice(self):
        environment = RebarEnvironment()

        directive = SourceDirs('foo', 'bar')
        directive.apply(environment)
        directive.finalize(environment)

        directive = SourceDirs('baz', 'qux')
        directive.apply(environment)
        directive.finalize(environment)

        self.assertEqual(
          environment['source_dirs'],
          ['foo', 'bar', 'baz', 'qux'],
        )

manager.add_test_case_class(TestSourceDirs)


class TestExportDirs(TestCase):
    def test_defaults(self):
        environment = RebarEnvironment()
        self.assertEqual(environment['export_dirs'], [])

    def test_apply(self):
        environment = RebarEnvironment()
        directive = ExportDirs('foo', 'bar', 'baz')
        directive.apply(environment)
        directive.finalize(environment)
        self.assertEqual(environment['export_dirs'], ['foo', 'bar', 'baz'])

    def test_apply_twice(self):
        environment = RebarEnvironment()

        directive = ExportDirs('foo', 'bar')
        directive.apply(environment)
        directive.finalize(environment)

        directive = ExportDirs('baz', 'qux')
        directive.apply(environment)
        directive.finalize(environment)

        self.assertEqual(
          environment['export_dirs'],
          ['foo', 'bar', 'baz', 'qux'],
        )

manager.add_test_case_class(TestExportDirs)


###


class ActionGroupTestCase(TestCase):
    cls = None
    env_name = None

    def test_loaded_from_string(self):
        environment = RebarEnvironment()
        directive = Exec('echo foo')
        environment.load_from_string('%s(%r)' % (self.cls.__name__, directive))
        directives = environment[self.env_name]
        self.assertEqual(len(directives), 1)
        # Note: We can't just compare directive with directives[0] because
        # they have differing action_group_names.
        self.assertEqual(type(directives[0]), type(directive))
        self.assertEqual(directives[0].init_args, directive.init_args)

    def test_with_target_directive(self):
        self.assertRaises(RebarSyntaxError, self.cls, PlatformImage('foo'))


class TestPreBuild(ActionGroupTestCase):
    cls = PreBuild
    env_name = 'pre_build_actions'

manager.add_test_case_class(TestPreBuild)


class TestPostBuild(ActionGroupTestCase):
    cls = PostBuild
    env_name = 'post_build_actions'

manager.add_test_case_class(TestPostBuild)


class TestPreInstall(ActionGroupTestCase):
    cls = PreInstall
    env_name = 'pre_install_actions'

manager.add_test_case_class(TestPreInstall)


class TestPostInstall(ActionGroupTestCase):
    cls = PostInstall
    env_name = 'post_install_actions'

manager.add_test_case_class(TestPostInstall)


class TestPreRemove(ActionGroupTestCase):
    cls = PreRemove
    env_name = 'pre_remove_actions'

manager.add_test_case_class(TestPreRemove)


class TestPostRemove(ActionGroupTestCase):
    cls = PostRemove
    env_name = 'post_remove_actions'

manager.add_test_case_class(TestPostRemove)


class TestPreShip(ActionGroupTestCase):
    cls = PreShip
    env_name = 'pre_ship_actions'

manager.add_test_case_class(TestPreShip)


class TestPostShip(ActionGroupTestCase):
    cls = PostShip
    env_name = 'post_ship_actions'

manager.add_test_case_class(TestPostShip)


class ActionDirectiveTestCase(TestCase):
    default_action_group_name = 'post_remove_actions'

    def test_loaded_from_string(self):
        environment = RebarEnvironment()
        directive = self.get_directive()
        environment.load_from_string(repr(directive))
        directives = environment[self.default_action_group_name]
        self.assertEqual(len(directives), 1)
        self.assertEqual(directives, [directive])

    def get_directive(self):
        raise NotImplementedError()


class TestExec(ActionDirectiveTestCase):
    def get_directive(self):
        return Exec('echo foo')

    def test_using_default(self):
        directive = Exec('echo foo')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.cmd, 'echo foo')
        self.assertEqual(action.values, ())
        self.assertEqual(action.using, PLATFORM)

    def test_using_platform(self):
        directive = Exec('echo foo', using = PLATFORM)
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.cmd, 'echo foo')
        self.assertEqual(action.values, ())
        self.assertEqual(action.using, PLATFORM)

    def test_using_virtualizer(self):
        directive = Exec('echo foo', using = VIRTUALIZER)
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.cmd, 'echo foo')
        self.assertEqual(action.values, ())
        self.assertEqual(action.using, VIRTUALIZER)

    def test_missing_required(self):
        self.assertRaises(RebarSyntaxError, Exec)

    def test_unexpected_keyword_argument(self):
        self.assertRaises(RebarSyntaxError, Exec, 'echo foo', bar = 'qux')

manager.add_test_case_class(TestExec)


class TestShell(ActionDirectiveTestCase):
    def get_directive(self):
        return Shell('echo foo')

    def test(self):
        directive = Shell('echo foo')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.cmd, 'echo foo')
        self.assertEqual(action.values, ())

    def test_missing_required(self):
        self.assertRaises(RebarSyntaxError, Shell)

manager.add_test_case_class(TestShell)


class TestCopyTo(ActionDirectiveTestCase):
    def get_directive(self):
        return CopyTo('src', '/dst')

    def test_defaults(self):
        directive = CopyTo('src', '/dst')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, False)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, None)
        self.assertEqual(action.group, None)

    def test_recurse(self):
        directive = CopyTo('src', '/dst', recurse = True)
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, True)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, None)
        self.assertEqual(action.group, None)

    def test_user(self):
        directive = CopyTo('src', '/dst', user = 'foo')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, False)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, 'foo')
        self.assertEqual(action.group, None)

    def test_group(self):
        directive = CopyTo('src', '/dst', group = 'foo')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, False)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, None)
        self.assertEqual(action.group, 'foo')

    def test_user_group(self):
        directive = CopyTo('src', '/dst', user = 'foo', group = 'bar')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, False)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, 'foo')
        self.assertEqual(action.group, 'bar')

    def test_user_group_recurse(self):
        directive = CopyTo(
          'src',
          '/dst',
          user = 'foo',
          group = 'bar',
          recurse = True,
        )
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, 'src')
        self.assertEqual(action.dst, '/dst')
        self.assertEqual(action.recurse, True)
        self.assertEqual(action.dereference, False)
        self.assertEqual(action.replace, False)
        self.assertEqual(action.user, 'foo')
        self.assertEqual(action.group, 'bar')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, CopyTo)
        self.assertRaises(MissingRequiredArgumentError, CopyTo, 'src')

manager.add_test_case_class(TestCopyTo)


class TestCopyFrom(ActionDirectiveTestCase):
    def get_directive(self):
        return CopyFrom('/src', 'dst')

    def test_defaults(self):
        directive = CopyFrom('/src', 'dst')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, '/src')
        self.assertEqual(action.dst, 'dst')
        self.assertEqual(action.recurse, False)
        self.assertEqual(action.dereference, False)

    def test_recurse(self):
        directive = CopyFrom('/src', 'dst', recurse = True)
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.src, '/src')
        self.assertEqual(action.dst, 'dst')
        self.assertEqual(action.recurse, True)
        self.assertEqual(action.dereference, False)

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, CopyFrom)
        self.assertRaises(MissingRequiredArgumentError, CopyFrom, '/src')

manager.add_test_case_class(TestCopyFrom)


# FIXME: Test other options.

class TestPatch(ActionDirectiveTestCase):
    def get_directive(self):
        return Patch('foo.patch', dir = '/bar')

    def test_with_dir(self):
        directive = Patch('foo', dir = '/bar')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.patch, 'foo')
        self.assertEqual(action.dir, '/bar')
        self.assertEqual(action.file, None)
        self.assertEqual(action.strip, 0)

    def test_with_file(self):
        directive = Patch('foo', file = '/bar')
        action = directive.get_build_step(RebarEnvironment())
        self.assertEqual(action.patch, 'foo')
        self.assertEqual(action.dir, None)
        self.assertEqual(action.file, '/bar')
        self.assertEqual(action.strip, 0)

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, Patch)
        self.assertRaises(RebarSyntaxError, Patch, 'patch')

manager.add_test_case_class(TestPatch)


###


class TargetDirectiveTestCase(ActionDirectiveTestCase):
    default_action_group_name = 'targets'


class TestPlatformImage(TargetDirectiveTestCase):
    def get_directive(self):
        return PlatformImage('target_path')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, PlatformImage)

manager.add_test_case_class(TestPlatformImage)


class TestDirImage(TargetDirectiveTestCase):
    def get_directive(self):
        return DirImage('target_path', 'source_dir')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, DirImage)
        self.assertRaises(MissingRequiredArgumentError, DirImage, 'target_path')

manager.add_test_case_class(TestDirImage)


class TestDiskImagingDir(TargetDirectiveTestCase):
    def get_directive(self):
        return DiskImagingDir('target_path', 'source_image')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, DiskImagingDir)
        self.assertRaises(
          MissingRequiredArgumentError,
          DiskImagingDir,
          'root_directory',
        )

manager.add_test_case_class(TestDiskImagingDir)


class TestDiskImagingImage(TargetDirectiveTestCase):
    def get_directive(self):
        return DiskImagingImage('target_path', 'source_image')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, DiskImagingImage)
        self.assertRaises(
          MissingRequiredArgumentError,
          DiskImagingImage,
          'filename',
        )

manager.add_test_case_class(TestDiskImagingImage)


class TestNetworkImagingDir(TargetDirectiveTestCase):
    def get_directive(self):
        return NetworkImagingDir('target_path', 'source_image', 'image_url')

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, NetworkImagingDir)
        self.assertRaises(
          MissingRequiredArgumentError,
          NetworkImagingDir,
          'root_directory',
        )
        self.assertRaises(
          MissingRequiredArgumentError,
          NetworkImagingDir,
          'root_directory',
          'source_image',
        )

manager.add_test_case_class(TestNetworkImagingDir)


class TestTarget(TargetDirectiveTestCase):
    def get_directive(self):
        return Target('target_path', [Exec('echo foo'), Exec('echo bar')])

    def test_target(self):
        directive = Target('target_path', [Exec('echo foo'), Exec('echo bar')])
        target = directive.get_build_step(RebarEnvironment())
        actions = target.subtasks
        self.assertEqual(len(actions), 2)
        self.assertTrue(isinstance(actions[0], ExecAction))
        self.assertTrue(isinstance(actions[1], ExecAction))

    def test_missing_required(self):
        self.assertRaises(MissingRequiredArgumentError, Target)

manager.add_test_case_class(TestTarget)
