# Copyright (c) 2007-2014 Outpost Embedded, LLC
# This file is part of the rebar software package.
#
# rebar is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, tempfile, shutil

from rebar.targets import (
  PlatformImageTarget,
  DirImageTarget,
  DiskImagingDirTarget,
  DiskImagingImageTarget,
  NetworkImagingDirTarget,
)
from rebar.environment import (
  DISK,
  GRUB,
  SYSLINUX,
  MBR,
  VBR,
  EXT2,
  VFAT,
)
from rebar.util import locate_data_file, locate_script

from manager import manager
from common import BuildStepTestCase


class TestPlatformImageTarget(BuildStepTestCase):
    def _test_run(self, location):
        tmpfile = '/tmp/tmp.img'
        mount_point = '/tmp/mnt'
        loopback_device = '/dev/loop0'
        part_device = loopback_device + 'p1'
        offset = 63 * 512
        size = 8193024

        self.virtualizer.shell.result_cache[
          ('mktemp', ())
        ] = (0, tmpfile)

        self.virtualizer.shell.result_cache[
          ('losetup --show --find ?', (tmpfile,))
        ] = (0, loopback_device)

        self.virtualizer.shell.result_cache[
          ('mktemp -d', ())
        ] = (0, mount_point)

        self.virtualizer.shell.result_cache[
          ('du --bytes --summarize /platform', ())
        ] = (0, '1000000 /platform')

        self.virtualizer.shell.result_cache[
          ('chroot ? /usr/bin/dpkg --print-architecture', (mount_point,))
        ] = (0, 'i386')

        target = PlatformImageTarget(
          self.env,
          'target_path',
          geometry = None,
          filesystem = EXT2,
          label = None,
          bootloader = (location and GRUB(location) or GRUB),
          fsexec = (),
          devexec = ('stat ?',),
          prepend_partitions = (),
          type = DISK,
          compression = None,
          name = 'PlatformImage',
        )
        self.run_build_step(target)

        self.assertVirtualizerCommand('du --bytes --summarize /platform')
        self.assertVirtualizerCommand('mktemp')
        self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
        self.assertVirtualizerCommand('sync')
        self.assertVirtualizerCommand('losetup --detach ?', loopback_device)
        self.assertVirtualizerCommand(
          'dd if=/dev/null of=? bs=512 seek=?',
          tmpfile,
          16065,
        )
        self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
        self.assertVirtualizerCommand(
          'dd if=? of=? conv=notrunc',
          '/usr/lib/syslinux/mbr.bin',
          loopback_device,
        )

        self.assertVirtualizerCommand(
          'sfdisk -uS -x -C? -H? -S? ?',
          1,
          255,
          63,
          loopback_device,
          wait = False,
        )
        # The cylinder count here is determined from the size returned by du.
        if location == VBR:
            bootable = ' *'
        else:
            bootable = ''
        self.assertVirtualizerCommand('SENDLINE', '63 16002 83' + bootable)
        for i in range(3):
            self.assertVirtualizerCommand('SENDLINE', '0 0 0')
        self.assertVirtualizerCommand('SENDLINE', 'y')

        self.assertVirtualizerCommand('partx --add ? ?', 1, loopback_device)
        self.assertVirtualizerCommand('mkfs -t ? ?', 'ext2', part_device)
        self.assertVirtualizerCommand('mktemp -d')
        self.assertVirtualizerCommand('mount -n ? ?', part_device, mount_point)

        self.assertVirtualizerCommand(
          "find /platform -mindepth 1 -maxdepth 1 -exec cp --archive --one-file-system '{}' ? \\;",
          mount_point,
        )
        self.assertVirtualizerCommand('umount -n ?', mount_point)
        self.assertVirtualizerCommand('rmdir ?', mount_point)
        self.assertVirtualizerCommand('partx --delete ? ?', 1, loopback_device)

        self.assertVirtualizerCommand('partx --add ? ?', 1, loopback_device)
        self.assertVirtualizerCommand('mktemp -d')
        self.assertVirtualizerCommand('mount -n ? ?', part_device, mount_point)
        self.assertVirtualizerCommand(
          'chroot ? /usr/bin/dpkg --print-architecture',
          mount_point,
        )
        self.assertVirtualizerCommand('mkdir -p ?/boot/grub', mount_point)
        self.assertVirtualizerCommand(
          'cp --archive --one-file-system --no-clobber ?/usr/lib/grub/?/* ?/boot/grub/',
          mount_point,
          'i386-pc',
          mount_point,
        )
        self.assertVirtualizerCommand('umount -n ?', mount_point)
        self.assertVirtualizerCommand('rmdir ?', mount_point)
        self.assertVirtualizerCommand('partx --delete ? ?', 1, loopback_device)

        self.assertVirtualizerCommand('which grub >/dev/null')
        self.assertVirtualizerCommand('sync')
        self.assertVirtualizerCommand('losetup --detach ?', loopback_device)
        self.assertVirtualizerCommand(
          'grub --batch --no-curses --device-map=/dev/null --no-floppy',
          wait = False,
        )
        self.assertVirtualizerCommand('SENDLINE', 'device (hd0) %s' % tmpfile)
        self.assertVirtualizerCommand('SENDLINE', 'geometry (hd0) 1 255 63')
        self.assertVirtualizerCommand('SENDLINE', 'root (hd0,0)')
        if location == VBR:
            grub_device = '(hd0,0)'
        else:
            grub_device = '(hd0)'
        self.assertVirtualizerCommand('SENDLINE', 'setup %s' % grub_device)
        self.assertVirtualizerCommand('SENDLINE', 'quit')

        self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)

        # Applying devexec:
        self.assertVirtualizerCommand('partx --add ? ?', 1, loopback_device)
        self.assertVirtualizerCommand('mktemp -d')
        self.assertVirtualizerCommand('mount -n ? ?', part_device, mount_point)
        self.assertVirtualizerCommand(
          'mount -nt proc proc ?/proc',
          mount_point,
        )
        self.assertVirtualizerCommand(
          'mount -n --bind /dev ?/dev',
          mount_point,
        )
        self.assertVirtualizerCommand('chroot ? /bin/sh', mount_point)
        self.assertVirtualizerCommand('test -b ?', loopback_device)
        self.assertVirtualizerCommand('stat ?', loopback_device)
        self.assertVirtualizerCommand('exit')
        self.assertVirtualizerCommand(
          'sync; umount -n ?/dev || umount -nfl ?/dev',
          mount_point,
          mount_point,
        )
        self.assertVirtualizerCommand(
          'umount -n ?/proc || umount -nfl ?/proc',
          mount_point,
          mount_point,
        )
        self.assertVirtualizerCommand('umount -n ?', mount_point)
        self.assertVirtualizerCommand('rmdir ?', mount_point)
        self.assertVirtualizerCommand('partx --delete ? ?', 1, loopback_device)

        self.assertVirtualizerCommand('sync')
        self.assertVirtualizerCommand('losetup --detach ?', loopback_device)

        self.assertCopyFrom(tmpfile, 'target_path')
        self.assertVirtualizerCommand('rm -f ?', tmpfile)

    def test_run(self):
        self._test_run(None)

    def test_run_mbr(self):
        self._test_run(MBR)

    def test_run_vbr(self):
        self._test_run(VBR)

manager.add_test_case_class(TestPlatformImageTarget)


class TestDirImageTarget(BuildStepTestCase):
    def test_run(self):
        tmpdir = tempfile.mkdtemp()
        try:
            with open(os.path.join(tmpdir, 'foo'), 'w') as f:
                f.write('foo\n')

            tmpfile = '/tmp/tmp.img'
            mount_point = '/tmp/mnt'
            loopback_device = '/dev/loop0'
            part_device = loopback_device + 'p1'
            offset = 63 * 512
            size = 8193024

            self.user_shell.result_cache[
              ('du --bytes --summarize ?', (tmpdir,))
            ] = (0, '1000 %s' % tmpdir)

            self.virtualizer.shell.result_cache[
              ('mktemp', ())
            ] = (0, tmpfile)

            self.virtualizer.shell.result_cache[
              ('losetup --show --find ?', (tmpfile,))
            ] = (0, loopback_device)

            self.virtualizer.shell.result_cache[
              ('mktemp -d', ())
            ] = (0, mount_point)

            target = DirImageTarget(
              self.env,
              'target_path',
              tmpdir,
              geometry = None,
              filesystem = VFAT,
              label = None,
              bootloader = SYSLINUX,
              fsexec = (),
              type = DISK,
              compression = None,
              name = 'DirImage',
            )
            self.run_build_step(target)

            self.assertUserCommand('du --bytes --summarize ?', tmpdir)
            self.assertVirtualizerCommand('mktemp')
            self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
            self.assertVirtualizerCommand('sync')
            self.assertVirtualizerCommand('losetup --detach ?', loopback_device)
            self.assertVirtualizerCommand(
              'dd if=/dev/null of=? bs=512 seek=?',
              tmpfile,
              16065,
            )
            self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
            self.assertVirtualizerCommand(
              'dd if=? of=? conv=notrunc',
              '/usr/lib/syslinux/mbr.bin',
              loopback_device,
            )

            self.assertVirtualizerCommand(
              'sfdisk -uS -x -C? -H? -S? ?',
              1,
              255,
              63,
              loopback_device,
              wait = False,
            )
            # The cylinder count here is determined from the size returned by
            # du.
            self.assertVirtualizerCommand('SENDLINE', '63 16002 b *')
            for i in range(3):
                self.assertVirtualizerCommand('SENDLINE', '0 0 0')
            self.assertVirtualizerCommand('SENDLINE', 'y')

            self.assertVirtualizerCommand('partx --add ? ?', 1, loopback_device)
            self.assertVirtualizerCommand('mkfs -t ? ?', 'vfat', part_device)
            self.assertVirtualizerCommand('syslinux ?', part_device)
            self.assertVirtualizerCommand('mktemp -d')
            self.assertVirtualizerCommand(
              'mount -n ? ?',
              part_device,
              mount_point,
            )

            self.assertCopyTo(
              os.path.join(tmpdir, 'foo'),
              os.path.join(mount_point, 'foo'),
              recurse = True,
            )

            self.assertVirtualizerCommand('umount -n ?', mount_point)
            self.assertVirtualizerCommand('rmdir ?', mount_point)
            self.assertVirtualizerCommand('partx --delete ? ?', 1, loopback_device)
            self.assertVirtualizerCommand('sync')
            self.assertVirtualizerCommand('losetup --detach ?', loopback_device)

            self.assertCopyFrom(tmpfile, 'target_path')
            self.assertVirtualizerCommand('rm -f ?', tmpfile)
        finally:
            shutil.rmtree(tmpdir)

manager.add_test_case_class(TestDirImageTarget)


class TestDiskImagingDirTarget(BuildStepTestCase):
    def test_run(self):
        tmpdir = tempfile.mkdtemp()
        try:
            target = DiskImagingDirTarget(
              self.env,
              tmpdir,
              'source_image',
              name = 'DiskImagingDir',
            )
            self.run_build_step(target)
        finally:
            shutil.rmtree(tmpdir)

        self.assertUserCommand('mkdir -p ?', tmpdir)
        self.assertCopyToPlatform(
          locate_script('gzimg_write'),
          '/etc/initramfs-tools/scripts/gzimg_write',
          dereference = True,
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/disk_imaging.script', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/disk_imaging',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/imaging_functions', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/disk_imaging.hook', ('indep', 'pkg')),
          '/etc/initramfs-tools/hooks/disk_imaging',
        )
        self.assertVirtualizerCommand(
          "sed -i.bak 's/^BOOT=.*$/BOOT=disk_imaging/' "
          '/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/initrd.img',
          os.path.join(tmpdir, 'imaging.ird'),
          dereference = True,
        )
        self.assertVirtualizerCommand(
          'cat /etc/initramfs-tools/initramfs.conf.bak '
          '>/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/initramfs.conf.bak',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/disk_imaging',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/hooks/disk_imaging',
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/vmlinuz',
          os.path.join(tmpdir, 'imaging.vml'),
          dereference = True,
        )
        self.assertUserCommand(
          'cp --link --dereference ? ?',
          'source_image',
          os.path.join(tmpdir, 'source_image'),
        )

manager.add_test_case_class(TestDiskImagingDirTarget)


class TestDiskImagingImageTarget(BuildStepTestCase):
    def test_run(self):
        tmpfile = '/tmp/tmp.img'
        loopback_device = '/dev/loop0'
        part_device = loopback_device + 'p1'
        mount_point = '/tmp/mnt'
        imaging_filenames = [
          'imaging.ird',
          'imaging.vml',
          'source_image',
          'syslinux.cfg',
        ]
        offset = 63 * 512
        size = 8193024

        tmpdir = tempfile.mkdtemp()
        try:
            self.user_shell.result_cache[('mktemp -d', ())] = (0, tmpdir)

            # We return 0 bytes here, which forces a geometry with one
            # cylinder.
            self.user_shell.result_cache[
              ('du --bytes --summarize ?', (tmpdir,))
            ] = (0, '0')

            self.virtualizer.shell.result_cache[('mktemp', ())] = (0, tmpfile)

            self.virtualizer.shell.result_cache[
              ('losetup --show --find ?', (tmpfile,))
            ] = (0, loopback_device)

            self.virtualizer.shell.result_cache[
              ('mktemp -d', ())
            ] = (0, mount_point)

            self.user_shell.result_cache[
              ('find ? -mindepth 1 -maxdepth 1 -print0', ('%s/' % tmpdir,))
            ] = (0, '%s\0' % '\0'.join([
              os.path.join(tmpdir, filename) for filename in imaging_filenames
            ]))

            target = DiskImagingImageTarget(
              self.env,
              'target_path',
              'source_image',
              geometry = None,
              filesystem = VFAT,
              label = None,
              bootloader = SYSLINUX,
              fsexec = (),
              type = DISK,
              compression = None,
              name = 'DiskImagingImage',
            )
            self.run_build_step(target)

        finally:
            shutil.rmtree(tmpdir)

        self.assertUserCommand('mktemp -d')
        self.assertCopyToPlatform(
          locate_script('gzimg_write'),
          '/etc/initramfs-tools/scripts/gzimg_write',
          dereference = True,
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/disk_imaging.script', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/disk_imaging',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/imaging_functions', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/disk_imaging.hook', ('indep', 'pkg')),
          '/etc/initramfs-tools/hooks/disk_imaging',
        )
        self.assertVirtualizerCommand(
          "sed -i.bak 's/^BOOT=.*$/BOOT=disk_imaging/' "
          '/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/initrd.img',
          os.path.join(tmpdir, 'imaging.ird'),
          dereference = True,
        )
        self.assertVirtualizerCommand(
          'cat /etc/initramfs-tools/initramfs.conf.bak '
          '>/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/initramfs.conf.bak',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/disk_imaging',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/hooks/disk_imaging',
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/vmlinuz',
          os.path.join(tmpdir, 'imaging.vml'),
          dereference = True,
        )
        self.assertUserCommand(
          'cp --link --dereference ? ?',
          'source_image',
          os.path.join(tmpdir, 'source_image'),
        )
        self.assertUserCommand('du --bytes --summarize ?', tmpdir)
        self.assertVirtualizerCommand('mktemp')
        self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
        self.assertVirtualizerCommand('sync')
        self.assertVirtualizerCommand('losetup --detach ?', loopback_device)
        self.assertVirtualizerCommand(
          'dd if=/dev/null of=? bs=512 seek=?',
          tmpfile,
          16065,
        )
        self.assertVirtualizerCommand('losetup --show --find ?', tmpfile)
        self.assertVirtualizerCommand(
          'dd if=? of=? conv=notrunc',
          '/usr/lib/syslinux/mbr.bin',
          loopback_device,
        )
        self.assertVirtualizerCommand(
          'sfdisk -uS -x -C? -H? -S? ?',
          1,
          255,
          63,
          loopback_device,
          wait = False,
        )
        self.assertVirtualizerCommand('SENDLINE', '63 16002 b *')
        for i in range(3):
            self.assertVirtualizerCommand('SENDLINE', '0 0 0')
        self.assertVirtualizerCommand('SENDLINE', 'y')
        self.assertVirtualizerCommand('partx --add ? ?', 1, loopback_device)
        self.assertVirtualizerCommand('mkfs -t ? ?', 'vfat', part_device)
        self.assertVirtualizerCommand('syslinux ?', part_device)
        self.assertVirtualizerCommand('mktemp -d')
        self.assertVirtualizerCommand('mount -n ? ?', part_device, mount_point)
        self.assertUserCommand(
          'find ? -mindepth 1 -maxdepth 1 -print0',
          ('%s/' % tmpdir),
        )

        for filename in imaging_filenames:
            self.assertCopyTo(
              os.path.join(tmpdir, filename),
              os.path.join(mount_point, filename),
              recurse = True,
            )

        self.assertVirtualizerCommand('umount -n ?', mount_point)
        self.assertVirtualizerCommand('rmdir ?', mount_point)
        self.assertVirtualizerCommand('partx --delete ? ?', 1, loopback_device)
        self.assertVirtualizerCommand('sync')
        self.assertVirtualizerCommand('losetup --detach ?', loopback_device)

        self.assertCopyFrom(tmpfile, 'target_path')

        self.assertVirtualizerCommand('rm -f ?', tmpfile)
        self.assertUserCommand('rm -fR ?', tmpdir)

manager.add_test_case_class(TestDiskImagingImageTarget)


class TestNetworkImagingDirTarget(BuildStepTestCase):
    def test_run(self):
        tmpdir = tempfile.mkdtemp()
        try:
            target = NetworkImagingDirTarget(
              self.env,
              tmpdir,
              'source_image',
              'image_url',
              name = 'NetworkImagingDir',
            )
            self.run_build_step(target)
        finally:
            shutil.rmtree(tmpdir)

        self.assertUserCommand('mkdir -p ?', tmpdir)
        self.assertCopyToPlatform(
          locate_script('gzimg_write'),
          '/etc/initramfs-tools/scripts/gzimg_write',
          dereference = True,
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/network_imaging.script', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/network_imaging',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/imaging_functions', ('indep', 'pkg')),
          '/etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertCopyToPlatform(
          locate_data_file('imaging/network_imaging.hook', ('indep', 'pkg')),
          '/etc/initramfs-tools/hooks/network_imaging',
        )
        self.assertVirtualizerCommand(
          "sed -i.bak 's/^BOOT=.*$/BOOT=network_imaging/' "
          '/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/initrd.img',
          os.path.join(tmpdir, 'imaging.ird'),
          dereference = True,
        )
        self.assertVirtualizerCommand(
          'cat /etc/initramfs-tools/initramfs.conf.bak '
          '>/etc/initramfs-tools/initramfs.conf'
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/initramfs.conf.bak',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/imaging_functions',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/scripts/network_imaging',
        )
        self.assertVirtualizerCommand(
          'rm /etc/initramfs-tools/hooks/network_imaging',
        )
        self.assertVirtualizerCommand('update-initramfs -u')
        self.assertCopyFromPlatform(
          '/vmlinuz',
          os.path.join(tmpdir, 'imaging.vml'),
          dereference = True,
        )
        self.assertUserCommand(
          'cp --link --dereference ? ?',
          'source_image',
          os.path.join(tmpdir, 'image_url'),
        )

manager.add_test_case_class(TestNetworkImagingDirTarget)
